unit pAssem;

interface
uses pCommon;

{ the following are used for method export }

type
  assemblytable =
    array [1..maxlabel] of integer;
var
  table: assemblytable;

function templength(labelno:
  integer): integer;

{ these are used in compiler}

procedure assemble;
procedure initassembler;

implementation
{
uses pCommon;
}

function templength(labelno:
  integer): integer;
begin
  { include block link
    (or process state)
    of length 5 (or 4) }
{WAS    of length 4 (or 3) }
  templength :=
    table[labelno] + arl
end;

procedure assemble;
type
  operations = set of minoperation
    ..maxoperation;
{
  assemblytable =
    array [1..maxlabel] of integer;
}
var
  labelno, address, op, arg1, arg2,
  arg3, arg4, arg5, arg6: integer;
  realarg: real;
  stringarg: stringg;
  casearg: casetable;
  idarg: identifier;
{
  table: assemblytable;
}

procedure nextinstruction;
begin
  get(op);
  if op in noarguments then
    { skip }
  else if op in oneargument then
    get(arg1)
  else if op in twoarguments then
    begin
      get(arg1); get(arg2)
    end
  else if op in threearguments then
    begin
      get(arg1); get(arg2);
      get(arg3)
    end
  else if op in fourarguments then
    begin
      get(arg1); get(arg2);
      get(arg3); get(arg4)
    end
  else if op in sixarguments then
    begin
      get(arg1); get(arg2);
      get(arg3); get(arg4);
      get(arg5); get(arg6);
    end
  else if op = realconst2 then
    getreal(realarg)
  else if op = stringconst2 then
    getstring(arg1, stringarg)
  else if op = case2 then
    getcase(arg1, arg2,
      casearg)
  else if op in descriptions then
    getmoduledesc(idarg, arg1, arg2)
  else haltt(implementation5)
end;

procedure emit1(op: integer);
begin
  debugarguments(op, 0);
  put(op);
  address := address + 1
end;

procedure emit2(
  op, arg: integer);
begin
  debugarguments(op, 1);
  put(op); put(arg);
  address := address + 2
end;

procedure emit3(
  op, arg1, arg2: integer);
begin
  debugarguments(op, 2);
  put(op); put(arg1);
  put(arg2);
  address := address + 3
end;

procedure emit4(
  op, arg1, arg2,
  arg3: integer);
begin
  debugarguments(op, 3);
  put(op); put(arg1);
  put(arg2); put(arg3);
  address := address + 4
end;

procedure emit5(
  op, arg1, arg2, arg3,
  arg4: integer);
begin
  debugarguments(op, 4);
  put(op); put(arg1);
  put(arg2); put(arg3);  
  put(arg4);
  address := address + 5
end;

procedure emit7(
  op, arg1, arg2, arg3, arg4,
  arg5, arg6: integer);
begin
  debugarguments(op, 6);
  put(op); put(arg1);
  put(arg2); put(arg3);
  put(arg4); put(arg5);
  put(arg6);
  address := address + 7
end;

procedure emitreal(
  value: real);
begin
  put(realconst2);
  putreal(value);
  address := address + 3
end;

procedure emitstring(
  length: integer;
  value: stringg);
begin
  put(stringconst2);
  putstring(length, value);
  address :=
    address + length + 2
end;

procedure emitcase(
  lineno, length: integer;
  var table: casetable);
begin
  put(case2);
  putcase(lineno, length,
    table);
  address :=
    address + 2*length + 3
end;

procedure emitmoduledesc(
  op: integer;
  id: identifier;
  key, lineno: integer);
begin
  put(op); putmoduledesc(id, key, lineno);
end;

function optimize(condition:
  boolean): boolean;
begin
  optimize :=
    optimizing and condition
end;

function jumpdispl(labelno:
  integer): integer;
begin
  jumpdispl :=
    table[labelno] - address
end;

procedure assign(
  length: integer);
begin
  if optimize(length = 1)
    then emit1(ordassign2)
  else if optimize(length = 2)
    then emit1(realassign2)
  else emit2(assign2, length);
  nextinstruction
end;

procedure casex(
  lineno, length: integer;
  table: casetable);
var i: integer;
begin
  for i := 1 to length do
    table[i].index :=
      jumpdispl(
        table[i].index);
  emitcase(lineno, length,
    table);
  nextinstruction
end;

procedure defaddr(labelno:
  integer);
begin
  table[labelno] := address;
  nextinstruction
end;

procedure defarg(labelno,
  value: integer);
begin
  table[labelno] := value;
  nextinstruction
end;

procedure endprocess(exitlabel,
  lineno: integer);
begin
  emit3(endprocess2,
    jumpdispl(exitlabel),
    lineno);
  nextinstruction
end;

procedure field(displ: integer);
begin
  if optimize(displ = 0)
    then { empty }
    else emit2(field2, displ);
  nextinstruction
end;

procedure forall(blockno,
  templabel, endlabel, lineno: integer);
begin
  emit5(forall2, blockno,
    templength(templabel),
    jumpdispl(endlabel),
    lineno);
  nextinstruction
end;

procedure jump(op, labelno:
  integer);
begin
  { op in [do2, downto2,
      enddown2, endto2,
        goto2, to2] }
  emit2(op,
    jumpdispl(labelno));
  nextinstruction
end;

procedure proccall(level,
  labelno: integer);
var displ: integer;
begin
  displ := jumpdispl(labelno);
  if optimize(level = 1) then
    emit2(globalcall2, displ)
  else
    emit3(mcall2, level,
      displ);
  nextinstruction
end;

procedure method(blockno,
  paramlength, varlabel, templabel,
  beginlabel, lineno: integer);
begin
  emit7(method2,
    blockno,
    paramlength,
    table[varlabel],
    templength(templabel),
    jumpdispl(beginlabel),
    lineno);
  nextinstruction
end;

procedure process(blockno,
  templabel, endlabel, lineno: integer);
begin
  emit5(process2, blockno,
    templength(templabel),
    jumpdispl(endlabel),
    lineno);
  nextinstruction
end;

procedure moduleinit(varlabel,
  templabel, beginlabel,
  lineno: integer);
begin
  emit5(moduleinit2,
    table[varlabel],
    templength(templabel),
    jumpdispl(beginlabel),
    lineno);
  nextinstruction
end;

procedure moduleenter(varlabel: integer);
begin
  emit2(moduleenter2,
    table[varlabel]);
  nextinstruction
end;

procedure value(length: integer);
begin
  if optimize(length = 1)
    then emit1(ordvalue2)
  else if optimize(length = 2)
    then emit1(realvalue2)
  else emit2(value2, length);
  nextinstruction
end;

procedure variable(level, displ:
  integer);
begin
{WAS:
  if displ >= 0 then
    / include block number
  block link of length 4
  displ := displ + 4;
}
  if displ < 0 then
    error(implementation5);
  nextinstruction;
  while optimize(op = field2) do
    begin
      displ := displ + arg1;
      nextinstruction
    end;
  if optimize(level = 0) then
    if (op = value2) and
      (arg1 = 1) then
        begin
          emit2(localvalue2,
            displ);
          nextinstruction
        end
    else if (op = value2) and
      (arg1 = 2) then
        begin
          emit2(localreal2,
            displ);
          nextinstruction
        end
    else emit2(localvar2, displ)
  else if optimize(level = 1) then
    if (op = value2) and
      (arg1 = 1) then
        begin
          emit2(globalvalue2,
            displ);
          nextinstruction
        end
    else emit2(globalvar2, displ)
  else
    emit3(variable2, level, displ)
end;

procedure modulehead(
  var moduleid: identifier;
  keylabel, lineno: integer);
begin
  emitmoduledesc(modulehead2,
    moduleid, table[keylabel], lineno);
  nextinstruction
end;

procedure xternal(op,
  mno, displ: integer);
begin
  if displ < 0 then
    error(implementation5);
{ include block link
      of length arl = 5 }
    displ := displ + arl;
  emit3(op, mno, displ);
  nextinstruction
end;

procedure imethod(
  proclabel: integer);
begin
  proclabel := jumpdispl(proclabel);
  emit2(installm2, proclabel);
  nextinstruction
end;

procedure copyinstruction;
begin
  if op in noarguments
    then emit1(op)
  else if op in oneargument
    then emit2(op, arg1)
  else if op in twoarguments
    then emit3(op, arg1, arg2)
  else if op in threearguments
    then
      emit4(op, arg1, arg2,
        arg3)
  else if op in fourarguments
    then
      emit5(op, arg1, arg2,
        arg3, arg4)
  else if op in sixarguments
    then
      emit7(op, arg1, arg2,
        arg3, arg4, arg5, arg6)
  else if op = realconst2
    then emitreal(realarg)
  else if op = stringconst2 then
    emitstring(arg1,
      stringarg)
  else if op in descriptions then
    emitmoduledesc(op, idarg, arg1, arg2)
  else
    haltt(implementation5);
  nextinstruction
end;

procedure assm;
begin
  address := 0;
  nextinstruction;
  while op <> endmodule2 do
    case op of
      assign2         :    assign(arg1);
      case2           :    casex(arg1, arg2, casearg);
      defaddr2        :    defaddr(arg1);
      defarg2         :    defarg(arg1, arg2);
      endprocess2     :    endprocess(arg1, arg2);
      field2          :    field(arg1);
      forall2         :    forall(arg1, arg2, arg3, arg4);
      do2,
      downto2,
      enddown2,
      endto2,
      goto2,
      to2             :     jump(op, arg1);
      mcall2          :     proccall(arg1, arg2);
      method2         :     method(arg1, arg2, arg3,
                                 arg4, arg5, arg6);
      process2        :     process(arg1, arg2, arg3, arg4);
      value2          :     value(arg1);
      variable2       :     variable(arg1, arg2);

      modulehead2    :     modulehead(idarg, arg1, arg2);
      moduleenter2   :     moduleenter(arg1);
      moduleinit2    :     moduleinit(arg1, arg2, arg3, arg4);
      xvariable2,
      xmcall2         :     xternal(op, arg1, arg2);
      installm2       :     imethod(arg1);
    else
      copyinstruction;
    end {case};
  emit1(endmodule2)
end;

begin
  for labelno := 1 to maxlabel
    do table[labelno] := 0;
  assm;
  rerun;
  assm;
end;

procedure initassembler;
begin end;

end.