{        PARADIGM/SP INTERPRETERER
                MAIN PROGRAM
              JANUARY 20, 1998
}

program pInter;

uses dos, pCommon;
{ dos is a nonstandard unit providing
  a procedure to read cpu time }

type
  address = 1 .. maxaddr;
  store =
    array [address{1..maxaddr}] of
      integer;
  blocktable =
    array [1..maxblock] of integer;
  imptAddrTable =
    array [0..maximps] of address {integer};
  tags = 
    (arraytag, classtag, xclasstag, channeltag);

  importrecord =
    record
      id: identifier;
      key: integer;
      isbase: boolean;
      baseaddr: address;
    end;
  imptRecTable =
    array [0..maximps] of importrecord;


var
  { permanent variables }
  s, v, t: integer;
  running, correct: boolean;
  stackbottom, blocknumber: integer;
  cmax, ready: integer;
  free: blocktable;
  st: store;
  impentry, impinit: imptAddrTable;
  t1, t2: real; { starting / ending time }

  glbimports: imptRecTable;
  nofglbimps: integer;

procedure run(thismno, p, b: integer);

type
  channeltable =
    array [1..maxchan] of integer;

var
  { permanent variables }
  blockno: integer;
  open: channeltable;

  { temporary variables }
  bi, vi, c, i,
  j, k, length, level,
  lineno, lower, m, n,
  pi, si, templength,
  typeno, typeno2, upper,
  width, x, y, z,
  mno, tag, t0, v0
    : integer;
  dx, dy: dualreal;
  sx: stringg;
  cx: char;
  new, dynamic: boolean;

  {DEBUG: remove after debugging} p0, st0: integer;

  { method activation record, its length is arl:

    s      staticLink       method call
    s+1    link to prevBlk  method
    s+2    returnAddress    method call
    s+3    dynamicLink1, b  method
    s+4    dynamicLink2, v  method
  }

  {        module activation record:

           s      own base address           load
           s+1    start of code              load
           s+2    this module's number, mno  load /tentative/
           s+3    unused
           s+4    variables, v               load

           s+...  import bases
  }

{ Local procedures in run }

procedure error1(
  lineno: integer;
  kind: phrase);
var
  mno: integer;
begin
  mno := st[b + 2];
  write(kind, sp,
        'line ', lineno:1, sp,
        'module ', glbimports[mno].id);
  running := false; correct := false
end;

procedure memorylimit(
  lineno: integer);
begin
  error1(lineno, maxaddr5)
end;

function newblock1: integer;
begin
  if blocknumber = maxblock then
    error1(lineno, maxblock5);
  blocknumber := blocknumber + 1;
  newblock1 := blocknumber;
end;

procedure rangeerror(
  lineno: integer);
begin
  error1(lineno, range4)
end;

procedure activate(
  bvalue, svalue, pvalue:
    integer);
begin
  svalue := svalue + 3;
  st[svalue - 2] := pvalue;
  st[svalue - 1] := bvalue;
  st[svalue] := ready;
  ready := svalue
end;

procedure select(
  lineno: integer);
begin
  if ready = 0 then
    error1(lineno, deadlock4)
  else
    begin
      s := ready;
      ready := st[s];
      b := st[s - 1];
      p := st[s - 2];
      v := st[b + 4];
      s := s - 3;
    end
end;

procedure popstring(
  var value: stringg);
var i: integer;
begin
  s := s - maxstring;
  for i := 1 to maxstring do
    value[i] := chr(st[s + i])
end;

procedure dassign(
  target, source: integer
);
var
  i: integer;
begin
  x  := target + st[target];  {deref target}
  y  := source + st[source];  {deref source (allocated by dvalue2)}
  z := st[y - 1];             {source description}
  blockno := st[z + 1];
  length  := st[z + 2];       {block length}
{DEBUG}
  if z <> st[x - 1] then
    error1(0, implementation5);
{DEBUG
  writeln('dassign2, source, target, descr: ',
    y: 8, x: 8, z:8);
  readln;
}
  {copy block starting at y
  onto block starting at x}
  for i := 0 to
    length - 2 do {descriptor excluded}
      st[x + i] :=
         st[y + i];
   {deallocate source}
   st[y - 1] := free[blockno];
   free[blockno] := y - 1;
end;

function blength(
  descr: integer): integer;
var
  len, fld, fldref: integer;

begin
  tag := st[descr];
  if (tag = ord(xclasstag)) or
     (tag = ord(classtag)) then
    begin
      len := 0;
      fldref := descr + 5; {ref to first fld descr}
      for fld := 1 to st[descr + 4] {number of fields} do
        begin
          len := len +
            blength(
              st[fldref + 1]);  {field descriptor}
          fldref := fldref + 2;
        end;
    end
  else if tag = ord(arraytag) then
    len := st[descr + 3]          {array length}
      * blength(st[descr + 4])    {elem descr}
  else begin
    error1(lineno, implementation5);
    len := 0;
  end;
  blength := 
    len
    + st[descr + 3] {type length}
    + 1;            {ref to descriptor}
end;

procedure construct(
  addr,
  descr: integer);
var comp, ref, len: integer;

begin {construct}
  if correct then begin
    x := t0 + 1; {beginning of block}
    t0 := x + st[descr + 3]; {length};
    st[x] := descr;
    x := x + 1;
    st[addr] := x - addr;
    addr := x;

    tag := st[descr];
    if (tag = ord(xclasstag)) or
       ( tag = ord(classtag))
      then {construct fields}
        begin
          ref := descr + 5; {ref to first fld descr}
          for comp := 1 to st[descr + 4] {number of fields} do
            begin
              construct(
                addr + st[ref],   {field address}
                st[ref + 1]       {field descriptor}
              );
              ref := ref + 2;
            end;
        end
    else if tag = ord(arraytag)
      then {construct elements}
        begin
          len := st[descr + 3];   {array length}
          ref := st[descr + 4];   {elem descr ref}
          comp := len - 1;        {(last) elem displ}
          while comp >= 0 do
            begin
              construct(
                addr + comp,      {elem addr}
                ref               {elem descr}
              );
              comp := comp - 1;
            end;
        end
    else
      error1(0, implementation5);
  end; {correct}
end; {construct}

procedure allocate(
  descr: integer;
  var addr: integer;
  var new: boolean;
  lineno: integer);
begin
  blockno := st[descr + 1];
  length  := st[descr + 2];   {block length}
  addr := free[blockno];
  new := addr = 0;
  if new then begin
  addr := t + 1;
  t := t + length;
  if t > maxaddr then
    memorylimit(lineno);
  end else
    free[blockno] := st[addr];
end;

procedure debug0;
begin
  if (p <=0) or (p > maxaddr) then begin
    writeln('address out of range; p, p0, st0 = ',
      p:5, p0:5, st0:5);
    error1(lineno, implementation5);
  end else if (st[p] < minoperation)
      or (st[p] > maxoperation) then  begin
    writeln('operation out of range; p, st[p], p0, st0 = ',
      p:5, st[p]:5, p0:5, st0:5);
    error1(lineno, implementation5);
  end;
  p0 := p; st0 := st[p];
end;

procedure xmcall; {mno, descdispl}
        begin
          s := s + 1;
          st[s + 2] := p + 3;           { return address  }
          x := b;
          while st[x] <> x do           { st[objbase] = objbase }
            x := st[x];                 { find base of object   }
{DEBUG
          writeln('xproccall, base of module: ', x);readln;
}
          vi := x + arl{ = 5};
          y := st[vi + st[p + 1]];   { base of descriptor   }

{WAS          y := st[x + 4 + st[p + 1]];   / base of descriptor   }

{DEBUG
          writeln('xmcall, base of descriptor owner: ', y);readln;
}
          z := y + st[p + 2];           { descriptor's address   }
          st[s] := st[z];               { base of method }
{DEBUG
          writeln('xmcall, base of method: ', st[s]);readln;
}
          p := st[z + 1];               { xproc's address }
{DEBUG
          writeln('xmcall, entry of method: ', p); readln;
}
        end;

procedure imcall; {mno, methoddispl}

        begin
          s := s + 1;
          st[s + 2] := p + 3;           { return address  }
          x := b;
          while st[x] <> x do           { st[objbase] = objbase }
            x := st[x];                 { find base of object   }
{DEBUG
          writeln('imcall, base of object: ', x);readln;
}
          vi := x + arl{ = 5};
          y := st[vi + st[p + 1]];      { base of method owner}
{WAS:          y := st[x + 4 + st[p + 1]];   / base of method owner}
          st[s] := y;
{DEBUG
          writeln('imcall, base of method owner: ', y); readln;
}
          p := st[y + 1] + st[p + 2];           { proc's address   }
{DEBUG
          writeln('imcall, entry of method: ', p);readln;
}
        end;

procedure process;
          {blockno, templength, displ, lineno}
        begin
          st[s] := st[s] + 1;
          blockno := st[p + 1];
          bi := free[blockno];
          if bi = 0 then
            begin
              bi := t + 1;
              t :=
                bi + st[p + 2] + arl{ = 5};
{WAS:                bi + st[p + 2] + 4;}
            end
          else
            free[blockno] :=
              st[bi];
          if t > maxaddr then
            memorylimit(st[p + 4])
          else
            begin
              st[bi] := b;
              st[bi + 1] := s;
{WAS set by select before: }
              st[bi + 4] := v;
              si := bi + arl { = 5};
{WAS:             si := bi + 4;}
              st[si] := blockno;
              activate(bi, si,
                p + 5);
              p := p + st[p + 3]
            end
        end;

procedure endprocess;
        begin
          blockno := st[s];
          x := b;
          b := st[x];
          v := st[b + 4];
          s := st[x + 1];
          st[x] := free[blockno];
          free[blockno] := x;
          st[s] := st[s] - 1;
          if st[s] > 0 then
            select(st[p + 2])
          else
            begin
              s := s - 1;
              p := p + st[p + 1]
            end
        end;

begin { run }
  running := correct;
  while running do begin

    debug0; {DEBUG: remove after debugging}

    case st[p] of

      variable2{level, displ}:
        begin
          level := st[p + 1];
          s := s + 1;
          x := b;
          while level > 0 do
            begin
              x := st[x];
              level := level - 1
            end;
          st[s] :=
            st[x + 4]   {variable base}
            + st[p + 2] {displacement};
{WAS:          st[s] := x + st[p + 2];}
          p := p + 3
        end;

      valueparam2{level, displ}:
        begin
          level := st[p + 1];
          s := s + 1;
          x := b;
          while level > 0 do
            begin
              x := st[x];
              level := level - 1
            end;
          st[s] := x + st[p + 2];
          p := p + 3
        end;

      xvariable2{mno, displ}:
        begin
          s := s + 1;
          x := b;
          while st[x] <> x do           { st[objbase] = objbase }
            x := st[x];                 { find base of object   }
          vi := x + arl{ = 5};
          y :=  st[vi + st[p + 1]];     { base of owner  }
{WAS:          y :=  st[x + 4 + st[p + 1]];  / base of owner  }
{          vi := y + arl;}
          st[s] := y + st[p + 2];       { base + diplacement    }
          p := p + 3
        end;

      varparam2{level, displ}:
        begin
          level := st[p + 1];
          s := s + 1;
          x := b;
          while level > 0 do
            begin
              x := st[x];
              level := level - 1
            end;
          st[s] := st[x + st[p + 2]];
          p := p + 3
        end;

      index2{lower, upper, length,
          lineno}:
        begin
          lower := st[p + 1];
          i := st[s];
          s := s - 1;
          if (i < lower) or
            (i > st[p + 2]) then
              rangeerror(st[p + 4])
          else
            begin
              st[s] := st[s] +
                (i - lower) *
                  st[p + 3];
               p := p + 5
            end
        end;

      field2{displ}:
        begin
          st[s] :=
            st[s] + st[p + 1];
          p := p + 2
        end;

      eof2{lineno}:
        begin
          s := s + 1;
          st[s] :=
            ord(eof(input));
          p := p + 2
        end;

      eoln2{lineno}:
        begin
          s := s + 1;
          st[s] :=
            ord(eoln(input));
          p := p + 2
        end;

      float2:
        begin
          dx.a := st[s];
          s := s + 1;
          st[s - 1] := dx.b;
          st[s] := dx.c;
          p := p + 1
        end;

      abs2{lineno}:
        begin
          dx.b := st[s - 1];
          dx.c := st[s];
          dx.a := abs(dx.a);
          st[s - 1] := dx.b;
          st[s] := dx.c;
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      absint2{lineno}:
        begin
          st[s] := abs(st[s]);
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      arctan2{lineno}:
        begin
          dx.b := st[s - 1];
          dx.c := st[s];
          dx.a := arctan(dx.a);
          st[s - 1] := dx.b;
          st[s] := dx.c;
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      chr2{lineno}:
        begin
          x := st[s];
          if (x < null) or (x > del)
            then
              rangeerror(st[p + 1])
            else p := p + 2
        end;

      cos2{lineno}:
        begin
          dx.b := st[s - 1];
          dx.c := st[s];
          dx.a := cos(dx.a);
          st[s - 1] := dx.b;
          st[s] := dx.c;
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      exp2{lineno}:
        begin
          dx.b := st[s - 1];
          dx.c := st[s];
          dx.a := exp(dx.a);
          st[s - 1] := dx.b;
          st[s] := dx.c;
        { if overflow then
             rangeerror(st[p + 1])
          else } p := p + 2
        end;

      ln2{lineno}:
        begin
          dx.b := st[s - 1];
          dx.c := st[s];
          dx.a := ln(dx.a);
          st[s - 1] := dx.b;
          st[s] := dx.c;
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      odd2:
        begin
          st[s] := ord(odd(st[s]));
          p := p + 1
        end;

      pred2{minvalue, lineno}:
        begin
          if st[s] > st[p + 1] then
            begin
              st[s] := pred(st[s]);
              p := p + 3
            end
          else
            rangeerror(st[p + 2])
        end;

      round2{lineno}:
        begin
          dx.b := st[s - 1];
          dx.c := st[s];
          s := s - 1;
          st[s] := round(dx.a);
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      sin2{lineno}:
        begin
          dx.b := st[s - 1];
          dx.c := st[s];
          dx.a := sin(dx.a);
          st[s - 1] := dx.b;
          st[s] := dx.c;
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      sqr2{lineno}:
        begin
          dx.b := st[s - 1];
          dx.c := st[s];
          dx.a := sqr(dx.a);
          st[s - 1] := dx.b;
          st[s] := dx.c;
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      sqrint2{lineno}:
        begin
          st[s] := sqr(st[s]);
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      sqrt2{lineno}:
        begin
          dx.b := st[s - 1];
          dx.c := st[s];
          dx.a := sqrt(dx.a);
          st[s - 1] := dx.b;
          st[s] := dx.c;
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      succ2{maxvalue, lineno}:
        begin
          if st[s] < st[p + 1] then
            begin
              st[s] := succ(st[s]);
              p := p + 3
            end
          else
            rangeerror(st[p + 2])
        end;

      trunc2{lineno}:
        begin
          dx.b := st[s - 1];
          dx.c := st[s];
          s := s - 1;
          st[s] := trunc(dx.a);
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      result2{length}:
        begin
          s := s + st[p + 1];
          p := p + 2
        end;

      mcall2{level, displ}:
        begin
          level := st[p + 1];
          s := s + 1;
          x := b;
          while level > 0 do
            begin
              x := st[x];
              level := level - 1
            end;
          st[s] := x;
          st[s + 2] := p + 3;
          p := p + st[p + 2]
        end;

      xmcall2{mno, descdispl}:
        xmcall;

      imcall2{mno, methoddispl}:
        imcall;

      ordconst2{value}:
        begin
          s := s + 1;
          st[s] := st[p + 1];
          p := p + 2
        end;

      realconst2{value}:
        begin
          st[s + 1] := st[p + 1];
          st[s + 2] := st[p + 2];
          s := s + 2;
          p := p + 3
        end;

      stringconst2{length, value}:
        begin
          length := st[p + 1];
          for i := 1 to length do
            st[s + i] :=
              st[p + i + 1];
          for i := length + 1
            to maxstring do
              st[s + i] := null;
          s := s + maxstring;
          p := p + length + 2
        end;

      value2{length}:
        begin
          length := st[p + 1];
          x := st[s];
          for i := 0 to
            length - 1 do
              st[s + i] :=
                st[x + i];
          s := s + length - 1;
          p := p + 2
        end;

      not2:
        begin
          if st[s] = ord(true)
            then
              st[s] := ord(false)
            else
              st[s] := ord(true);
          p := p + 1
        end;

      floatleft2:
        begin
          dx.a := st[s - 2];
          s := s + 1;
          st[s] := st[s - 1];
          st[s - 1] := st[s - 2];
          st[s - 3] := dx.b;
          st[s - 2] := dx.c;
          p := p + 1
        end;

      multiply2{lineno}:
        begin
          s := s - 1;
          st[s] :=
            st[s] * st[s + 1];
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      divide2{lineno}:
        begin
          s := s - 1;
          st[s] :=
            st[s] div st[s + 1];
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      modulo2{lineno}:
        begin
          s := s - 1;
          st[s] :=
            st[s] mod st[s + 1];
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      multreal2{lineno}:
        begin
          dy.b := st[s - 1];
          dy.c := st[s];
          s := s - 2;
          dx.b := st[s - 1];
          dx.c := st[s];
          dx.a := dx.a * dy.a;
          st[s - 1] := dx.b;
          st[s] := dx.c;
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      divreal2{lineno}:
        begin
          dy.b := st[s - 1];
          dy.c := st[s];
          s := s - 2;
          dx.b := st[s - 1];
          dx.c := st[s];
          dx.a := dx.a / dy.a;
          st[s - 1] := dx.b;
          st[s] := dx.c;
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      and2:
        begin
          s := s - 1;
          if st[s] = ord(true)
            then
              st[s] := st[s + 1];
          p := p + 1
        end;

      minus2{(lineno}:
        begin
          st[s] := - st[s];
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      minusreal2{lineno}:
        begin
          dx.b := st[s - 1];
          dx.c := st[s];
          dx.a := - dx.a;
          st[s - 1] := dx.b;
          st[s] := dx.c;
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      add2{lineno}:
        begin
          s := s - 1;
          st[s] :=
            st[s] + st[s + 1];
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      subtract2{lineno}:
        begin
          s := s - 1;
          st[s] :=
            st[s] - st[s + 1];
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      addreal2{lineno}:
        begin
          dy.b := st[s - 1];
          dy.c := st[s];
          s := s - 2;
          dx.b := st[s - 1];
          dx.c := st[s];
          dx.a := dx.a + dy.a;
          st[s - 1] := dx.b;
          st[s] := dx.c;
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      subreal2{lineno}:
        begin
          dy.b := st[s - 1];
          dy.c := st[s];
          s := s - 2;
          dx.b := st[s - 1];
          dx.c := st[s];
          dx.a := dx.a - dy.a;
          st[s - 1] := dx.b;
          st[s] := dx.c;
        { if overflow then
            rangeerror(st[p + 1])
          else } p := p + 2
        end;

      or2:
        begin
          s := s - 1;
          if st[s] = ord(false) then
            st[s] := st[s + 1];
          p := p + 1
        end;

      lsord2:
        begin
          s := s - 1;
          st[s] :=
            ord(st[s] < st[s + 1]);
          p := p + 1
        end;

      eqord2:
        begin
          s := s - 1;
          st[s] :=
            ord(st[s] = st[s + 1]);
          p := p + 1
        end;

      grord2:
        begin
          s := s - 1;
          st[s] :=
            ord(st[s] > st[s + 1]);
          p := p + 1
        end;

      ngord2:
        begin
          s := s - 1;
          st[s] :=
            ord(st[s] <= st[s + 1]);
          p := p + 1
        end;

      neord2:
        begin
          s := s - 1;
          st[s] :=
            ord(st[s] <> st[s + 1]);
          p := p + 1
        end;

      nlord2:
        begin
          s := s - 1;
          st[s] :=
            ord(st[s] >= st[s + 1]);
          p := p + 1
        end;

      lsreal2:
        begin
          dy.b := st[s - 1];
          dy.c := st[s];
          s := s - 3;
          dx.b := st[s];
          dx.c := st[s + 1];
          st[s] := ord(dx.a < dy.a);
          p := p + 1
        end;

      eqreal2:
        begin
          dy.b := st[s - 1];
          dy.c := st[s];
          s := s - 3;
          dx.b := st[s];
          dx.c := st[s + 1];
          st[s] := ord(dx.a = dy.a);
          p := p + 1
        end;

      grreal2:
        begin
          dy.b := st[s - 1];
          dy.c := st[s];
          s := s - 3;
          dx.b := st[s];
          dx.c := st[s + 1];
          st[s] := ord(dx.a > dy.a);
          p := p + 1
        end;

      ngreal2:
        begin
          dy.b := st[s - 1];
          dy.c := st[s];
          s := s - 3;
          dx.b := st[s];
          dx.c := st[s + 1];
          st[s] := ord(dx.a <= dy.a);
          p := p + 1
        end;

      nereal2:
        begin
          dy.b := st[s - 1];
          dy.c := st[s];
          s := s - 3;
          dx.b := st[s];
          dx.c := st[s + 1];
          st[s] := ord(dx.a <> dy.a);
          p := p + 1
        end;

      nlreal2:
        begin
          dy.b := st[s - 1];
          dy.c := st[s];
          s := s - 3;
          dx.b := st[s];
          dx.c := st[s + 1];
          st[s] := ord(dx.a >= dy.a);
          p := p + 1
        end;

      lsstring2:
        begin
          y := s - maxstring + 1;
          s := y - maxstring;
          i := 0;
          while (i < maxstring - 1)
            and 
              (st[s + i] = st[y + i])
                 do i := i + 1;
          st[s] := ord(st[s + i] <
            st[y + i]);
          p := p + 1
        end;

      eqstring2:
        begin
          y := s - maxstring + 1;
          s := y - maxstring;
          i := 0;
          while (i < maxstring - 1)
            and
              (st[s + i] = st[y + i])
                do i := i + 1;
          st[s] := ord(st[s + i] =
            st[y + i]);
          p := p + 1
        end;

      grstring2:
        begin
          y := s - maxstring + 1;
          s := y - maxstring;
          i := 0;
          while (i < maxstring - 1)
            and
              (st[s + i] = st[y + i])
                do i := i + 1;
          st[s] := ord(st[s + i] >
            st[y + i]);
          p := p + 1
       end;

      ngstring2:
        begin
          y := s - maxstring + 1;
          s := y - maxstring;
          i := 0;
          while (i < maxstring - 1)
            and
              (st[s + i] = st[y + i])
                do i := i + 1;
          st[s] := ord(st[s + i] <=
            st[y + i]);
          p := p + 1
        end;

      nestring2:
        begin
          y := s - maxstring + 1;
          s := y - maxstring;
          i := 0;
          while (i < maxstring - 1)
            and
              (st[s + i] = st[y + i])
                do i := i + 1;
          st[s] := ord(st[s + i] <>
            st[y + i]);
          p := p + 1
        end;

      nlstring2:
        begin
          y := s - maxstring + 1;
          s := y - maxstring;
          i := 0;
          while (i < maxstring - 1)
            and
              (st[s + i] = st[y + i])
                do i := i + 1;
          st[s] := ord(st[s + i] >=
            st[y + i]);
          p := p + 1
        end;

      equal2{length}:
        begin
          length := st[p + 1];
          y := s - length + 1;
          s := y - length;
          i := 0;
          while (i < length - 1) and
            (st[s + i] = st[y + i])
              do i := i + 1;
          st[s] := ord(st[s + i] =
            st[y + i]);
          p := p + 2
        end;

      notequal2{length}:
        begin
          length := st[p + 1];
          y := s - length + 1;
          s := y - length;
          i := 0;
          while (i < length - 1) and
           (st[s + i] = st[y + i])
             do i := i + 1;
          st[s] := ord(st[s + i] <>
            st[y + i]);
           p := p + 2
      end;

      assign2{length}:
        begin
          length := st[p + 1];
          s := s - length - 1;
          x := st[s + 1];
          y := s + 2;
          for i := 0 to
            length - 1 do
              st[x + i] :=
                st[y + i];
          p := p + 2
        end;

      read2{lineno}:
        begin
          read(input, cx);
          st[st[s]] := ord(cx);
          s := s - 1;
          p := p + 2
        end;

      readint2{lineno}:
        begin
          read(input,
            st[st[s]]);
          s := s - 1;
          p := p + 2
        end;

      readreal2{lineno}:
        begin
          read(input, dx.a);
          y := st[s];
          s := s - 1;
          st[y] := dx.b;
          st[y + 1] := dx.c;
          p := p + 2
       end;

      readln2{lineno}:
        begin
          readln(input);
          p := p + 2
        end;

      write2{option, lineno}:
        begin
          if st[p + 1] = ord(true)
            then
              begin
                write(output,
                  chr(st[s - 1]):
                    st[s]);
               s := s - 2
              end
            else
              begin
                write(output,
                  chr(st[s]));
                s := s - 1
              end;
          p := p + 3
        end;

      writebool2{option, lineno}:
        begin
          if st[p + 1] = ord(true)
            then
              begin
                write(output,
                  (st[s - 1] = 1)
                    :st[s]);
                s := s - 2
              end
            else
              begin
                write(output,
                  st[s] = 1);
                s := s - 1
              end;
          p := p + 3
        end;

      writeint2{option, lineno}:
        begin
          if st[p + 1] = ord(true)
            then
              begin
                write(output,
                  st[s - 1]:
                    st[s]);
                s := s - 2
              end
            else
              begin
                write(output,
                  st[s]);
                s := s - 1
              end;
          p := p + 3
        end;

      writereal2{option1, option2,
          lineno}:
        begin
          if st[p + 1] = ord(true)
            then
              if st[p + 2] =
                ord(true) then
                  begin
                    s := s - 4;
                    dx.b :=
                      st[s + 1];
                    dx.c :=
                      st[s + 2];
                    m :=
                      st[s + 3];
                    n :=
                      st[s + 4];
                    write(output,
                      dx.a:m:n)
                  end
                else
                  begin
                    s := s - 3;
                    dx.b :=
                      st[s + 1];
                    dx.c :=
                      st[s + 2];
                    m :=
                      st[s + 3];
                    write(output,
                      dx.a:m)
                  end
            else
              begin
                s := s - 2;
                dx.b := st[s + 1];
                dx.c := st[s + 2];
                write(output,
                  dx.a);
              end;
          p := p + 4
        end;

      writestring2{option, lineno}:
        begin
          if st[p + 1] = ord(true)
            then
              begin
              { write(output,
                  sx:width) }
                width := st[s];
                s := s - 1;
                popstring(sx);
                writestring(output,
                  sx, width)
              end
            else
              begin
              { write(output, sx) }
                popstring(sx);
                writestring(output,
                  sx,
                  stringlength(sx))
              end;
          p := p + 3
        end;

      writeln2{lineno}:
        begin
          writeln(output);
          p := p + 2
        end;

      open2{lineno}:
        begin
          cmax := cmax + 1;
          if cmax > maxchan then
            error1(st[p + 1], maxchan5)
          else
            begin
              open[cmax] := 0;
              st[st[s]] := cmax;
              s := s - 1;
              p := p + 2
            end
        end;

      checkio2{lineno}:
        begin
          c := st[s];          
          if (c < 1) or (c > cmax)
            then
              error1(st[p + 1], channel4)
          else p := p + 2
        end;

      endio2:
        begin
          s := s - 1;
          p := p + 1
        end;

      receive2{typeno, length,
          dynamic, lineno}:
        begin
          typeno := st[p + 1];
          length := st[p + 2];
          dynamic :=
            st[p + 3] = ord(true);
          lineno := st[p + 4];
          c := st[s - 1];
          si := open[c];
          if si = 0 then
            begin
              s := s + 3;
              st[s - 2] := -typeno;
              st[s - 1] := p + 5;
              st[s] := b;
              open[c] := s;
              select(lineno)
            end
          else
            begin
              typeno2 := st[si - 2];
              if typeno = typeno2 then
                begin
                  pi := st[si - 1];
                  bi := st[si];
                  si :=
                    si - length - 3;
                  if dynamic then
                    begin
                      {dassign2:}
                      x := st[s];                {target variable}
                      z := si + 1;               {source}
                      dassign(x, z);
                    end
                  else begin
                    x := st[s] - 1;
                    for i := 1 to length do
                      st[x + i] :=
                        st[si + i];
                  end;
                  s := s - 1;
                  open[c] := 0;
                  activate(bi, si, pi);
                  p := p + 5
                end
              else if typeno2 < 0 then
                error1(lineno, contention4)
              else
                error1(lineno, type4)
            end
        end;

      send2{typeno, length,
          dynamic, lineno}:
        begin
          typeno := st[p + 1];
          length := st[p + 2];
          dynamic :=
            st[p + 3] = ord(true);
          lineno := st[p + 4];
          c := st[s - length];
          si := open[c];
          if si = 0 then
            begin
              s := s + 3;
              st[s - 2] := typeno;
              st[s - 1] := p + 5;
              st[s] := b;
              open[c] := s;
              select(lineno)
            end
          else
            begin
              typeno2 := st[si - 2];
              if typeno = -typeno2 then
                begin
                  pi := st[si - 1];
                  bi := st[si];
                  si := si - 4;
                  s := s - length;
                  if dynamic then
                    begin
                      {dassign2:}
                      x := st[si + 1];           {target variable}
                      z := s + 1;                {source}
                      dassign(x, z);
                    end
                  else begin
                    x := st[si + 1] - 1;
                    for i := 1 to length do
                      st[x + i] :=
                        st[s + i];
                  end;
                  open[c] := 0;
                  activate(bi, si, pi);
                  p := p + 5
                end
              else if typeno2 > 0 then
                error1(lineno, contention4)
              else
                error1(lineno, type4)
            end
        end;

      do2{displ}:
        begin
          if st[s] = ord(true) then
            p := p + 2
          else p := p + st[p + 1];
          s := s - 1
        end;

      goto2{displ}:
        p := p + st[p + 1];

      for2:
        begin
          st[st[s - 1]] := st[s];
          s := s - 1;
          p := p + 1
        end;

      to2{displ}:
        begin
          if st[st[s - 1]] <= st[s]
            then p := p + 2
            else
              begin
                s := s - 2;
                p := p + st[p + 1]
              end
        end;

      endto2{disp}:
        begin
          x := st[s - 1];
          st[x] := st[x] + 1;
          p := p + st[p + 1]
        end;

      downto2{displ}:
        begin
          if st[st[s - 1]] >= st[s]
            then p := p + 2
            else
              begin
                s := s - 2;
                p := p + st[p + 1]
              end
        end;

      enddown2{displ}:
        begin
          x := st[s - 1];
          st[x] := st[x] - 1;
          p := p + st[p + 1]
        end;

      case2{
        lineno, length, table}:
          begin
            x := st[s];
            s := s - 1;
            { binary search }  
            i := 1;
            j := st[p + 2];
            while i < j do
              begin
                k :=
                 (i + j) div 2;
                m := p + 2*k + 1;
                if st[m] < x
                  then i := k + 1
                  else j := k
              end;
            m := p + 2*i + 1;
            if st[m] = x then
              p := p + st[m + 1]
            else
              error1(st[p + 1], case4)
          end;

      parallel2:
        begin
           s := s + 1;
           st[s] := 0;
           p := p + 1
        end;

      endparallel2{lineno}:
        select(st[p + 1]);

      process2{blockno,
        templength, displ,
          lineno}:
          process;

      endprocess2{displ, lineno}:
          endprocess;

      forall2{blockno,
        templength, displ,
          lineno}:
        begin
          upper := st[s];
          lower := st[s - 1];
          if lower <= upper then
            begin
              s := s - 1;
              st[s] :=
                upper - lower + 1;
              blockno :=
                st[p + 1];
              templength :=
                st[p + 2];
              lineno := st[p + 4];
              for i := lower to
                upper do
              begin
                bi :=
                  free[blockno];
                if bi = 0 then
                  begin
                    bi := t + 1;
                    t := bi +
                      templength + (arl + 1){ = 6};
                  end
                else
                  free[blockno] :=
                    st[bi];
                if t > maxaddr then
                  memorylimit(lineno)
                else
                  begin
                    st[bi] := b;
                    st[bi + 1] := s;

                    vi := bi + arl { = 5};
                    st[vi] := i;
                    si := vi + 1;
                    st[bi + 4] := vi;
{WAS
                    st[bi + 4] := i;
                    si := bi + 5;
}
                    st[si] :=
                      blockno;
                    activate(bi, si,
                      p + 5)
                  end
              end;
              select(lineno)
            end
          else { lower > upper }
            begin
              s := s - 2;
              p := p + st[p + 3]
            end
        end;

      endall2{lineno}:
        begin
          blockno := st[s];
          x := b;
          b := st[x];
          v := st[b + 4];
          s := st[x + 1];
          st[x] := free[blockno];
          free[blockno] := x;
          st[s] := st[s] - 1;
          if st[s] > 0 then
            select(st[p + 2])
          else
            begin
              s := s - 1;
              p := p + 2
            end
        end;

      assume2{lineno}:
        begin
          x := st[s];
          s := s - 1;
          if x = ord(false) then
            error1(st[p + 1], assume4)
          else p := p + 2
        end;

      method2{blockno,          st[p + 1]
                 paramlength,   st[p + 2]
                 varlength,     st[p + 3]
                 templength,    st[p + 4]
                 displ,         st[p + 5]
                 lineno         st[p + 6]}:
        begin
          st[s + 1] :=
            s - st[p + 2] - 1;
          st[s + 3] := b;
          b := s;
          blockno := st[p + 1];
          v0 := free[blockno];
          if v0 = 0 then
            begin
              v0 := t + 1;
              s := v0 + st[p + 3];
              t := s + st[p + 4];
              if t > maxaddr then
                memorylimit(
                  st[p + 6])
            end
          else
            begin
              free[blockno] := st[v0];
              s := v0 + st[p + 3];
            end;
          st[v0] := blockno;
          v := v0 + 1;
          st[b + 4] := v;
          p := p + st[p + 5]
        end;

{OLD:
        begin
          st[s + 1] :=
            s - st[p + 2] - 1;
          st[s + 3] := b;
          b := s;
          blockno := st[p + 1];
          s := free[blockno];
          if s = 0 then
            begin
              s := t + 1;
              t := s + st[p + 4];
              if t > maxaddr then
                memorylimit(
                  st[p + 6])
            end
          else
            free[blockno] := st[s];
          st[s] := blockno;
          p := p + st[p + 5]
        end;
}

      endmethod2:
        begin
{WAS:
          blockno := st[s];
          x := s;
}
          s := st[b + 1];
          p := st[b + 2];
          b := st[b + 3];

          v0 := v - 1;
          blockno := st[v0];
{          if blockno > maxblock then
             readln;
}
          st[v0] := free[blockno];
          free[blockno] := v0;
{WAS
          st[x] := free[blockno];
          free[blockno] := x
}
          v := st[b + 4];
        end;

      moduleenter2 {varlength}:
        begin
          v := b + arl { = 5 };
          s := v - 1 + st[p + 1];
          p := p + 2;
        end;

      moduleinit2 {varlength, templength, displ, lineno}:
        begin
          st[b]     := b; { own base, needed for localxcall2 }
          st[b + 1] := p; { start of own code, needed for imcall2}
          st[b + 2] := thismno; { owm module number; needed for error }
          v := b + arl { = 5};
          s := v - 1 + st[p + 1];
          t := s + st[p + 2];
          st[b + 4] := v;
          if t > maxaddr then
            memorylimit(st[p + 4])
          else p := p + st[p + 3]
        end;

      endmodule2:
          running := false;

      endinit2:
          running := false;

    { localvar(displ) =
        variable(0, displ) . }

      localvar2{displ}:
        begin
          s := s + 1;
          st[s] := v {WAS b} + st[p + 1];
          p := p + 2
        end;

    { localvalue(displ) =
        localvar(displ)
          value(1) . }

      localvalue2{displ}:
        begin
          s := s + 1;
          st[s] :=
            st[v {WAS b} + st[p + 1]];
          p := p + 2
        end;

    { localreal(displ) =
        localvar(displ)
          value(2) . }

      localreal2{displ}:
        begin
          x := v {WAS b} + st[p + 1];
          s := s + 2;
          st[s - 1] := st[x];
          st[s] := st[x + 1];
          p := p + 2
        end;

    { globalvar(displ) =
        variable(1, displ) . }

      globalvar2{displ}:
        begin
          s := s + 1;
          st[s] :=
            st[v {WAS b}] + st[p + 1];
          p := p + 2
        end;

    { globalvalue(displ) =
        globalvar(displ)
          value(1) . }

      globalvalue2{displ}:
        begin
          s := s + 1;
          st[s] := st[
            st[v {WAS b}] + st[p + 1]];
          p := p + 2
        end;

    { ordvalue =
        value(1) . }

      ordvalue2:
        begin
          st[s] := st[st[s]];
          p := p + 1
        end;

    { realvalue =
        value(2) . }

      realvalue2:
        begin
          x := st[s];
          s := s + 1;
          st[s - 1] := st[x];
          st[s] := st[x + 1];
          p := p + 1
        end;

    { ordassign =
        assign(1) . }

      ordassign2:
        begin
          st[st[s - 1]] := st[s];
          s := s - 2;
          p := p + 1
        end;

    { realassign =
        assign(2) . }

      realassign2:
        begin
          s := s - 3;
          x := st[s + 1];
          st[x] := st[s + 2];
          st[x + 1] := st[s + 3];
          p := p + 1
        end;

      globalcall2{displ}:
        begin
          s := s + 1;
          st[s] := st[b];
          st[s + 2] := p + 2;
          p := p + st[p + 1]
        end;

      installm2 {procdispl}:
        begin
          x := st[s];
          st[x] := b;
          st[x + 1] := p + st[p + 1];
          s := s - 1;
          p := p + 2
        end;

      iarray2 {arraylength}:
        begin
          x := st[s - 1];               {array descriptor}
          st[x]     := ord(arraytag);   {tag}
          st[x + 1] := newblock1;       {blockno}
          st[x + 2] := 0;               {blocklength}
          st[x + 3] := st[p + 1];       {array length}
          st[x + 4] := st[s];           {element descr reference}
          s := s - 2;
          p := p + 2;
        end;

      iclass2 {recordlength}:
        begin
          x := st[s];                   {class descriptor}
          st[x]     := ord(classtag);   {tag}
          st[x + 1] := newblock1;        {blockno}
          st[x + 2] := 0;               {blocklength}
          st[x + 3] := st[p + 1];       {record length}
          st[x + 4] := 0;               {number of dynamic fields}
{DEBUG:
          writeln('iclass; descr, length: ');
          writeln(x:5, st[x+3]:5); readln;
}
          s := s - 1;
          p := p + 2;
        end;

      ixclass2 {recordlength}:
        begin
          x := st[s];                   {xclass descriptor}
          st[x]     := ord(xclasstag);  {tag}
          st[x + 1] := newblock1;        {blockno}
          st[x + 2] := 0;               {blocklength}
          st[x + 3] := st[p + 1];       {record length}
          st[x + 4] := 0;               {number of dynamic fields}
{DEBUG
          writeln('ixclass; descr, length: ');
          writeln(x:5, st[x+3]:5); readln;
}
          s := s - 1;
          p := p + 2;
        end;

      oxclass2 {recordlength}:
        begin
          x := st[s];                   {xclass descriptor}
          st[x + 1] := newblock1;       {blockno}
          st[x + 2] := 0;               {blocklength}
          st[x + 3] := st[p + 1];       {record length}
{DEBUG
          writeln('oxclass; descr, length: ');
          writeln(x:5, st[x+3]:5); readln;
}
          s := s - 1;
          p := p + 2;
        end;

      ifield2 {displacement}:
        begin
          x := st[s - 1];               {class descriptor}
          y := st[x + 4];               {number of dynamic fields}
{DEBUG
          writeln('ifield; descr, length, field no: ');
          writeln(x:5, st[x+3]:5, y:5); readln;
}
          st[x + 4] := y + 1;
          x := x + 5 + 2*y;             {field descriptor}
          st[x]     := st[p + 1];       {field displacement}
          st[x + 1] := st[s];           {field descr reference}
{DEBUG
          writeln('ifield; field descr, field displ, field descr: ');
          writeln(x:5, st[x]:5, st[s]:5); readln;
}
          s := s - 2;
          p := p + 2;
        end;

      len2 {}:
        begin
          x := st[s];  {type decsriptor}
          st[x + 2] := {block length}
            blength(x);
{DEBUG
          writeln('len2; descr, BLOCK length: ');
          writeln(x:5, st[x+2]:5); readln;
}
          s := s - 1;
          p := p + 1;
        end;

      alloc2: {lineno}
        begin
          y := st[s];      {address of variable}
          z := st[s - 1];  {type decsriptor}
          t0 := t;
          allocate(
            z,          {descriptor}
            x,          {addr}
            new,
            st[p + 1]   {lineno});
          if new then
            construct(
              y,      {address of variable}
              z       {type decsriptor}
            )
          else begin
            st[x] := z {descr};
            x := x + 1;
            st[y {addr}] := x - y{addr};
          end;
          s := s - 2;
          p := p + 2;
        end;

      dealloc2:
        begin
          x := st[s] - 1;        {beginning of block}
          y := st[x];            {descriptor}
          blockno := st[y + 1];
          st[x] := free[blockno];
          free[blockno] := x;
          s := s - 1;
          p := p + 1;
        end;

      dresult2: {lineno}
        begin
          z := st[s];     {type decsriptor}
          allocate(
            z,            {descriptor}
            x,            {addr}
            new,
            st[p + 1]     {lineno});
          st[x] := z      {descr};
          x := x + 1;
          st[s] := x - s; {relative reference}
          p := p + 2;
        end;

      random2{lineno}:
        begin
          if st[s] < 0 then
            rangeerror(st[p + 1])
          else
            st[s] := random(st[s]);
          p := p + 2
        end;

      deref2{}:
        {deref is also performed by dassign2, send2, receive2}
        begin
          st[s] := st[s] + st[st[s]];
          p := p + 1
        end;

      dassign2 {}:
        begin
          x  := st[s - 1];            {target variable}
          x  := x + st[x];            {deref target}
          y  := s + st[s];            {deref source (allocated by dvalue2)}
          z := st[y - 1];             {source description}
          blockno := st[z + 1];
          length  := st[z + 2];       {block length}
{DEBUG}
          if z <> st[x - 1] then
            error1(0, implementation5);
{DEBUG
          writeln('dassign2, source, target, descr: ',
            y: 8, x: 8, z:8);
          readln;
}
          {copy block starting at y
          onto block starting at x}
          for i := 0 to
            length - 2 do {descriptor excluded}
              st[x + i] :=
                st[y + i];
          {deallocate source}
          st[y - 1] := free[blockno];
          free[blockno] := y - 1;
          s := s - 2;
          p := p + 1;
        end; {dassign2}

      dvalue2 {lineno}:
        begin
          {allocate whole block}
          y  := st[s] - 1;        {beginning of block}
          z := st[y];             {descriptor}
          allocate(
            z,          {descriptor}
            x,          {addr}
            new,
            st[p + 1]   {lineno});
          st[s] := x + 1 - s;
{DEBUG
          writeln('dvalue, valueStart, blockno, block len: ');
          writeln(x+1:5, z:5, blockno:5, length:5); readln;
}
                        
          {copy block starting at y
          onto block starting at x}
          for i := 0 to
            length - 1 do
              st[x + i] :=
                st[y + i];
          p := p + 2;
        end; {dvalue2}
      else
        error1(lineno, implementation5);
    end;
  end;
end { run };

procedure readtime(
 var t: real);
 var hour, minute, second, sec100: word;
 { A nonstandard type needed
   by a nonstandard procedure which
   reads the processor time in seconds }
begin
  gettime(hour, minute, second, sec100);
  t := hour*3600 + minute*60 + second
end;

procedure writetime(
  var output: text;
  t1, t2: real);
begin
  { Outputs the time interval
    t2 - t1 sec in seconds }
  writeln(output);
  writeln(output, 'Elapsed time: ', round(t2 - t1), ' seconds.')
end;

procedure initinterp;
  var i: integer;
begin
  correct := true;
  ready := 0; cmax := 0; t := 0;
  blocknumber := 0;
  for i := 1 to maxblock
    do free[i] := 0;
  randomize;

  {DEBUG: remove after debugging}
  for i := 1 to maxaddr
    do st[i] := -1;

end;

procedure load;

type
{
Taken out of load:
  importrecord =
    record
      id: identifier;
      key: integer;
      isbase: boolean;
      baseaddr: address;
    end;
  imptRecTable =
    array [0..maximps] of importrecord;
}
  basetable = array[0..maximps] of address;
  baserecord =
    record
      no: integer;
      bases: basetable;
    end;
  imports = array[0..maximps] of integer;
  queuerec = record
    first, last: integer;
    mno: imports;
  end;

var
  op, arg1, arg2,
  arg3, arg4, arg5, arg6: integer;
  realarg: real;
  stringarg: stringg;
  casearg: casetable;
  idarg, objid: identifier;
{
Taken out of load:
  glbimports: imptRecTable;
  nofglbimps: integer;
}
  queue: queuerec;

{ Local procedures for load }

procedure initqueue;
begin
  queue.first := 0;
  queue.last := -1;
end;

procedure enqueue(mno: integer);
begin
  queue.last := queue.last + 1;
  queue.mno[queue.last] := mno;
end;

procedure loadthis(
    var impname: identifier;
    var key: integer;
    var baseaddr, initaddr, entryaddr: address;
    var iprovided: baserecord;
    lineno: integer);

  var
    codename: stringg;
    codefile: binary;
    iseen: baserecord;
    blockno: integer;
    ib,              {imported bases}
    bno:             {base number}
      integer;

{ nested procedures follow }

  procedure resetcode
    (var codef: binary);
  begin
    {$I-} reset(codef); {$I+}
    if (ioresult <> 0) then
      error(file6);
  end;

  procedure error2(
    lineno: integer;
    kind: phrase);
  begin
    write(kind, sp,
        'line ', lineno:1, sp,
        'module ', impname);
    running := false; correct := false
  end;

  function newblock2: integer;
  begin
    if blocknumber = maxblock then
      error2(lineno, maxblock5);
    blocknumber := blocknumber + 1;
    newblock2 := blocknumber;
  end;

  procedure get(var value: integer);
  begin read(codefile, value) end;

  procedure put(value: integer);
  begin
    t := t + 1;
    st[t] := value
  end;

  procedure getreal(
    var value: real);
  var dual: dualreal;
  begin
    dual.split := true;
    get(dual.b);
    get(dual.c);
    dual.split := false;
    value := dual.a
  end;

  procedure putreal(value: real);
  var dual: dualreal;
  begin
    dual.split := false;
    dual.a := value;
    dual.split := true;
    put(dual.b);
    put(dual.c)
  end;

  procedure getstring(
    var length: integer;
    var value: stringg);
  var c, i: integer;
  begin
    get(length);
    for i := 1 to length do
      begin
        get(c);
        value[i] := chr(c)
      end;
    for i := length + 1
      to maxstring do
        value[i] := chr(null)
  end;

  procedure putstring(
    length: integer;
    value: stringg);
  var i: integer;
  begin
    put(length);
    for i := 1 to length do
      put(ord(value[i]))
  end;

  procedure getcase(
    var lineno, length: integer;
    var table: casetable);
  var i: integer;
  begin
    get(lineno); get(length);
    for i := 1 to length do
      begin
        get(table[i].value);
        get(table[i].index)
      end
  end;

  procedure putcase(
    lineno, length: integer;
    table: casetable);
  var i: integer;
  begin
    put(lineno); put(length);
    for i := 1 to length do
      begin
        put(table[i].value);
        put(table[i].index)
      end
  end;

  procedure getobjdesc(
    var id: identifier;
    var key, lineno: integer);
  var i, j, k, c: integer;

    procedure gettident(var name: identifier);
      var i, k, c: integer;
    begin i := 1;
      repeat get(c); name[i] := chr(c); i := i + 1
      until c = null;
      for k := i to maxident do name[k] := chr(null);
    end;

  begin
    gettident(id); get(key); get(lineno)
  end;

  procedure putobjdesc(
    id: identifier;
    key, lineno: integer);
  var i: integer;

    procedure putident(var id: identifier);
      var i, c: integer;
    begin i := 1;
      repeat c := ord(id[i]); put(c); i := i + 1
      until c = null
    end;

  begin
    put(lineno); putident(id); put(key)
  end;

  procedure nextinstruction;
  begin
    get(op);
    if op in noarguments then
      { skip }
    else if op in oneargument then
      get(arg1)
    else if op in twoarguments then
      begin
        get(arg1); get(arg2)
      end
    else if op in threearguments then
      begin
        get(arg1); get(arg2);
        get(arg3)
      end
    else if op in fourarguments then
      begin
        get(arg1); get(arg2);
        get(arg3); get(arg4)
      end
    else if op in sixarguments then
      begin
        get(arg1); get(arg2);
        get(arg3); get(arg4);
        get(arg5); get(arg6);
      end
    else if op = realconst2 then
      getreal(realarg)
    else if op = stringconst2 then
      getstring(arg1, stringarg)
    else if op = case2 then
      getcase(arg1, arg2,
        casearg)
    else if op in descriptions then
      getobjdesc(idarg, arg1, arg2)
    else error2(
      arg2, implementation5)
  end;

  procedure emit1(op: integer);
  begin
    debugarguments(op, 0);
    put(op);
  end;

  procedure emit2(
    op, arg: integer);
  begin
    debugarguments(op, 1);
    put(op); put(arg);
  end;

  procedure emit3(
    op, arg1, arg2: integer);
  begin
    debugarguments(op, 2);
    put(op); put(arg1);
    put(arg2);
  end;

  procedure emit4(
    op, arg1, arg2,
    arg3: integer);
  begin
    debugarguments(op, 3);
    put(op); put(arg1);
    put(arg2); put(arg3);
  end;

  procedure emit5(
    op, arg1, arg2, arg3,
    arg4: integer);
  begin
    debugarguments(op, 4);
    put(op); put(arg1);
    put(arg2); put(arg3);
    put(arg4);
  end;

  procedure emit7(
    op, arg1, arg2, arg3,
    arg4, arg5, arg6: integer);
  begin
    debugarguments(op, 6);
    put(op); put(arg1);
    put(arg2); put(arg3);
    put(arg4); put(arg5);
    put(arg6);
  end;

  procedure emitreal(
    value: real);
  begin
    put(realconst2);
    putreal(value);
  end;

  procedure emitstring(
    length: integer;
    value: stringg);
  begin
    put(stringconst2);
    putstring(length, value);
  end;

  procedure emitcase(
    lineno, length: integer;
    table: casetable);
  begin
    put(case2);
    putcase(lineno, length,
      table);
  end;

  procedure forall(blockno {0},
    templength, enddispl, lineno: integer);
  begin
    blockno := newblock2;
    emit5(forall2,
      blockno,
      templength,
      enddispl,
      lineno);
    nextinstruction
  end;

  procedure procedur(blockno {0},
    paramlength, varlength, templength,
    begindispl, lineno: integer);
  begin
    blockno := newblock2;
    emit7(method2,
      blockno,
      paramlength,
      varlength,
      templength,
      begindispl,
      lineno);
    nextinstruction
  end;

  procedure process(blockno {0},
    templength, enddispl, lineno: integer);
  begin
    blockno := newblock2;
    emit5(process2,
      blockno,
      templength,
      enddispl,
      lineno);
    nextinstruction
  end;

  procedure moduleinit(
    varlength, templength,
    begindispl, lineno: integer
    );
  begin
    initaddr := t + 1;
    emit5(moduleinit2,
      varlength,
      templength,
      begindispl,
      lineno);
    nextinstruction
  end;

  procedure modulehead(
    hdid: identifier;
    hdkey, lineno: integer);
  begin
    if (hdid <> impname) then
      error2(lineno, name6)
    else
      key := hdkey;
    nextinstruction
  end;

  procedure endinit;
  begin
    emit1(endinit2);
    entryaddr := t + 1;
    nextinstruction;
  end;

  procedure copyinstruction;
  begin
    if op in noarguments
      then emit1(op)
    else if op in oneargument
      then emit2(op, arg1)
    else if op in twoarguments
      then emit3(op, arg1, arg2)
    else if op in threearguments
      then
        emit4(op, arg1, arg2,
          arg3)
    else if op in fourarguments
      then
        emit5(op, arg1, arg2,
          arg3, arg4)
    else if op in sixarguments
      then
        emit7(op, arg1, arg2,
          arg3, arg4, arg5, arg6)
    else if op = realconst2
      then emitreal(realarg)
    else if op = stringconst2 then
      emitstring(arg1,
        stringarg)
    else if op = case2 then
      emitcase(arg1, arg2,
        casearg)
    else
      error2(lineno, implementation5);
    nextinstruction
  end;

  procedure insert(
    b: integer;
    var t: baserecord
  );
  var
    i: integer;
  begin
    t.bases[0] := b;
    i := t.no;
    while t.bases[i] <> b do
      i := i - 1;
    if i = 0 then begin
      t.no := t.no + 1;
      t.bases[t.no] := b;
    end;
  end;

  procedure loadx(
    id: identifier;
    key, lineno: integer;
    isbase: boolean);

  var
    m, mno, is, ip: integer;
    over, loadd, runn: boolean;
    lprovided: baserecord;

  begin
    over := false;
    loadd := true;
    runn := true;
    iseen.no := iseen.no + 1;
    is := iseen.no;
    if isbase then begin
      iprovided.no := iprovided.no + 1;
      ip := iprovided.no;
    end;
    mno := nofglbimps;
    glbimports[0].id := id;
    repeat
      if mno <= 0 then over := true
      else if id <> glbimports[mno].id then mno := mno - 1
      else { same id }
        begin
          m := mno; loadd := false;
          if isbase then over := true
          else if glbimports[mno].isbase then mno := mno -1
          else begin runn := false; over := true end
        end
    until over;
    if loadd or runn then begin { import }
      if nofglbimps < maximps then
        nofglbimps := nofglbimps + 1
      else
        error2(0, maximport5);
      mno := nofglbimps;
      glbimports[mno].id := id;
      glbimports[mno].isbase := isbase;

      if loadd then begin
        loadthis(id,
          glbimports[mno].key,
          glbimports[mno].baseaddr,
          impinit[mno],
          impentry[mno],
          lprovided,
          lineno)
      end else begin
        glbimports[mno].key := glbimports[m].key;
        glbimports[mno].baseaddr := t + 1;
        impinit[mno] := impinit[m];
        impentry[mno] := impentry[m];
      end;
      if runn then begin
        run(mno,
          impinit[mno],
          glbimports[mno].baseaddr);
        enqueue(mno);
      end;
    end;

    {add loaded object to those seen/provided by importer}
    iseen.bases[is] := glbimports[mno].baseaddr;
    if isbase then
      iprovided.bases[ip] := glbimports[mno].baseaddr;

    {add bases provided by loaded object
    to those SEEN by importer}
    for m := 1 to lprovided.no do
      insert(lprovided.bases[m], iseen);

    {if loaded object is a base (i.e., private),
    add bases provided by loaded object
    to those PROVIDED by importer}
    if isbase then
      for m := 1 to lprovided.no do
        insert(lprovided.bases[m], iprovided);

    if key <> glbimports[mno].key then
      error2(lineno, version6);

    nextinstruction
  end;

begin { loadthis }

  makefilename(impname, codeext, codename);
  { nonstandard assign }
  assign(codefile, codename);
  resetcode(codefile);
  iseen.no := 0; iprovided.no := 0;
  nextinstruction;
  while correct and (op <> endmodule2) do
    if op = forall2 then
      forall(arg1 {=blockno=0},
        arg2, arg3, arg4)
    else if op = method2 then
      procedur(arg1{=blockno=0},
        arg2, arg3, arg4, arg5, arg6)
    else if op = process2 then
      process(arg1{=blockno=0},
        arg2, arg3, arg4)
    else if op = moduleinit2 then
      moduleinit(arg1, arg2, arg3, arg4)
    else if op = loadshared2 then
      loadx(idarg, arg1, arg2, {isbase=}false)
    else if op = loadprivate2 then
      loadx(idarg, arg1, arg2, {isbase=}true)
    else if op = modulehead2 then
      modulehead(idarg, arg1, arg2)
    else if op = endinit2 then
      endinit
    else copyinstruction;
  emit1(endmodule2);
  close(codefile);
  baseaddr := t + 1;
  {install bases}
  ib := baseaddr + arl{ = 5}; {import bases}
{WAS:  ib := baseaddr + 4; /import bases}
  st[ib] := baseaddr; {self}
{DEBUG
  writeln('module ', impname:1,
    '  baseaddr ', baseaddr:1,
    '  iseen.no:', iseen.no:1);
}
  bno := iseen.no;
  while bno > 0 do begin
    st[ib + bno] := iseen.bases[bno];
{DEBUG
    writeln('base ', iseen.bases[bno], ' into ', ib + bno);
}
    bno := bno - 1;
  end;
end; { loadthis }

procedure loadmain;
var
  unused1, i, mno: integer;
  unused2: baserecord;
begin
  initqueue;
  loadthis(objid,
    unused1,
    glbimports[0].baseaddr,
    impinit[0],
    impentry[0],
    unused2,
    0);
    glbimports[0].id := objid;
{DEBUG
        writeln('Initializing MAIN module; <id, base> = ',
          glbimports[0].id,
          glbimports[0].baseaddr);
}
  run(0, impinit[0], glbimports[0].baseaddr);
  enqueue( {main} 0 );
  for i := queue.first to queue.last do begin
    mno := queue.mno[i];
    run(mno,
      impentry[mno],
      glbimports[mno].baseaddr
    );
  end;
end;

begin { load }
  nofglbimps := 0;
  write('    module = ');
  readident(objid);
  writeln(output);
  loadmain;
end; { load }

begin { interpreter }
  initcommon;
  initinterp;

  readtime(t1);
  load;
  readtime(t2);
  writetime(output, t1, t2)
end.