unit pParse;

interface

procedure parse;
procedure initparser;

implementation
  uses pCommon, pTexim, pScan;

procedure parse;

type
  tokens = set of mintoken..maxtoken;
{
  typenode = ^ typerecord; OBJSET
  typerecord =
    record
      prev: typenode;
      obj: objectt;
    end;
}
var
  addtokens, blocktokens,
  constanttokens, declarationtokens,
  doubletokens, expressiontokens,
  factortokens, literaltokens,
  longtokens, multiplytokens,
  parametertokens, relationtokens,  
  methodtokens, selectortokens,
  signtokens, simpleexprtokens,
  statementtokens, termtokens,
  unsignedtokens: tokens;
  functions, leftparts, localsubroutines,
  variables, parameters, 
  procedures, userfunctions,
  xmethods, methods: kinds;
  ordtypes, structures: forms;
  modulevarlen,
    {length of vars & descriptors at main level}
  labelno, minint, unused
           : integer;
  externals: objset;
{
  typelist
    {list of structures that need to be installed/
           : typenode;
}

{ FORWARD DECLARATIONS }

procedure push(length: integer);
forward;

procedure pop(length: integer);
forward;

procedure expression(
  var typex: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
forward;

procedure statementsequence(
  var used: closure;
  restricted: boolean;
  stop: tokens);
forward;

procedure statement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
forward;

procedure declarationpart(
  var varlength: integer;
  stop: tokens);
forward;

procedure statementpart(
  var templength: integer;
  var used: closure;
  stop: tokens);
forward;

{ INPUT }

procedure nexttoken;
begin
  gettoken;
  while token = newline1 do
    begin
      newline(lineno);
      gettoken
    end;
end;

{ OUTPUT }

procedure emit1(op: integer);
begin
  debugarguments(op, 0);
  put(op)
end;

procedure emit2(
  op, arg: integer); 
begin
  debugarguments(op, 1);
  put(op); put(arg)
end;

procedure emit3(
  op, arg1, arg2: integer);
begin
  debugarguments(op, 2);
  put(op); put(arg1);
  put(arg2)
end;

procedure emit4(
  op, arg1, arg2,
  arg3: integer);
begin
  debugarguments(op, 3);
  put(op); put(arg1);
  put(arg2); put(arg3)
end;

procedure emit5(
  op, arg1, arg2, arg3,
  arg4: integer);
begin
  debugarguments(op, 4);
  put(op); put(arg1);
  put(arg2); put(arg3);
  put(arg4)
end;

procedure emit7(
  op, arg1, arg2, arg3,
  arg4, arg5, arg6: integer);
begin
  debugarguments(op, 6);
  put(op); put(arg1);
  put(arg2); put(arg3);
  put(arg4); put(arg5);
  put(arg6);
end;

procedure emitreal(
  value: real);
begin
  put(realconst2);
  putreal(value)
end;

procedure emitstring(
  value: stringg);
begin
  put(stringconst2);
  putstring(
    stringlength(value),
    value);
end;

procedure emitcase(
  length: integer;
  table: casetable;
  lineno: integer);
begin
  put(case2);
  putcase(lineno, length,
    table);
end;

procedure emitmoduledesc(
  op: integer;
  id: identifier;
  key, lineno: integer);
begin
  put(op);
  putmoduledesc(id, key, lineno)
end;

{ object sets }

function empty(x: objset)
  : boolean;
{ empty = (x = []) }
begin
  empty := x = nil
end;

function member(x: objset;
  y: objectt): boolean;
{ member = y in x }
var correct: boolean;
begin
  correct := false;
  while not correct
    and (x <> nil) do
      if x^.item = y then
        correct := true
      else x := x^.next;
  member := correct
end;

function disjoint(
  x, y: objset): boolean;
{ disjoint = empty(x*y) }
var correct: boolean;
begin
  correct := true;
  while correct and
    (x <> nil) do
      if member(y, x^.item)
        then correct := false
        else x := x^.next;
  disjoint := correct
end;

procedure include(
  var x: objset;
  y: objectt);
{ x := x + [y] }
var v: objset;
begin
  if not member(x, y) then
    begin
      new(v);
      v^.next := x;
      v^.item := y;
      x := v;
    end
end;

procedure addset(
  var x: objset;
  y: objset);
{ x := x + y }
begin
  while y <> nil do
    begin
      include(x, y^.item);
      y := y^.next
    end
end;

function unrestrictedm(
  obj: objectt)
    : boolean;
{ unrestrictedm = the object
  is an unrestricted exported method }
begin
  if obj^.kind in xmethods then
    unrestrictedm :=
      not obj^.typex^.restricted
  else
    unrestrictedm := false
end;

procedure copyset(
  var x: objset;
  y: objset);
{ x := y }
begin
  newset(x);
  addset(x, y)
end;

{ global objects }

function nonlocalobj(
  proc, obj: objectt)
    : boolean;
{ nonlocalobj = the object
  is a variable or procedure that is
  nonlocal to a procedure
  (or function) }
var kind: objectkind;
begin
  kind := obj^.kind;
  if kind in variables then
    nonlocalobj :=
      obj^.varlevel <=
        proc^.methodlevel
  else if kind in methods then
{WAS:
    nonlocalobj :=
      (obj^.methodlevel <=
        proc^.methodlevel)
      and (proc <> obj)
}
    nonlocalobj :=
      unrestrictedm(obj)
      and (proc <> obj)
  else if kind = standardfile
    then nonlocalobj := true
  else if kind in xmethods
    then nonlocalobj := true
  else nonlocalobj := false
end;

procedure nonlocalset(
  proc: objectt;
  var left: objset;
  right: objset);
{ left := all members of
  right which are nonlocal
  to a procedure (or
  function) }
var item: objectt;
begin
  newset(left);
  while right <> nil do
    begin
      item := right^.item;
      if nonlocalobj(proc, item)
        then
          include(left, item);
      right := right ^.next
    end
end;

function globalvar(
  obj: objectt)
    : boolean;
{ globalvar = the object
  is a variable that is
  global }
var kind: objectkind;
begin
  kind := obj^.kind;
  if kind in variables then
    globalvar :=
      obj^.varlevel =
        mainlevel
  else if kind = standardfile
    then globalvar := true
  else
    globalvar := false
end;

function globalset(
  x: objset): boolean;
{ globalset := true
  if a member of x is global }
var item: objectt;
  more: boolean;
begin
  more := x <> nil;
  while more do
    begin
      item := x^.item;
      if globalvar(item) then
        more := false
      else begin
        x := x^.next;
        more := x <> nil;
      end;
    end;
  globalset := x <> nil;
end;

function unrestrictedset(
  x: objset): boolean;
{ unrestrictedset := a member of x is
  an exported unrestricted method }
var item: objectt;
  more: boolean;
begin
  more := x <> nil;
  while more do
    begin
      item := x^.item;
      if unrestrictedm(item) then
        more := false
      else begin
        x := x^.next;
        more := x <> nil;
      end;
    end;
  unrestrictedset := x <> nil;
end;

{ statement closures }

function emptyclosure(
  used: closure): boolean;
{ emptyclosure =
    (used = []) }
begin
  emptyclosure :=
    empty(used.exprobj) and
    empty(used.targetobj)
end;

procedure addclosure(
  var left: closure;
  right: closure);
{ left := left + right }
begin
  addset(left.targetobj,
    right.targetobj);
  addset(left.exprobj,
    right.exprobj);
end;

procedure copyclosure(
  var left: closure;
  right: closure);
{ left := right }
begin
  newclosure(left);
  addclosure(left, right)
end;

procedure nonlocalclosure(
  proc: objectt;
  var left: closure;
  right: closure);
{ left := all members of
  right which are nonlocal
  to a procedure or
  function }
begin
  nonlocalset(proc,
    left.targetobj,
    right.targetobj);
  nonlocalset(proc,
    left.exprobj,
    right.exprobj)
end;

function globalclosure(
  x: closure): boolean;
{ globalclosure := true
  if x contains a global var }
begin
  globalclosure :=
    globalset(x.targetobj)
      or
    globalset(x.exprobj)
end;

function unrestricted(
  x: closure): boolean;
{ unrestricted := x contains
  a unrestricted exported procedure }
begin
  unrestricted :=
    unrestrictedset(x.targetobj)
      or
    unrestrictedset(x.exprobj)
end;

{ type analysis }

procedure checktypes(
  var type1: structure;
  type2: structure);
begin
  if type1 <> type2 then
    begin
      if (type1 <> typevoid)
        and
          (type2 <> typevoid)
            then error(type3);
      type1 := typevoid
    end
end;

procedure typeerror(
  var typex: structure);
begin
  if typex <> typevoid then
    begin
      error(type3);
      typex := typevoid
    end
end;

procedure checktype(
  var typex: structure;
  f: forms);
begin
  if not (typex^.form in f) then
    typeerror(typex)
end;

procedure checkmessage(
  var type1, type2: structure;
  var typeno: integer);
{ typeno > 0 }
var mess: messagenode;
  found: boolean;
begin
  typeno := 1;
  if type1^.form = channeltype then
    begin
      mess := type1^.messagelist;
      found := false;
      while not found and
          (mess <> nil) do
        if mess^.typex <> type2 then
          begin
            typeno := typeno + 1;
            mess := mess^.prev
          end
        else found := true;
      if not found then
        typeerror(type2)
    end
  else typeerror(type1)
end;

procedure kinderror(
  obj: objectt);
begin
  if obj^.kind <> undefined
    then error(kind3)
end;

procedure convert1(
  var type1: structure;
  type2: structure);
begin
  if (type1 = typeinteger) and
    (type2 = typereal) then
      begin
        emit1(float2);
        pop(1); push(2);
        type1 := typereal
      end
end;

procedure convert2(
  var type1, type2: structure);
begin
  if (type1 = typeinteger) and
    (type2 = typereal) then
      begin
        emit1(floatleft2);
        pop(1); push(2);
        type1 := typereal
      end
   else convert1(type2, type1)
end;

{ labels }

procedure newlabel(
  var no: integer);
begin
  if labelno = maxlabel then
    haltt(maxlabel5);
  labelno := labelno + 1;
  no := labelno
end;

{ variable addressing }

function typelength(
  typex: structure) : integer;
begin
  if dynamic(typex) then {reference}
    typelength := 1
  else if typex^.form in ordtypes then
    typelength := 1
  else if typex^.form = realtype then
    typelength := 2
  else if typex^.form = arraytype then
    typelength := typex^.arraylength
  else if typex^.form = channeltype then
    typelength := 1
  else if typex^.form = classtype then
    typelength := typex^.recordlength
  else
    haltt(implementation5)
end;

procedure fieldaddressing(
  recordlength: integer;
  lastfield: objectt);
var displ: integer;
begin
  displ := recordlength;
  while lastfield <> nil do
    begin
      { skip undefined
        obj (if any) }
      if lastfield^.kind = field
        then
          begin
            displ := displ -
              typelength(lastfield^.typex);
            lastfield^.fielddispl
              := displ
          end;
      lastfield :=
        lastfield^.prev
    end
end;

procedure variableaddressing(
  varlength: integer;
  lastvar: objectt);
var displ: integer;
begin
  displ := varlength;
  while lastvar <> nil do
    begin
      { skip undefined
        obj (if any) }
      if lastvar^.kind = variable
        then
          begin
            displ := displ -
              typelength(
                lastvar^.typex);
            lastvar^.varlevel :=
              blocklevel;
            lastvar^.vardispl :=
              displ
          end;
      lastvar := lastvar^.prev
    end
end;

function valuereference(
  obj: objectt): boolean;
begin
  if obj^.kind = valueparameter then
    valuereference :=
      obj^.typex^.form in structures
  else
    valuereference := false;
end;

procedure parameteraddressing(
  paramlength: integer;
  lastparam: objectt);
var displ: integer;
begin
  displ := 0;
  while displ > - paramlength do
    begin
      { skip undefined
        objects (if any) }
      if (lastparam^.kind = varparameter)
         or valuereference(lastparam) then
        begin
          displ := displ - 1;
          lastparam^.varlevel :=
            blocklevel;
          lastparam^.vardispl :=
            displ
        end
      else if lastparam^.kind =
          valueparameter then
        begin
          displ := displ -
            typelength(
              lastparam^.typex);
          lastparam^.varlevel :=
            blocklevel;
          lastparam^.vardispl :=
            displ
        end;
      lastparam :=
        lastparam^.prev
    end
end;

{ temporaries }

procedure push{length: integer};
begin
  with block[blocklevel] do
    begin
      templength :=
        templength + length;
      if maxtemp < templength then
        maxtemp := templength
    end
end;

procedure pop{length: integer};
begin
  with block[blocklevel] do
    templength :=
      templength - length
end;

{ case etables }

procedure insertcase(
  value, index: integer;
  var length: integer;
  var table: casetable;
  var found: boolean);
var x, y: caserecord;
  i: integer;
begin
  { 0 <= length < maxcase
    and table ordered by
    case values }
  x.value := value;
  x.index := index;
  found := false;
  for i := 1 to length do
    begin
      y := table[i];
      if x.value < y.value
        then
          begin
            table[i] := x;
            x := y
          end
      else
        if x.value = y.value
          then found := true
     end;
  length := length + 1;
  table[length] := x
end;

{ overriding }

procedure register(
  var id: identifier;
  marked: boolean;
  kind: objectkind;
  var obj: objectt;
  var other: objectt);
var
  impno: integer;
  isnew: boolean;

begin
  if marked {(blocklevel = mainlevel} then begin
    searchobj(id, blocklevel, other, impno);
    if (other <> nil) and (impno > 0) then
      if other^.kind in [xproc, xfunc] then
        isnew := false {other^.typex^.form <> methodtype}
      else if (other^.kind = typ) then
        isnew := other^.typex^.form <> xclasstype
      else isnew := true
    else isnew := true;
  end else
    isnew := true;
  if isnew then
    begin
      declareobj(id, kind, obj);
      other := nil;
    end
  else
    enterobj(id, kind, obj);
  obj^.marked := marked
end;

procedure sameProcType(
  procobj, other: objectt);
var
  same, more: boolean;
  t1, t2: structure;
  p1, p2: objectt;
begin
  t1 := procobj^.typex; t2 := other^.typex;
  same := (t1^.paramlength = t2^.paramlength)
    and (t1^.resulttype = t2^.resulttype)
    and (t1^.restricted = t2^.restricted);
  p1 := t1^.lastparam; p2 := t2^.lastparam;
  more := (p1 <> nil) and (p2 <> nil);
  while same and more do begin
    same := (p1^.id = p2^.id) and (p1^.typex = p2^.typex);
    p1 := p1^.prev; p2 := p2^.prev;
    more := (p1 <> nil) and (p2 <> nil);
  end;
  if same then 
    procobj^.typex := other^.typex
  else
    error(syntax3)
end;

{ Descriptor =
    ArrayDescriptor     |
    ClassDescriptor     |
    XclassDescriptor    |
    MethodDescriptor .
  MethodDescriptor =
    ModuleBase
    Displacement .
  ArrayDescriptor =
    ArrayTypeTag
    BlockNumber
    BlockLength
    ArrayLength
    ElementDescriptorReference .
  ClassDescriptor =
    [ ClassTypeTag | XclassTypeTag ]
    BlockNumber
    BlockLength
    RecordLength
    NumberOfDynamicFields
    [ DynamicFieldDescriptor ] * .
  DynnamicFieldDescriptor =
    FieldDispl
    FieldTypeDescriptorReference . }

procedure descriptor(var str: structure);
var
  dlen: integer;
  form: structform;
  current: objset;
{  current: typenode;
}

{
  procedure append(
    str: structure
  );
  var
    current: typenode;
    obj: objectt;
  begin
    obj := str^.obj;
    if obj^.kind = typ then
      if obj^.typelevel > mainlevel then
        begin
          new(current);
          current^.obj := obj;
          current^.prev := typelist;
          typelist := current;
        end;
  end;
}
begin {descriptor}
  form := str^.form;
  str^.descr := modulevarlen;
  dlen := 0;
  if form = methodtype then
    dlen := 2
  else if form = arraytype then
    dlen := 5
  else if form in [xclasstype, classtype] then
    dlen := 5 + 2 * maxfields
    {descr must be allocated for new class only}
  else
     haltt(implementation5);
  modulevarlen := modulevarlen + dlen;
{
  include(externals, str^.obj);
}
{
  append(str);
}
end; {decriptor}

procedure installOne(
  obj: objectt
);
var
  str, comp: structure;

  procedure fields(str: structure);
  var fld: objectt;
    fldtype: structure;
  begin
    if str^.form = xclasstype then
      fld := str^.baseclass^.lastfield
    else
      fld := str^.lastfield;
    while fld <> nil do begin
      fldtype := fld^.typex;
      if dynamic(fldtype) then begin
        {class descriptor}
        emit3(xvariable2,
          str^.cno,
          str^.{type}descr);
        emit3(xvariable2,
          fldtype^.cno,
          fldtype^.{type}descr);
        push(2);
        emit2(ifield2,
          fld^.fielddispl);
        pop(2);
      end;
      fld := fld^.prev;
    end;
  end; {fields}

begin {installOne}
  str := obj^.typex;
  if str^.form = methodtype then begin
    if obj^.marked then begin
      emit3(xvariable2,
        str^.cno,
        str^.{procedure}descr);
      push(1);
      emit2(installm2, obj^.methodlabel);
      pop(1);
    end
  end else if (obj^.kind = typ) then begin
    if dynamic(str) then begin
      if str^.form = arraytype then
        begin
          emit3(xvariable2,
            str^.cno,
            str^.{type}descr);
          push(1);
          {push element descriptor reference}
          comp := str^.elementtype;
          emit3(xvariable2,
            comp^.cno,
            comp^.{type}descr);
          push(1);
          emit2(iarray2,
            str^.arraylength);
          pop(2);
        end
      else if str^.form = classtype then
        begin
          emit3(xvariable2,
            str^.cno,
            str^.{type}descr)  ;
          push(1);
          emit2(iclass2,
            str^.recordlength);
          pop(1);
          fields(str);
        end
      else if str^.form = xclasstype then
        begin
          emit3(xvariable2,
            str^.cno,
            str^.{type}descr);
          push(1);
          if str^.baseclass^.baseclass = typevoid then
            {install}
            emit2(ixclass2,
              str^.recordlength)
          else {override}
            emit2(oxclass2,
              str^.recordlength);
          pop(1);
          fields(str);
        end
      else
        haltt(implementation5);
    end; {dynamic}
  end; {typ}
end; {installOne}

procedure install(
  list: objset);
begin
{
  if blocklevel <> mainlevel then
    haltt(implementation5);
}
  if not empty(list) then begin
    install(list^.next);
    installOne(list^.item);
  end;
end; {install}

procedure findlenOne(
  obj: objectt
);
var
  str: structure;
begin
  str := obj^.typex;
  if dynamic(str) then begin
    emit3(xvariable2,
      str^.cno,
      str^.{type}descr);
    push(1);
    emit1(len2);
    pop(1);
  end;
end; {findlenOne}

procedure findlen(
  list: objset);
begin
  if not empty(list) then begin
    findlen(list^.next);
    findlenOne(list^.item);
  end;
end; {findlen}

{replaced 3-13-98
procedure installglobal;
var
  obj: objectt;
begin
  if blocklevel <> mainlevel then
    haltt(implementation5);
  obj := block[blocklevel].lastobject;
  while obj <> nil do begin
    install(obj);
    obj := obj^.prev;
  end;
end;

procedure installlocal;
var
  obj: objectt;
  list: typenode;
begin
  if blocklevel <> mainlevel then
    haltt(implementation5);
  list := typelist;
  while list <> nil do begin
    obj := list^.obj;
    install(obj);
    list := list^.prev;
  end;
end;

procedure findlenglobal;
var
  obj: objectt;
begin
  obj := block[blocklevel].lastobject;
  while obj <> nil do begin
    if obj^.kind = typ then
      findlen(obj);
    obj := obj^.prev;
  end
end;

procedure findlenlocal;
var
  obj: objectt;
  list: typenode;
begin
  list := typelist;
  while list <> nil do begin
    obj := list^.obj;
    if obj^.kind = typ then
      findlen(obj);
    list := list^.prev;
  end;
end;
}

procedure allocate;
var
  obj: objectt;
  str: structure;
  level: integer;
begin
  obj := block[blocklevel].lastobject;
  while obj <> nil do begin
    str := obj^.typex;
    if obj^.kind = variable then
      begin
        if dynamic(str) then
          begin
            emit3(xvariable2,
              str^.cno,
              str^.{type}descr);
            push(1);
            level := blocklevel - obj^.varlevel;
            emit3(variable2,
              level,
              obj^.vardispl);
            push(1);
            emit2(alloc2,
              lineno);
            pop(2);
          end;
      end;
    obj := obj^.prev;
  end;
end; {allocate}

procedure deallocate;
var
  obj: objectt;
  str: structure;
  level: integer;
begin
  obj := block[blocklevel].lastobject;
  while obj <> nil do begin
    str := obj^.typex;
    if obj^.kind = variable then
{keep this as an option: r/w structured value parameters:
    if obj^.kind in [variable, valueparameter] then
}
      begin {value par allocated by dvalue2}
        if dynamic(str) then
          begin
            level := blocklevel - obj^.varlevel;
            emit3(variable2,
              level,
              obj^.vardispl);
            push(1);
            emit1(deref2);
            emit1(dealloc2);
            pop(1);
          end;
      end;
    obj := obj^.prev;
  end;
end;

{ PARSER INITIALIZATION }

procedure initialize;
begin
  modulevarlen := 0;
{CHECK if all these are used}
  addtokens :=
    [minus1, or1, plus1];
  declarationtokens :=
    [const1, function1,
     procedure1, type1, var1];
  blocktokens := [begin1]
    + declarationtokens;
  doubletokens := [charconst1,
    intconst1, identifier1];
  literaltokens :=
    [charconst1, intconst1,
     realconst1, stringconst1];
  longtokens := [realconst1,
    stringconst1] + doubletokens;
  multiplytokens :=
    [and1, asterisk1, div1,
     mod1, slash1];
  parametertokens :=
    [identifier1, var1];
  relationtokens := [equal1,
    greater1, less1, notequal1,
    notgreater1, notless1];
  methodtokens :=
    [function1, procedure1];
  selectortokens :=
    [leftbracket1, period1];
  signtokens := [minus1, plus1];
  statementtokens :=
    [assume1, begin1, case1,
     for1, forall1, if1,
     identifier1, leftbracket1,
     parallel1, repeat1, while1,
     inherit1];
  unsignedtokens := [identifier1]
    + literaltokens;
  constanttokens := signtokens
    + unsignedtokens;
  factortokens := [leftparenthesis1,
    not1, inherit1] + unsignedtokens;
  termtokens := factortokens;
  simpleexprtokens := signtokens
    + termtokens;
  expressiontokens :=
    simpleexprtokens;
  structures := [arraytype,
    classtype, xclasstype];
  userfunctions := [funktion, xfunc];
  localsubroutines := [procedur, funktion];
  functions := userfunctions + [standardfunc];
  parameters := [valueparameter,
    varparameter];
  procedures :=
    [procedur, xproc, standardproc];
  xmethods :=
    [xproc, xfunc];
  methods := [procedur, xproc, funktion, xfunc];
  ordtypes := [inttype, chartype, booltype, enumtype];
  variables :=
    [variable] + parameters;
  leftparts :=
    userfunctions + variables;
  labelno := 0;
  minint := - maxint - 1;
  newset(externals);
{
  typelist := nil;
}
end;

{ SYNTAX ANALYSIS }

procedure syntaxerror(
  stop: tokens);
begin
  error(syntax3);
  while not (token in stop) do
    nexttoken
end;

procedure syntaxcheck(
  stop: tokens);
begin
  if not (token in stop) then
    syntaxerror(stop)
end;

procedure expect(t: integer;
  stop: tokens);
begin
  if token = t then
    begin
      nexttoken;
      syntaxcheck(stop)
    end
  else syntaxerror(stop)
end;

procedure expectid(
  var id: identifier;
  stop: tokens);
begin
  if token = identifier1 then
    begin
      id := idarg;
      nexttoken
    end
  else
    begin
      id := nameless;
      syntaxerror(stop)
    end;
  syntaxcheck(stop)
end;

procedure expectmark(
  var mk: boolean;
  markpermitted: boolean);
begin
  if token = export1 then begin
    if markpermitted then mk := true
    else begin
      mk := false; error(syntax3)
    end;
    gettoken
  end else mk := false
end;


procedure expectmethodmark(
  var mk, restr: boolean;
  markpermitted: boolean);
begin
  mk := false; restr := false;
  if token = export1 then begin
    if markpermitted then begin
      mk := true; restr := true;
    end else
      error(syntax3);
    gettoken
  end else if token = unrestricted1 then begin
    if markpermitted then
      mk := true
    else
      error(syntax3);
    gettoken;
  end;
end;

{OLD
procedure expectmethodmark(
  var mk, restr: boolean;
  markpermitted: boolean);
begin
  mk := false; restr := false;
  if token = export1 then begin
    if markpermitted then
      mk := true
    else
      error(syntax3);
    gettoken
  end else if token = separate1 then begin
    if markpermitted then begin
      restr := true; mk := true;
    end else
      error(syntax3);
    gettoken;
  end;
end;
}

{ QualIdent =
    [ ModuleIdentifier "." ] Identifier . }

procedure qualident(
    var obj: objectt;
    var cno: integer;
    stop: tokens);
begin { token = identifier1 }
  cno := 0;
  findobj(idarg, obj, cno);
  expect(identifier1, [period1] + stop);
  if obj^.kind = module then begin
    expect(period1, [identifier1] + stop);
    if token = identifier1 then
      findimport(obj, {copy of xbase}
        idarg, obj, cno);
    expect(identifier1, stop);
  end{ else
    cno := 0;}
end; { token = identifer1 }

{  TypeIdentifier = QualIdent . }

procedure typeidentifier(
  var typex: structure;
  stop: tokens);                    
var obj: objectt; 
begin
  if token = identifier1 then
    begin
      qualident(obj, unused, stop);
      if obj^.kind = typ then typex := obj^.typex
      else
        begin
          kinderror(obj);
          typex := typevoid
        end
    end
  else begin
    typex := typevoid;
    syntaxerror(stop)
  end
end;

{ UnsignedOrdinal =
    CharacterConstant |
      UnsignedInteger . }

procedure unsignedordinalliteral(
  var value: integer;
  var typex: structure;
  stop: tokens);
var obj: objectt; 
begin
  if token = charconst1 then
    begin
      value := argument;
      typex := typechar;
      expect(token, stop)
    end
  else if token = intconst1 then
    begin
      value := argument;
      typex := typeinteger;
      expect(token, stop)
    end
  else
    begin
      syntaxerror(stop);
      value := 0;
      typex := typevoid
    end
end;

{ UnsignedConstant =
    UnsignedOrdinal |
    UnsignedReal |
    CharacterString |
    ConstantQualIdent . }

procedure unsignedconstant(
  var value: integer;
  var realvalue: real;
  var stringvalue: stringg;
  var typex: structure;
  stop: tokens);
var obj: objectt;

begin
  { returns the type of an
    unsigned constant and
    its ordinal, real or
    stringg value }
  if token = realconst1 then
    begin
      realvalue := realarg;
      typex := typereal;
      expect(token, stop)
    end
  else if token = stringconst1
    then
      begin
        stringvalue := textarg;
        typex := typestring;
        expect(token, stop)
      end
  else if token = identifier1 then
    begin
      qualident(obj, unused, stop);
      if obj^.kind = constant then begin
        if obj^.typex = typereal then begin
          loadreal(obj, realvalue);
          typex := typereal;
        end else if obj^.typex = typestring then begin
          loadstring(obj, stringvalue);
          typex := typestring;
        end else if obj^.typex^.form in ordtypes then begin
          value := obj^.value;
          typex := obj^.typex
        end else begin
          kinderror(obj);
          value := 0;
          typex := typevoid
        end
      end
    end
  else
    unsignedordinalliteral(value, typex,
      stop)
end;

{ Constant =
    [ Sign ] UnsignedConstant .
  Sign =
    "+" | "-" . }

procedure doconstant(
  var value: integer;
  var realvalue: real;
  var stringvalue: stringg;
  var typex: structure;
  stop: tokens);
var sign: integer;
begin
  syntaxcheck(constanttokens
    + stop);
  if token in signtokens then
    begin
      sign := token;
      expect(sign, unsignedtokens
        + stop);
      unsignedconstant(value,
        realvalue, stringvalue,
        typex, stop);
      if typex = typeinteger then
        if sign = minus1 then
          if value <> maxint then
            value := - value
          else error(number3)
        else { sign = plus1 }
      else if typex = typereal then
        if sign = minus1 then
          realvalue := - realvalue
        else { skip }
      else
        begin
          typeerror(typex);
          value := 0
        end
    end
  else
    unsignedconstant(value,
      realvalue, stringvalue,
      typex, stop)
end;

{ OrdinalConstant =
    Constant . }

procedure ordinalconstant(
  var value: integer;
  var typex: structure;
  stop: tokens);
var realvalue: real;
  stringvalue: stringg;
begin
  doconstant(value, realvalue,
    stringvalue, typex, stop);
  if not (typex^.form in ordtypes) then
    begin
      typeerror(typex);
      value := 0
    end
end;

{ ExportMark = "*" .
  ConstantDeclaration =
    ConstantIdentifier
      [ ExportMark ]
        "=" Constant . }

procedure constantdefinition(
  stop: tokens);
var
  id: identifier;
  value: integer;
  realvalue: real;
  stringvalue: stringg;
  stringptr: stringpointer;
  constx: objectt;
  typex: structure;
  marked: boolean;
  stop2: tokens;
begin
  stop2 := constanttokens + stop;
  expectid(id, [export1, equal1] + stop2);
  expectmark(marked, blocklevel = mainlevel);
  expect(equal1, stop2);
  doconstant(value, realvalue,
    stringvalue, typex, stop);
  if typex^.form in ordtypes then
    begin
      declareobj(id, constant, constx);
      constx^.value := value;
      constx^.typex := typex;
      if marked and (typex^.form = enumtype) then
        begin
          haltt(implementation5);
          marked := false
        end;
      constx^.marked := marked
    end
  else if typex = typereal then
    begin
      declareobj(id, constant, constx);
      storereal(constx, realvalue);
      constx^.typex := typereal;
      constx^.marked := marked
    end
  else if typex = typestring then
    begin
      declareobj(id, constant, constx);
      constx^.typex := typestring;
      new(stringptr);
      stringptr^ := stringvalue;
      storestring(constx, stringptr);
      constx^.marked := marked
    end
end;

{ ConstantDeclarations =
    "const" ConstantDeclaration ";"
      [ ConstantDeclaration ";" ]*
        . }

procedure constantdefinitions(
  stop: tokens);
var stop2, stop3: tokens;
begin
  stop2 := [identifier1] + stop;
  stop3 := [semicolon1] + stop2;
  expect(const1, stop3);
  constantdefinition(stop3);
  expect(semicolon1, stop2);
  while token = identifier1 do
    begin
      constantdefinition(stop3);
      expect(semicolon1, stop2)
    end
end;

{ NewConstantList =
    ConstantIdentifier [ ","
      NewConstantList ] . }

procedure newconstantlist(
  value: integer;
  typex: structure;
  marked: boolean;
  stop: tokens);
var constx: objectt;
  id: identifier;
begin
  expectid(id, [export1, comma1]
    + stop);
  declareobj(id, constant, constx);
{ enum constants are assumed marked as their type ident}
  constx^.marked := marked;
  constx^.value := value;
  constx^.typex := typex;
  if token = comma1 then
    begin
      expect(comma1,
        [identifier1] + stop);
      newconstantlist(
        value + 1, typex, marked, stop)
    end
  else
    begin
      typex^.minvalue := 0;
      typex^.maxvalue := value;
    end
end;

{ EnumeratedType =
    "(" NewConstantList ")" . }

procedure enumeratedtypex(var id: identifier;
  marked: boolean; stop: tokens);
var obj: objectt;
  typex: structure;
  stop2: tokens;
begin
  stop2 :=
    [rightparenthesis1] + stop;
  declareobj(id, typ, obj);
  obj^.marked := marked;
  newstruct(enumtype, obj, typex);
  obj^.typex := typex;
  expect(leftparenthesis1,
    [identifier1] + stop2);
  newconstantlist(0, typex, marked,
    stop2);
  expect(rightparenthesis1,
    stop)
end;

{ IndexRange =
    "[" OrdinalConstant ".."
      OrdinalConstant "]" .  }

procedure indexrange(
  var lowerbound, upperbound:
    integer;
  var indextype: structure;
  stop: tokens);
var stop2, stop3: tokens;
  uppertype: structure;
begin
  stop2 := [rightbracket1] + stop;
  stop3 := constanttokens + stop2;
  expect(leftbracket1,
    [doubledot1] + stop3);
  ordinalconstant(
    lowerbound, indextype,
    [doubledot1] + stop3);
  expect(doubledot1, stop3);
  ordinalconstant(upperbound,
    uppertype, stop2);
  checktypes(indextype, uppertype);
  if lowerbound > upperbound then
    begin
      error(range3);
      lowerbound := upperbound
    end;
  expect(rightbracket1, stop)
end;

{ ArrayType =
    "array" IndexRange "of"
      TypeQualIdent . }

procedure arraytypex(
  var id: identifier;
  marked: boolean;
  stop: tokens);
var
  obj: objectt;
  newtype, indextype,
  elementtype: structure;
  lowerbound, upperbound, dyncomps: integer;
  stop2, stop3: tokens;
begin
  stop2 := [identifier1] + stop;
  stop3 := [of1] + stop2;
  expect(array1, [leftbracket1]
    + stop3);
  indexrange(lowerbound, upperbound,
    indextype, stop3);
  expect(of1, stop2);
  typeidentifier(elementtype, stop);
  declareobj(id, typ, obj);
  newstruct(arraytype, obj, newtype);
  obj^.typex := newtype;
  obj^.marked := marked;
  obj^.typelevel := blocklevel;
  newtype^.lowerbound := lowerbound;
  newtype^.upperbound := upperbound;
  newtype^.arraylength :=
    (upperbound - lowerbound + 1) *
      typelength(elementtype);
  newtype^.indextype := indextype;
  newtype^.elementtype := elementtype;
  if dynamic(elementtype) then begin
    newtype^.dyncomps := 1;
    descriptor(newtype);
    include(externals, obj);
  end;
end;

{ ClassSection =
    FieldIdentifier
      [ ExportMark ]
        SectionTail .
  SectionTail =
    "," ClassSection |
    ":" TypeQualIdent . }

procedure ClassSection(
  base: structure;
  var number: integer;
  var lastfield: objectt;
  var typex: structure;
  exported: boolean;
  stop: tokens);
var id: identifier; fieldx: objectt;
  stop2: tokens;
begin
  stop2 := [identifier1] + stop;
  expectid(id, [export1, comma1, colon1]
    + stop);
  if base <> typevoid then begin
    findfield(base, id, fieldx);
    if fieldx <> nil then
      error(ambiguous3)
  end;
  declareobj(id, field, fieldx);
  expectmark(fieldx^.marked, exported);
  if token = comma1 then
    begin
      expect(comma1, stop2);
      ClassSection(base, number,
        lastfield, typex, exported, stop);
      number := number + 1
    end
  else
    begin
      expect(colon1, stop2);
      typeidentifier(typex, stop);
      lastfield := fieldx;
      number := 1
    end;
  fieldx^.typex := typex
end;

{ FieldList =
    ClassSection
      [ ";" ClassSection ]*
        [ ";" ] . }

procedure fieldlist(
  base: structure;
  var lastfield: objectt;
  var length: integer;
  var dyncomps: integer;
  exported: boolean;
  stop: tokens);
var
  number: integer;
  typex: structure;
  stop2: tokens;
begin
  stop2 := [semicolon1] + stop;
  ClassSection(
    base, number, lastfield,
    typex, exported, stop2);
  length := length +
    number * typelength(typex);
  if dynamic(typex) then begin
    dyncomps := dyncomps + number;
    if dyncomps > maxfields then
      error(maxdynamic5);
  end;
  while token = semicolon1 do
    begin
      { use "stop" here }
      expect(semicolon1,
        [identifier1] + stop);
      if token = identifier1 then
        begin
          ClassSection(
            base, number, lastfield,
            typex, exported, stop2);
          length :=
            length + number *
              typelength(typex);
          if dynamic(typex) then begin
            dyncomps := dyncomps + number;
            if dyncomps > maxfields then
              error(maxdynamic5);
          end;
        end;
    end;
  fieldaddressing(
    length, lastfield)
end;

{ ClassKeyword =
    "class" | "record" .
  ClassType =
    ".." |
    ClassKeyword FieldList "end" . }

procedure classtyp(
  var id: identifier;
  marked: boolean;
  stop: tokens); forward;

procedure xclasstyp(
  var id: identifier;
  marked: boolean;
  stop: tokens); forward;

procedure classtypex(
  var id: identifier;
  marked: boolean;
  stop: tokens);

begin
  if marked then
    xclasstyp(id, marked, {desclength,} stop)
  else
    classtyp(id, marked, {desclength,} stop)
end;

  procedure classtyp(
    var id: identifier;
    marked: boolean;
    stop: tokens);
  var
    newstr: structure;
    obj, lastfield: objectt;
    length: integer;
    dyncomps: integer;
    stop2, stop3: tokens;
  begin {classtyp}
    stop2 := [end1] + stop;
    stop3 := [identifier1] + stop2;
    length := 0; dyncomps := 0;
    if token = nullclass1 then begin
      expect(doubledot1, stop + [semicolon1]);
      lastfield := nil;
    end else begin
      expect(class1, stop3);
      newblock(nil);
      fieldlist(typevoid, lastfield, length,
        dyncomps, marked, stop2);
      expect(end1, stop);
      endblock;
    end;
    declareobj(id, typ, obj);
    newstruct(classtype, obj, newstr);
    obj^.typex := newstr;
    newstr^.recordlength := length;
    newstr^.dyncomps := dyncomps;
    newstr^.lastfield := lastfield;
    obj^.marked := marked;
    obj^.typelevel := blocklevel;
    if dynamic(newstr) then begin
      descriptor(newstr);
      include(externals, obj);
    end;
  end; {classtyp}

  procedure xclasstyp(
    var id: identifier;
    marked: boolean;
    stop: tokens);
  var
    anchor, str: structure;
    newobj, oldobj, lastfield: objectt;
    length: integer;
    dyncomps: integer;
    stop2, stop3: tokens;
  begin {xclasstyp}
    stop2 := [end1] + stop;
    stop3 := [identifier1] + stop2;
    register(id, marked, typ,
      newobj, {isnew,} oldobj);
    newobj^.typelevel := blocklevel;
    include(externals, newobj);  
    if oldobj = nil {isnew} then begin
      newstruct(xclasstype, newobj, anchor);
      anchor^.baseclass := typevoid;
      length := 0;
      dyncomps := 0;
      descriptor(anchor);
    end else begin {class extension}
      anchor := oldobj^.typex;
      length :=
        anchor^.recordlength;
      dyncomps :=
        anchor^.dyncomps;
    end;
    newobj^.typex := anchor;
    newstruct(classtype, nil, str);
    if token = nullclass1 then begin
      expect(doubledot1, stop + [semicolon1]);
      lastfield := nil;
    end else begin
      expect(class1, stop3);
      newblock(nil);
      fieldlist(anchor^.baseclass,
        lastfield, length,
        dyncomps, marked, stop2);
      expect(end1, stop);
      endblock;
    end;
    str^.lastfield := lastfield;
    str^.baseclass := anchor^.baseclass;
    anchor^.baseclass := str;
    anchor^.recordlength := length;
    anchor^.dyncomps := dyncomps;
  end; {xclasstyp}

{ MessageTypeQualIdent =
    QualIdent . }

procedure messagetypequalident(
  var last: messagenode;
    stop: tokens);
var typex: structure;
begin
  typeidentifier(typex, stop);
  new(last);
  last^.prev := nil;
  last^.typex := typex
end;

{ MessageTypeList =
    MessageTypeQualIdent [ ","
      MessageTypeQualIdent ]* . }

procedure messagetypelist(
  var last: messagenode;
  stop: tokens);
var next: messagenode;
  stop2: tokens;
begin
  stop2 := [comma1] + stop;
  messagetypequalident(
    last, stop2);
  while token = comma1 do
    begin
      expect(comma1,
        [identifier1] + stop2);
      messagetypequalident(
        next, stop2);
      next^.prev := last;
      last := next;
    end
end;

{ ChannelType =
    "*" "(" MessageTypeList ")"
      . }

procedure channeltypex(
  var id: identifier;
  marked: boolean;
  stop: tokens);
var stop2, stop3: tokens;
  newtype: structure; obj: objectt;
  last: messagenode; {dynamic: boolean;}
begin
  stop2 := [rightparenthesis1]
    + stop;
  stop3 :=
    [identifier1] + stop2;
  expect(asterisk1,
    [leftparenthesis1] + stop3);
  expect(leftparenthesis1,
    stop3);
  messagetypelist(last, {dynamic,} stop2);
  expect(rightparenthesis1,
    stop);
  declareobj(id, typ, obj);
  newstruct(channeltype, obj, newtype);
  obj^.typex := newtype;
  newtype^.messagelist := last;
  newtype^.dyncomps := 0 ;
  obj^.marked := marked;
  obj^.typelevel := blocklevel;
end;

{ ExistingType =
     TypeIdentifier . }

procedure existingtypex(var id: identifier;
  marked: boolean; stop: tokens);
var
  obj: objectt;
  oldtype: structure;
begin
  typeidentifier(oldtype, stop + [semicolon1]);
  declareobj(id, typ, obj);
  obj^.typex := oldtype;
  obj^.marked := marked;
  obj^.typelevel := blocklevel;
end;

{ TypeDefinition =
    TypeIdentifier [ ExportMark ]
      "=" NewType | ExistingType .
  NewType =
    EnumeratedType |
    ArrayType |
    ClassType |
    ChannelType . }

procedure typedefinition(
  stop: tokens);
var stop2: tokens; id: identifier;
  obj: objectt;
  marked: boolean;
begin
  stop2 := [array1, asterisk1,
    leftparenthesis1, {record1,} class1, doubledot1]
      + stop;
  expectid(id, [export1, equal1] + stop2);
  expectmark(marked, blocklevel = mainlevel);
  expect(equal1, stop2);
  if token = leftparenthesis1
    then enumeratedtypex(id, marked, stop)
  else if token = array1
    then arraytypex(id, marked, {desclength,} stop)
  else if token in [class1, nullclass1] then
    classtypex(id, marked, {desclength,} stop)
  else if token = asterisk1
    then channeltypex(id, marked, {desclength,} stop)
  else if token = identifier1
    then existingtypex(id, marked, stop)
  else
    begin
      declareobj(id, undefined, obj);
      syntaxerror(stop)
    end
end;

{ TypeDeclaration =
    "type" TypeDefinition ";"
      [ TypeDefinition ";" ]* . }

procedure typedefinitions(
  stop: tokens);
var stop2, stop3: tokens;
begin
  stop2 := [identifier1] + stop;
  stop3 := [semicolon1] + stop2;
  expect(type1, stop3);
  typedefinition({desclength, }stop3);
  expect(semicolon1, stop2);
  while token = identifier1 do
    begin
      typedefinition({desclength, }stop3);
      expect(semicolon1, stop2)
    end
end;

{ VariableDeclaration =
    VariableIdentifier [ ExportMark ]
      VariableTail .
  VariableTail =
    "," VariableDeclaration |
    ":" TypeQualIdent . }

procedure variabledeclaration(
  kind: objectkind;
  var number: integer;
  var lastvar: objectt;
  var typex: structure;
  stop: tokens);
var id: identifier; varx: objectt;
  stop2: tokens;
begin
  stop2 := [identifier1] + stop;
  expectid(id, [export1, comma1, colon1]
    + stop);
  declareobj(id, kind, varx);
  expectmark(varx^.marked, blocklevel = mainlevel);
  if token = comma1 then
    begin
      expect(comma1, stop2);
      variabledeclaration(kind,
        number, lastvar, typex,
        stop);
      number := number + 1
    end
  else
    begin
      expect(colon1, stop2);
      typeidentifier(typex, stop);
      lastvar := varx;
      number := 1
    end;
  varx^.typex := typex
end;

{ VariableDeclarations =
    "var" VariableDeclaration ";"
      [ VariableDeclaration ";" ]*
        . }

procedure variabledeclarations(
  var length: integer; stop: tokens);
var lastvar: objectt;
  typex: structure;
  number: integer;
  stop2, stop3: tokens;
begin
  stop2 := [identifier1] + stop;
  stop3 := [semicolon1] + stop2;
  expect(var1, stop3);
  variabledeclaration(variable,
    number, lastvar, typex, stop3);
  length := {NEW} length +
    number * typelength(typex);
  expect(semicolon1, stop2);
  while token = identifier1 do
    begin
      variabledeclaration(variable,
        number, lastvar, typex,
        stop3);
      length := length +
        number * typelength(typex);
      expect(semicolon1, stop2)
    end;
  variableaddressing(length,
    lastvar)
end;

{ FormalParameterSection =
    [ "var" ]
      VariableDeclaration . }

procedure formalparametersection(
  proc: objectt;
  var lastparam: objectt;
  var length: integer;
  stop: tokens);
var stop2: tokens; number: integer;
  typex: structure;
begin
  stop2 := [identifier1] + stop;
  syntaxcheck([var1] + stop2);
  if token = var1 then
    begin
      expect(var1, stop2);
      variabledeclaration(
        varparameter, length,
        lastparam, typex, stop);
      if proc^.kind in userfunctions
        then error(parameter3)
    end
  else
    begin
      variabledeclaration(
        valueparameter, number,
        lastparam, typex, stop);
      if typex^.form in structures then
        { valuereference }
        length := number
      else length :=
        number * typelength(typex)
    end
end;

{ FormalParameters =
    FormalParameterSection [ ";"
      FormalParameterSection ]* . }

procedure formalparameters(
  proc: objectt;
  var lastparam: objectt;
  var length: integer;
  stop: tokens);
var more: integer; stop2: tokens;
begin
  stop2 := [semicolon1] + stop;
  formalparametersection(
    proc, lastparam,
    length, stop2); 
  while token = semicolon1 do
    begin
      expect(semicolon1,
        parametertokens + stop2);
      formalparametersection(
        proc, lastparam,
        more, stop2);
      length := length + more
    end;
  parameteraddressing(length,
    lastparam)
end;

{ FormalParameterList =
    "(" FormalParameters ")" . }

procedure formalparameterlist(
  proc: objectt;
  var lastparam: objectt;
  var length: integer;
  stop: tokens);
var stop2: tokens;
begin
  stop2 := [rightparenthesis1]
    + stop;
  expect(leftparenthesis1,
    parametertokens + stop2);
  formalparameters(proc,
    lastparam, length, stop2);
  expect(rightparenthesis1, stop)
end;

{  }

procedure formalparameterpart(
  proc: objectt;
  var lastparam: objectt;
  var length: integer;
  stop: tokens);
begin
  syntaxcheck([leftparenthesis1]
    + stop);
  if token = leftparenthesis1
    then
      formalparameterlist(
        proc, lastparam,
        length, stop)
    else { no parameter list }
      begin
        lastparam := nil;
        length := 0
      end
end;

{ UnrestrictedMark = "-" .
  RestrictedMark = "*" .
  MethodMark = UnrestrictedMark | RestrictedMark .
  FunctionHeading =
    "function" FunctionIdentifier [ MethodMark ]
      FormalParameterPart ":"
        TypeQualIdent . }


procedure functionheading(
  var procobj: objectt;
  stop: tokens);
var stop2, stop3: tokens;
  procstr: structure;
  other: objectt;
  id: identifier;
  impno: integer;
  kind: class;
  marked, restricted: boolean;
begin
  stop2 := [identifier1] + stop;
  stop3 := [colon1] + stop2;
  expect(function1,
    [leftparenthesis1] + stop3);
  expectid(id, [export1, unrestricted1, leftparenthesis1]
    + stop3);
  expectmethodmark(marked,
    restricted, blocklevel = mainlevel);
  if marked then kind := xfunc
  else kind := funktion;
  register(id,
    marked, kind, procobj, other);
  newstruct(methodtype, procobj, procstr);
  procobj^.methodlevel := blocklevel;
  procstr^.restricted := restricted;
  if procobj^.kind = xfunc then begin
    if other = nil then {newfunc}
      descriptor(procstr);
    include(externals, procobj);
  end;
{
  if (other = nil {newfunc/)
      and (procobj^.kind = xfunc) then
    descriptor(procstr);
}
  newlabel(procobj^.methodlabel);
  newblock(procobj);
  formalparameterpart(
    procobj, procstr^.lastparam,
    procstr^.paramlength, stop3);
  expect(colon1, stop2);
  typeidentifier(
    procstr^.resulttype, stop);
  procobj^.typex := procstr;
  if other <> nil {not newfunc} then
    sameProcType(procobj, other);
end;

{ MethodBlock =
    Block .
  Block =
    DeclarationPart
      StatementPart . }

procedure procedureblock(
  procobj: objectt;
  stop: tokens);
var
  implicit, used: closure;
  varlabel, templabel,
    beginlabel, templength: integer;
  procstr: structure;
begin
  procstr := procobj^.typex;
  newlabel(varlabel);
  newlabel(templabel);
  newlabel(beginlabel);
  emit2(defaddr2, procobj^.methodlabel);
  emit7(method2,
    {blockno} 0,
    procstr^.paramlength,
    varlabel, templabel,
    beginlabel, lineno);
  procobj^.recursive := false;
  newclosure(procobj^.implicit);
  procobj^.varlength := 0;
  declarationpart(
    procobj^.varlength,
    [begin1] + stop);
  emit3(defarg2, varlabel,
    procobj^.varlength);
  emit2(defaddr2, beginlabel);
  statementpart(templength,
    used, stop);
  nonlocalclosure(procobj,
    implicit, used);
  if (procobj^.kind in userfunctions)
      and procstr^.restricted
      and not
    empty(implicit.targetobj)
      then error(parameter3)
  else if procobj^.recursive
      and not
    emptyclosure(implicit)
      then error(recursion3)
  else if procstr^.restricted
      and globalclosure(implicit) then
    error(restricted3)
  else
    procobj^.implicit := implicit;
  emit3(defarg2, templabel,
    templength);
  emit1(endmethod2);
  endblock
end;

{ FunctionDeclaration =
    FunctionHeading ";"
      MethodBlock . }

procedure functiondeclaration(
  stop: tokens);
var proc: objectt;
  restricted: boolean;
  stop2: tokens;
begin
  stop2 := blocktokens + stop;
  functionheading(proc,
    [semicolon1] + stop2);
  expect(semicolon1, stop2);
  procedureblock(proc, stop)
end;

{ ProcedureHeading =
    "procedure"
      ProcedureIdentifier [ MethodMark ]
        FormalParameterPart . }

procedure procedureheading(
  var procobj: objectt;
  stop: tokens);
var
  procstr: structure;
  other: objectt;
  id: identifier;
  impno: integer;
  kind: class;
  marked, restricted: boolean;
  stop2: tokens;
begin
  stop2 :=
    [export1, unrestricted1, leftparenthesis1] + stop;
  expect(procedure1,
    [identifier1] + stop2);
  expectid(id, stop2);
  expectmethodmark(marked,
    restricted, blocklevel = mainlevel);
  if marked then kind := xproc
  else kind := procedur;
  register(id, marked, kind, procobj, other);
  newstruct(methodtype, procobj, procstr);
  procobj^.methodlevel := blocklevel;
  procstr^.restricted := restricted;
  if procobj^.kind = xproc then begin
    if other = nil then {newfunc}
      descriptor(procstr);
    include(externals, procobj);
  end;
{
  if (other = nil {newproc/)
      and (procobj^.kind = xproc) then
    descriptor(procstr);
}
  newlabel(procobj^.methodlabel);
  newblock(procobj);
  formalparameterpart(
    procobj, procstr^.lastparam,
    procstr^.paramlength, stop);
  procstr^.resulttype := typevoid;
  procobj^.typex := procstr;
  if other <> nil {not newproc} then
    sameProcType(procobj, other);
end;

{ ProcedureDeclaration =
    ProcedureHeading ";"
      MethodBlock . }

procedure proceduredeclaration(
  stop: tokens);
var proc: objectt;
  stop2: tokens;
begin
  stop2 := blocktokens + stop;
  procedureheading(proc,
    [semicolon1] + stop2);
  expect(semicolon1, stop2);
  procedureblock(proc, stop)
end;

{ MethodDeclarations =
    [ FunctionDeclaration ";" |
      ProcedureDeclaration ";"
        ]* . }

procedure methoddeclarations(
{  var desclength: integer;}
  stop: tokens);
var stop2, stop3: tokens;
begin
  stop2 := methodtokens + stop;
  stop3 := [semicolon1] + stop2;
  syntaxcheck(stop2);
  while token in methodtokens do
    begin
      if token = function1 then
        functiondeclaration({desclength, }stop3)
      else { symbol = procedure1 }
        proceduredeclaration({desclength, }stop3);
      expect(semicolon1, stop2)
    end
end;

{ DeclarationPart =
    [ ConstantDeclarations ]
      [ TypeDeclaration ]
        [ VariableDeclarations ]
          MethodDeclarations . }

procedure declarationpart{
  var varlength: integer;
  stop: tokens};
var stop2: tokens;
begin
  stop2 := methodtokens + stop;
  syntaxcheck(declarationtokens
    + stop);
  if token = const1 then
    constantdefinitions(
      [type1, var1] + stop2);
  if token = type1 then
    typedefinitions({varlength, }[var1] + stop2);
  if token = var1 then
    variabledeclarations(varlength,
      stop2);
  methoddeclarations({varlength, }stop)
end;

{ IndexExpression =
    OrdinalExpression . }

procedure indexexpression(
  var typex: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var exprtype: structure;
begin
  if dynamic(typex) then
    emit1(deref2);
  expression(exprtype, used,
    restricted, stop);
  if typex^.form = arraytype then
    begin
      checktypes(exprtype,
        typex^.indextype);
      emit5(index2,
        typex^.lowerbound,
        typex^.upperbound,
        typelength(
          typex^.elementtype),
        lineno);
      pop(1);
      typex := typex^.elementtype
    end
  else
    begin
      typeerror(typex);
      typex := typevoid
    end
end;

{ IndexExpressions =
    IndexExpression
      [ "," IndexExpression ]* . }

procedure indexexpressions(
  var typex: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var used2: closure; stop2: tokens;
begin
  stop2 := [comma1] + stop;
  indexexpression(typex, used,
    restricted, stop2);
  while token = comma1 do
    begin
      expect(comma1,
        expressiontokens + stop2);
      indexexpression(typex, used2,
        restricted, stop2);
      addclosure(used, used2)
    end
end;

{ IndexedSelector =
    "[" OrdinalExpressions "]" . }

procedure indexedselector(
  var typex: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var stop2: tokens;
begin
  stop2 := [rightbracket1] + stop;
{
  if dynamic(typex) then
    emit1(deref2);
}
  expect(leftbracket1,
    expressiontokens + stop2);
  indexexpressions(typex, used,
    restricted, stop2);
  expect(rightbracket1, stop)
end;

{ FieldSelector =
    "." FieldIdentifier . }

procedure fieldselector(
  var typex: structure;
  stop: tokens);
var {found: boolean;}
  fieldx: objectt;
begin
  expect(period1, [identifier1]
    + stop);
  if token = identifier1 then
    begin
      if typex^.form in [classtype, xclasstype]
          then
        begin

          findfield(typex, idarg, fieldx);

          if fieldx <> nil then
            begin
              if dynamic(typex) then
                emit1(deref2);
              typex := fieldx^.typex;
              emit2(field2, fieldx^.fielddispl)
            end
          else
            error(undefined3);
        end
      else
        begin
          typeerror(typex);
          typex := typevoid
        end;
      expect(identifier1, stop)
    end
  else
    begin
      syntaxerror(stop);
      typex := typevoid
    end
end;

{ EntireVariableAccess =
    VariableIdentifier . }

procedure entirevariableaccess(
  var typex: structure;
  var used: closure;
  istarget: boolean;
  stop: tokens);
var level, displ, cno: integer;
  obj: objectt;
begin
  newclosure(used);
  if token = identifier1 then
    begin
      findobj(idarg, obj, cno);
      expect(identifier1, stop);
      if istarget and valuereference(obj) then
        begin
          kinderror(obj);
          typex := typevoid
        end
      else if obj^.kind in variables then
        begin
          typex := obj^.typex;
          displ := obj^.vardispl;
          level := blocklevel - obj^.varlevel;
          if (obj^.kind = varparameter)
              or valuereference(obj) then
            emit3(varparam2, level, displ)
          else if cno > 0 then
            emit3(xvariable2, cno, displ)
          else {cno = 0 } if (obj^.kind = valueparameter) then
            emit3(valueparam2, level, displ)
          else {cno = 0 and obj is local variable}
            emit3(variable2, level, displ);
          if istarget then
            include(used.targetobj, obj)
          else
            include(used.exprobj, obj);
          push(1);
        end
      else
        begin
          kinderror(obj);
          typex := typevoid
        end
    end
  else
    begin
      syntaxerror(stop);
      typex := typevoid
    end
end;

{ VariableAccess =
    EntireVariableAccess
      [ ComponentSelector ]* .
  ComponentSelector =
    IndexedSelector |
    FieldSelector . }

procedure variableaccess(
  var typex: structure;
  var used: closure;
  target, restricted: boolean;
  stop: tokens);
var used2: closure;
  stop2: tokens;
begin
  stop2 :=
    selectortokens + stop;
  entirevariableaccess(
    typex, used, target,
    stop2);
  while token in selectortokens do
    begin
      if token = leftbracket1
        then
          begin
            indexedselector(
              typex, used2,
              restricted,
              stop2);
            addclosure(used,
              used2)
           end
         else
           { token = period1 }
           fieldselector(
             typex, stop2);
    end;
end;

{ FileFunctionDesignator =
    "eof" | "eoln" . }

procedure filefunctiondesignator(
  obj: objectt;
  var typex: structure;
  var used: closure;
  stop: tokens);
var id: identifier;
begin
  { symbol = (file function)
      identifier }
  newclosure(used);
  include(used.exprobj,
    inputfile);
  if obj^.arg = eof0 then
    emit2(eof2, lineno)
  else { id = eoln0 }
    emit2(eoln2, lineno);
  push(1);
  typex := typeboolean
end;

{ MathFunctionDesignator =
    MathFunctionIdentifier
      "(" Expression ")" .
  MathFunctionIdentifier =
    "abs" | "arctan" | "chr" |
    "cos" | "exp" | "ln" |
    "odd" | "ord" | "pred" |
    "round" | "sin" | "sqr" |
    "sqrt" | "succ" | "trunc"
    "random". }

procedure mathfunctiondesignator(
  arg: integer;
  var typex: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var stop2, stop3: tokens;
  exprtype: structure;
begin
  { (math function) }
  stop2 :=
    [rightparenthesis1] + stop;
  stop3 :=
    expressiontokens + stop2;
  expect(leftparenthesis1,
    stop3);
  expression(exprtype, used,
    restricted, stop2);
  case arg of
    arctan0, cos0, exp0, ln0,
      sin0, sqrt0:
        begin
          convert1(exprtype,
            typereal);
          checktypes(exprtype,
            typereal);
          if arg = arctan0 then
            emit2(arctan2, lineno)
          else if arg = cos0 then
            emit2(cos2, lineno)
          else if arg = exp0 then
            emit2(exp2, lineno)
          else if arg = ln0 then
            emit2(ln2, lineno)
          else if arg = sin0 then
            emit2(sin2, lineno)
          else { arg = sqrt0 }
            emit2(sqrt2, lineno);
          typex := typereal
        end;
    abs0, sqr0:
      begin
        if exprtype = typeinteger
          then
            if arg = abs0 then
              emit2(absint2,
                lineno)
            else { arg = sqr0 }
              emit2(sqrint2,
                lineno)
          else
            begin
              checktypes(exprtype,
                typereal);
              if arg = abs0 then
                emit2(abs2, lineno)
              else { arg = sqr0 }
                emit2(sqr2, lineno)
            end;
        typex := exprtype
      end;
    odd0:
      begin
        checktypes(exprtype,
          typeinteger);
        typex := typeboolean;
        emit1(odd2)
      end;
    random0:
      begin
        checktypes(exprtype,
          typeinteger);
        typex := typeinteger;
        emit2(random2, lineno)
      end;
    ord0:
      begin
        checktype(exprtype,
          ordtypes);
        typex := typeinteger
      end;
    pred0, succ0:
      begin
        checktype(exprtype,
          ordtypes);
        if arg = pred0 then
          emit3(pred2,
            exprtype^.minvalue,
            lineno)
        else { arg = succ0 }
          emit3(succ2,
            exprtype^.maxvalue,
            lineno);
        typex := exprtype
      end;
    chr0:
      begin
        checktypes(exprtype,
          typeinteger);
        emit2(chr2, lineno);
        typex := typechar
      end;
    round0, trunc0:
      begin
        convert1(exprtype,
          typereal);
        checktypes(exprtype,
          typereal);
        if arg = round0 then
          emit2(round2, lineno)
        else { arg = trunc0 }
          emit2(trunc2, lineno);
        typex := typeinteger
      end
  end;
  expect(rightparenthesis1,
    stop);
  pop(typelength(exprtype));
  push(typelength(typex))
end;

{ StandardFunctionDesignator =
    FileFunctionDesignator |
    MathFunctionDesignator . }

procedure
  standardfunctiondesignator(
    obj: objectt;
    var typex: structure;
    var used: closure;
    restricted: boolean;
    stop: tokens);
{ access restrictions always
  satisfied }
begin
  if obj^.arg in [eof0, eoln0]
    then
      filefunctiondesignator(obj,
        typex, used, stop)
    else
      mathfunctiondesignator(obj^.arg,
        typex, used, restricted,
        stop)
end;

{ ActualParameters =
    [ ActualParameters "," ]
      ActualParameter .
  ActualParameter =
    Expression |
    VariableAccess . }

procedure actualparameters(
  lastparam: objectt;
  var length: integer;
  var distinct: objset;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var typex: structure;
  more: integer;
  used2: closure;
  stop2: tokens;
begin
  { lastparam <> nil }
  if lastparam^.prev <> nil
    then
      begin
        stop2 :=
          expressiontokens
            + stop;
        actualparameters(
          lastparam^.prev,
          more, distinct,
          used, restricted,
          [comma1] + stop2);
        expect(comma1, stop2)
      end
  else more := 0;
  if lastparam^.kind =
        varparameter then
      begin
        variableaccess(
          typex, used2, true,
          restricted, stop);
        if restricted and not
          disjoint(distinct,
            used2.targetobj)
          then error(procedure3);
        addset(distinct,
          used2.targetobj);
        addclosure(used, used2);
        length := 1 + more
      end
  else if valuereference(lastparam) then
      begin
        variableaccess(
          typex, used2, false,
          restricted, stop);
        addclosure(used, used2);
        length := 1 + more
      end
  else { lastparam^.kind =
    valueparameter }
      begin
        expression(typex, used2,
          restricted, stop);
        addclosure(used, used2);
        convert1(typex,
          lastparam^.typex);
        length :=
          typelength(typex)
            + more
      end;
  checktypes(typex,
    lastparam^.typex)
end;

{ ActualParameterPart =
    [ ActualParameterList ] .
  ActualParameterList =
    "(" ActualParameters ")" . }

procedure actualparameterpart(
  procobj: objectt;
  var length: integer;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var distinct: objset;
  procstr: structure;
  stop2: tokens;
begin
  procstr := procobj^.typex;
  copyclosure(used,
    procobj^.implicit);
  include(used.exprobj, procobj);
  if procstr^.lastparam <> nil then
    begin
      stop2 :=
        [rightparenthesis1]
          + stop;
      expect(leftparenthesis1,
        expressiontokens
          + stop2);
      copyset(distinct,
        used.exprobj);
      addset(distinct,
        used.targetobj);
      actualparameters(
        procstr^.lastparam,
        length, distinct,
        used, restricted,
        stop2);
      expect(rightparenthesis1,
        stop)
    end
  else { no parameter list }
    begin
      length := 0;
      syntaxcheck(stop)
    end
end;

{ OverloadedIdent =
    " ^ " Identifier .
  FunctionIdent =
    OverloadedIdent | QualIdent .
  FunctionDesignator =
    FunctionIdent
      ActualParameterPart |
    StandardFunctionDesignator
      . }

{ QualIdent is pre-handled by factor
  which passes procobj and cno }
procedure functiondesignator(
  procobj: objectt;
  var typex: structure;
  var used: closure;
  restricted, isinherited: boolean;
  stop: tokens);
var
  procstr: structure;
  paramlength, resultlength, cno: integer;
begin
  procstr := procobj^.typex;
  if procobj^.kind = standardfunc
    then
      standardfunctiondesignator(procobj,
        typex, used, restricted,
        stop)
    else
      begin
        typex := procstr^.resulttype;
        if dynamic(typex) then
          begin
            emit3(xvariable2,
              typex^.cno,
              typex^.descr);
            push(1);
            emit2(dresult2, lineno);
{            pop(1); push(1);}
          end
        else
          begin
            resultlength := typelength(typex);
            emit2(result2, resultlength);
            push(resultlength);
          end;
        actualparameterpart(
          procobj, paramlength,
          used, restricted,
          stop);
        if within(procobj) then
          procobj^.recursive := true;
        if procobj^.kind = funktion then
          emit3(mcall2,
             blocklevel - procobj^.methodlevel,
             procobj^.methodlabel)
        else { xfunc } begin
          cno := procstr^.cno;
          if isinherited then
            emit3(imcall2, cno,
              procobj^.methodlabel)
          else
            emit3(xmcall2, cno,
              procstr^.descr);
        end;
{WAS
        push(procobj^.varlength);
        pop(procstr^.paramlength
          + procobj^.varlength)
}
        pop(procstr^.paramlength);
      end
end;

{ QualifiedVariableAccess =
  Qualident [ ComponentSelector ] * . }

{ QualIdent is pre-handled by factor
  which passes obj and cno }
procedure qualifiedvariableaccess(
  obj: objectt;
  cno: integer;
  var typex: structure;
  var used: closure;
  target, restricted: boolean;
  stop: tokens);
var used2: closure;
  stop2: tokens;

  procedure entirevariableaccess(
    obj: objectt;
    cno: integer;
    var typex: structure;
    var used: closure;
    istarget: boolean;
    stop: tokens);
  var level, displ: integer;
  begin
    newclosure(used);
    if istarget and valuereference(obj) then
      begin
        kinderror(obj);
        typex := typevoid
      end
    else if obj^.kind in variables then
      begin
        typex := obj^.typex;
        displ := obj^.vardispl;
        if cno = 0 then begin
          { variable not imported }
          level := blocklevel - obj^.varlevel;
          if (obj^.kind = varparameter)
               or valuereference(obj) then
            emit3(varparam2, level, displ)
          else if (obj^.kind = valueparameter) then
            emit3(valueparam2, level, displ)
          else
            emit3(variable2, level, displ);
        end else
          emit3(xvariable2, cno, displ);
        push(1);
        if istarget then
          include(used.targetobj, obj)
        else
          include(used.exprobj, obj)
      end
    else
      begin
        kinderror(obj);
        typex := typevoid
      end
  end;

begin
  stop2 :=
    selectortokens + stop;
  entirevariableaccess(
    obj, cno,
    typex, used, target,
    stop2);
  while token in
    selectortokens do
      if token = leftbracket1
        then
          begin
            indexedselector(
              typex, used2,
              restricted,
              stop2);
            addclosure(used,
              used2)
           end
         else
           { token = period1 }
           fieldselector(
             typex, stop2)
end;

{ Factor =
    UnsignedConstant |
    QualifiedVariableAccess |
    FunctionDesignator |
    "(" Expression ")" |
    "not" Factor . }

procedure factor(
  var typex: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var length, value, cno: integer;
  obj: objectt;
  realvalue: real;
  stringvalue: stringg;
  stop2: tokens;
begin
  if token in literaltokens then
    begin
      newclosure(used);
      unsignedconstant(value,
        realvalue, stringvalue,
        typex, stop);
      if typex^.form in ordtypes
        then
          emit2(ordconst2, value)
      else if typex = typereal
        then
          emitreal(realvalue)
      else { typex = typestring }
        emitstring(stringvalue);
      push(typelength(typex))
    end
  else if token = identifier1 then
    begin
      qualident(obj, cno,
        [leftparenthesis1] + selectortokens + stop);
      if obj^.kind = constant then
        begin
          newclosure(used);
          typex := obj^.typex;
          if obj^.typex = typereal then begin
            loadreal(obj, realvalue);
            typex := typereal;
            emitreal(realvalue)
          end else if obj^.typex = typestring then begin
            loadstring(obj, stringvalue);
            typex := typestring;
            emitstring(stringvalue);
          end else if obj^.typex^.form in ordtypes then begin
            value := obj^.value;
            typex := obj^.typex;
            emit2(ordconst2, value)
          end else begin
            kinderror(obj);
            typex := typevoid
          end;
          push(typelength(typex))
        end
      else if obj^.kind in
        variables then
          begin
            qualifiedvariableaccess(
              obj, cno,
              typex, used, false,
              restricted, stop);
            if dynamic(typex) then
              begin
                emit1(deref2);
                emit2(dvalue2, lineno);
                {pop(1); push(1);}
              end
            else
              begin
                length := typelength(typex);
                emit2(value2, length);
                pop(1); push(length);
              end;
          end
      else if obj^.kind in
        functions then
          functiondesignator(
            obj, {cno,} typex,
            used, restricted,
            { inherited = } false, stop)
      else
        begin
          newclosure(used);
          kinderror(obj);
          typex := typevoid;
          expect(identifier1,
            stop)
        end
    end
  else if token = inherit1 then
    begin
      stop2 := [leftparenthesis1] + stop;
      expect(inherit1, [identifier1] + stop);
      findimport(
        mainmodule^.xbase, {copy of xbase}
        idarg, obj, cno);
      expect(identifier1, stop2);
      if obj^.kind in
        functions then
          functiondesignator(
            obj, {cno,} typex,
            used, restricted,
            { isinherited= } true, stop)
      else
        begin
          newclosure(used);
          kinderror(obj);
          typex := typevoid;
          expect(identifier1,
            stop)
        end;
    end
  else if token = leftparenthesis1
      then
    begin
      stop2 := [rightparenthesis1]
        + stop;
      expect(leftparenthesis1,
        expressiontokens + stop2);
      expression(typex, used,
        restricted, stop2);
      expect(rightparenthesis1,
        stop)
    end
  else if token = not1 then
    begin
      expect(not1,
        factortokens + stop);
      factor(typex, used,
        restricted, stop);
      checktypes(typex,
        typeboolean);
      emit1(not2)
    end
  else
    begin
      newclosure(used);
      syntaxerror(stop);
      typex := typevoid
    end
end;

{ Term =
    Factor [ MultiplyingOperator
      Factor ]* .
  MultiplyingOperator =
    "*" | "/" | "div" | "mod" |
    "and" . }

procedure term(
  var typex: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var operator: integer;
  type2: structure;
  used2: closure;
  stop2: tokens;
begin
  stop2 := multiplytokens + stop;
  factor(typex, used, restricted,
    stop2);
  while token in multiplytokens do
    begin
      operator := token;
      expect(token, factortokens
        + stop2);
      factor(type2, used2,
        restricted, stop2);
      addclosure(used, used2);
      convert2(typex, type2);
      if typex = typeboolean then
        begin
          checktypes(typex, type2);
          if operator = and1 then
            emit1(and2)
          else
            { arithmetic operator }
            typeerror(typex)
        end
      else if typex = typeinteger
          then
        begin
          checktypes(typex, type2);
          if operator = asterisk1
            then 
              emit2(multiply2,
                lineno)
          else
            if operator = div1
              then
                emit2(divide2,
                  lineno)
          else
            if operator = mod1
              then
                emit2(modulo2,
                  lineno)
          else { operator = and1 }
            typeerror(typex)
        end
      else if typex = typereal then
        begin
          checktypes(typex, type2);
          if operator = asterisk1
            then
              emit2(multreal2,
                lineno)
          else
            if operator = slash1
              then
                emit2(divreal2,
                  lineno)
          else { operator in
            [and1, div1, mod1] }
              typeerror(typex)
        end
      else typeerror(typex);
      pop(typelength(typex))
    end
end;

{ SimpleExpression =
    [ Sign ] Term
       [ AddingOperator Term ]* .
  AddingOperator =
    "+" | "-" | "or" . }

procedure simpleexpression(
  var typex: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var operator: integer;
  type2: structure;
  used2: closure;
  stop2: tokens;
begin
  stop2 := addtokens + stop;
  syntaxcheck(signtokens
    + termtokens + stop2);
  if token in signtokens then
    begin
      operator := token;
      expect(token, termtokens
        + stop2);
      term(typex, used,
        restricted, stop2);
      if typex = typeinteger then
        if operator = minus1 then
          emit2(minus2, lineno)
        else { operator = plus1 }
      else if typex = typereal then
        if operator = minus1 then
          emit2(minusreal2, lineno)
        else { operator = plus1 }
      else typeerror(typex)
    end
  else
    term(typex, used, restricted,
      stop2);
  while token in addtokens do
    begin
      operator := token;
      expect(token, termtokens
        + stop2);
      term(type2, used2,
        restricted, stop2);
      addclosure(used, used2);
      convert2(typex, type2);
      if typex = typeboolean then
        begin
          checktypes(typex, type2);
          if operator = or1 then
            emit1(or2)
          else
            { arithmetic operator }
            typeerror(typex)
        end
      else if typex = typeinteger
        then
          begin
            checktypes(typex,
              type2);
            if operator = plus1
              then
                emit2(add2, lineno)
            else
              if operator = minus1
                then
                  emit2(subtract2,
                    lineno)
            else { operator = or1 }
              typeerror(typex)
          end
      else if typex = typereal then
        begin
          checktypes(typex, type2);
          if operator = plus1 then
            emit2(addreal2, lineno)
          else
            if operator = minus1
              then
                emit2(subreal2,
                  lineno)
          else { operator = or1 }
            typeerror(typex)
        end
      else typeerror(typex);
      pop(typelength(typex))
    end
end;

{ Expression =
    SimpleExpression
      [ RelationalOperator
        SimpleExpression ] .
  RelationalOperator =
    "<" | "=" | ">" |
    "<=" | "<>" | ">=" . }

procedure expression{
  typex: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens};
var length, operator: integer;
  type2: structure; used2: closure;
begin
  simpleexpression(typex,
    used, restricted,
    relationtokens + stop);
  if token in relationtokens then
    begin
      operator := token;
      expect(token, simpleexprtokens
        + stop);
      simpleexpression(type2, used2,
        restricted, stop);
      addclosure(used, used2);
      convert2(typex, type2);
      if typex^.form in ordtypes then
        begin
          checktypes(typex, type2);
          if operator = less1
            then emit1(lsord2)
          else if operator = equal1
            then emit1(eqord2)
          else if operator = greater1
            then emit1(grord2)
          else
            if operator = notgreater1
              then emit1(ngord2)
          else
            if operator = notequal1
              then emit1(neord2)
          else
            { operator = notless1 }
              emit1(nlord2)
        end
      else if typex = typereal then
        begin
          checktypes(typex, type2);
          if operator = less1
            then emit1(lsreal2)
          else if operator = equal1
            then emit1(eqreal2)
          else if operator = greater1
            then emit1(grreal2)
          else
            if operator = notgreater1
              then emit1(ngreal2)
          else
            if operator = notequal1
              then emit1(nereal2)
          else
            { operator = notless1 }
              emit1(nlreal2)
        end
      else if typex = typestring then
        begin
          checktypes(typex, type2);
          if operator = less1
            then emit1(lsstring2)
          else if operator = equal1
            then emit1(eqstring2)
          else if operator = greater1
            then emit1(grstring2)
          else
            if operator = notgreater1
              then emit1(ngstring2)
          else
            if operator = notequal1
              then emit1(nestring2)
          else
            { operator = notless1 }
              emit1(nlstring2)
        end
      else { array, record or
         channel type }
        begin
          checktypes(typex, type2);
          length :=
            typelength(typex);
          if operator = equal1 then
            emit2(equal2, length)
          else
            if operator = notequal1
              then
                emit2(notequal2,
                  length)
          else typeerror(typex)
        end;
      pop(2*typelength(typex));
      push(1);
      typex := typeboolean
    end
end;

{ FunctionVariable =
    FunctionIdentifier . }

procedure functionvariable(
  var typex: structure;
  var used: closure;
  stop: tokens);
var procobj: objectt; procstr: structure;
begin
  { token = (function)
      identifier }
  findobj(idarg, procobj, unused);
  newclosure(used);
  if procobj =
      block[blocklevel].heading
    then { in function block }
      begin
        procstr := procobj^.typex;
        typex := procstr^.resulttype;
        emit3(function2, 0,
          - procstr^.paramlength
            - typelength(typex));
        include(used.targetobj,
          procobj);
        push(1)
      end
    else
      begin
        kinderror(procobj);
        typex := typevoid
      end;
  expect(identifier1, stop)
end;

{ AssignmentStatement =
    LeftPart ":=" Expression .
  LeftPart =
    VariableAccess |
    FunctionVariable . }

procedure assignmentstatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var leftpart: objectt;
  lefttype, exprtype: structure;
  length: integer;
  used2: closure;
  stop2, stop3: tokens;
begin
  { token = (left part)
      identifier }
  stop2 :=
    expressiontokens + stop;
  stop3 := [becomes1] + stop2;
  findobj(idarg, leftpart, unused);
  if leftpart^.kind in variables then
    begin
      variableaccess(lefttype,
        used, true, restricted,
        stop3);
    end
  else
    functionvariable(
      lefttype, used, stop3);
  expect(becomes1, stop2);
  expression(exprtype, used2,
    restricted, stop);
  addclosure(used, used2);
  convert1(exprtype, lefttype);
  checktypes(lefttype, exprtype);
  if dynamic(exprtype) then
    begin {reference length = 1}
      emit1(dassign2);
      pop(2);
    end
  else
    begin
      length := typelength(exprtype);
      emit2(assign2, length);
      pop(1 + length);
    end;
end;

{ ReadParameter =
    VariableAccess . }

procedure readparameter(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var typex: structure;
begin
  variableaccess(typex, used,
    true, restricted, stop);
  include(used.targetobj,
    inputfile);
  if typex = typechar then
    emit2(read2, lineno)
  else if typex = typeinteger then
    emit2(readint2, lineno)
  else if typex = typereal then
    emit2(readreal2, lineno)
  else typeerror(typex);
  pop(1)
end;

{ ReadParameters =
    ReadParameter
      [ "," ReadParameter ]* . }

procedure readparameters(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var used2: closure; stop2: tokens;
begin
  stop2 := [comma1] + stop;
  readparameter(used,
    restricted, stop2);
  while token = comma1 do
    begin
      expect(comma1,
        [identifier1] + stop2);
      readparameter(used2,
        restricted, stop2);
      addclosure(used, used2)
    end
end;

{ ReadParameterList =
    "(" ReadParameters ")" . }

procedure readparameterlist(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var stop2: tokens;
begin
  stop2 :=
    [rightparenthesis1] + stop;
  expect(leftparenthesis1,
    [identifier1] + stop2);
  readparameters(used,
    restricted, stop2);
  expect(rightparenthesis1, stop)
end;

{ ReadStatement =
    "read" ReadParameterList |
    "readln"
       [ ReadParameterList ] . }

procedure readstatement(
  arg: integer;
  var used: closure;
  restricted: boolean;
  stop: tokens);
begin
  { (read procedure) }
  expect(identifier1, [leftparenthesis1]
    + stop);
  if arg = read0 then
    readparameterlist(used,
      restricted, stop)
  else { arg = readln0 }
    begin
      if token = leftparenthesis1
        then
          readparameterlist(used,
            restricted, stop)
        else
          begin
            newclosure(used);
            include(used.targetobj,
              inputfile)
          end;
      emit2(readln2, lineno)
    end;
end;

{ WriteOption =
    [ ":" Expression ] . }

procedure writeoption(
  var option: integer;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var typex: structure;
begin
  syntaxcheck([colon1] + stop);
  if token = colon1 then
    begin
      expect(colon1,
        expressiontokens + stop);
      expression(typex, used,
        restricted, stop);
      checktypes(typex,
        typeinteger);
      option := ord(true)
    end
  else
    begin
      newclosure(used);
      option := ord(false)
    end
end;

{ WriteParameter =
    Expression WriteOption
      WriteOption . }

procedure writeparameter(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var option1, option2: integer;
  typex: structure;
  used2: closure;
  stop2: tokens;
begin
  stop2 := [colon1] + stop;
  expression(typex, used,
    restricted, stop2);
  if typex = typereal then
    begin
      writeoption(option1, used2,
        restricted, stop2);
      addclosure(used, used2);
      writeoption(option2, used2,
        restricted, stop);
      addclosure(used, used2);
      emit4(writereal2, option1,
        option2, lineno)
    end
  else
    begin
      writeoption(option1, used2,
        restricted, stop);
      addclosure(used, used2);
      if typex = typechar then
        emit3(write2, option1,
          lineno)
      else if typex = typeboolean
        then
          emit3(writebool2,
            option1, lineno)
      else if typex = typeinteger
        then
          emit3(writeint2, option1,
            lineno)
      else if typex = typestring
        then
          emit3(writestring2,
            option1, lineno)
      else typeerror(typex)
    end;
  include(used.targetobj,
    outputfile);
  pop(typelength(typex))
end;

{ WriteParameters =
    WriteParameter
      [ "," WriteParameter ]* . }

procedure writeparameters(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var used2: closure; stop2: tokens;
begin
  stop2 := [comma1] + stop;
  writeparameter(used,
    restricted, stop2);
  while token = comma1 do
    begin
      expect(comma1,
        expressiontokens + stop2);
      writeparameter(used2,
        restricted, stop2);
      addclosure(used, used2)
    end
end;

{ WriteParameterList =
    "(" WriteParameters ")" . }

procedure writeparameterlist(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var stop2: tokens;
begin
  stop2 :=
    [rightparenthesis1] + stop;
  expect(leftparenthesis1,
    expressiontokens + stop2);
  writeparameters(used,
    restricted, stop2);
  expect(rightparenthesis1, stop)
end;

{ WriteStatement =
    "write" WriteParameterList |
    "writeln"
      [ WriteParameterList ] . }

procedure writestatement(
  arg: integer;
  var used: closure;
  restricted: boolean;
  stop: tokens);
begin
  { (write procedure) }
  expect(identifier1, [leftparenthesis1]
    + stop);
  if arg = write0 then
    writeparameterlist(used,
      restricted, stop)
  else { arg = writeln0 }
    begin
      if token = leftparenthesis1
        then
          writeparameterlist(used,
            restricted, stop)
        else
          begin
            newclosure(used);
            include(used.targetobj,
              outputfile)
          end;
      emit2(writeln2, lineno)
    end
end;

{ OpenParameter =
    ChannelVariableAccess . }

procedure openparameter(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var typex: structure;
begin
  variableaccess(typex, used,
    true, restricted, stop);
  checktype(typex, [channeltype]);
  emit2(open2, lineno);
  pop(1)
end;

{ OpenParameters =
    OpenParameter [ ","
      OpenParameter ]* . }

procedure openparameters(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var used2: closure; stop2: tokens;
begin
  stop2 := [comma1] + stop;
  openparameter(used,
    restricted, stop2);
  while token = comma1 do
    begin
      expect(comma1,
        [identifier1] + stop2);
      openparameter(used2,
        restricted, stop2);
      addclosure(used, used2)
    end
end;

{ OpenStatement =
    "open" "("
      OpenParameters ")" . }

procedure openstatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var stop2, stop3: tokens;
begin
  { token = (open procedure)
      identifier }
  stop2 := [rightparenthesis1]
    + stop;
  stop3 := [identifier1]
    + stop2;
  expect(identifier1,
    [leftparenthesis1]
      + stop3);
  expect(leftparenthesis1,
    stop3);
  openparameters(used,
    restricted, stop2);
  expect(rightparenthesis1,
    stop)
end;

{ InputVariableAccess =
    VariableAccess . }

procedure inputvariableaccess(
  chantype: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var vartype: structure;
  typeno: integer;
begin
  variableaccess(
    vartype, used, true,
    restricted, stop);
  checkmessage(chantype,
    vartype, typeno);
  emit5(receive2, typeno,
    typelength(vartype),
    ord(dynamic(vartype)),
    lineno);
  push(1); pop(2)
end;

{ InputVariableList =
    InputVariableAccess
      [ ","
        InputVariableAccess ]*
          . }

procedure inputvariablelist(
  chantype: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var used2: closure;
  stop2: tokens;
begin
  stop2 := [comma1] + stop;
  inputvariableaccess(
    chantype, used,
    restricted, stop2);
  while token = comma1 do
    begin
      expect(comma1,
        [identifier1]
          + stop2);
      inputvariableaccess(
        chantype, used2,
        restricted, stop2);
      addclosure(used,
        used2)
    end
end;

{ ReceiveParameters =
    ChannelExpression ","
      InputVariableList . }

procedure receiveparameters(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var chantype: structure;
  used2: closure;
  stop2: tokens;
begin
  stop2 := [identifier1]
    + stop;
  expression(chantype,
    used, restricted,
    [comma1] + stop2);
  emit2(checkio2, lineno);
  expect(comma1, stop2);
  inputvariablelist(chantype,
    used2, restricted, stop);
  addclosure(used, used2);
  emit1(endio2);
  pop(1)
end;

{ ReceiveStatement =
    "receive" "("
      ReceiveParameters ")" . }

procedure receivestatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var stop2, stop3: tokens;
begin
  { token = (receive procedure)
      identifier }
  stop2 := [rightparenthesis1]
    + stop;
  stop3 := expressiontokens
    + stop2;
  expect(identifier1,
    [leftparenthesis1]
      + stop3);
  expect(leftparenthesis1,
    stop3);
  receiveparameters(used,
    restricted, stop2);
  expect(rightparenthesis1,
    stop)
end;

{ OutputExpression =
    Expression . }

procedure outputexpression(
  chantype: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var typeno, length: integer;
  exprtype: structure;
begin
  expression(exprtype,
    used, restricted, stop);
  checkmessage(chantype,
    exprtype, typeno);
  length :=
    typelength(exprtype);
  emit5(send2, typeno, length,
    ord(dynamic(exprtype)),
    lineno);
  push(1); pop(length + 1)
end;

{ OutputExpressionList =
    OutputExpression [ ","
      OutputExpression ]*
        . }

procedure
  outputexpressionlist(
    chantype: structure;
    var used: closure;
    restricted: boolean;
    stop: tokens);
var used2: closure;
  stop2: tokens;
begin
  stop2 := [comma1] + stop;
  outputexpression(
    chantype, used,
    restricted, stop2);
  while token = comma1 do
    begin
      expect(comma1,
        expressiontokens
          + stop2);
      outputexpression(
        chantype, used2,
        restricted, stop2);
      addclosure(used,
        used2)
    end
end;

{ SendParameters =
    ChannelExpression ","
      OutputExpressionList
        . }

procedure sendparameters(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var chantype: structure;
  used2: closure;
  stop2: tokens;
begin
  stop2 :=
    expressiontokens
      + stop;
  expression(chantype,
    used, restricted, 
    [comma1] + stop2);
  emit2(checkio2, lineno);
  expect(comma1, stop2);
  outputexpressionlist(
    chantype, used2,
    restricted, stop);
  addclosure(used, used2);
  emit1(endio2);
  pop(1)
end;

{ SendStatement =
    "send" "("
      SendParameters ")"
        . }

procedure sendstatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var stop2, stop3: tokens;
begin
  { token =
      (send procedure)
        identifier }
  stop2 :=
    [rightparenthesis1]
      + stop;
  stop3 :=
    expressiontokens
      + stop2;
  expect(identifier1,
    [leftparenthesis1]
      + stop3);
  expect(leftparenthesis1,
    stop3);
  sendparameters(used,
    restricted, stop2);
  expect(rightparenthesis1,
    stop)
end;

{ StandardProcedureStatement =
    ReadStatement |
    WriteStatement |
    OpenStatement |
    ReceiveStatement |
    SendStatement . }

procedure
  standardprocedurestatement(
    arg: integer;
    var used: closure;
    restricted: boolean;
    stop: tokens);
{ restrictions always satisfied }
begin
  { token =
      (standard procedure)
        identifier }
  if arg in
    [read0, readln0] then
      readstatement(arg, used,
        restricted, stop)
  else if arg in
    [write0, writeln0] then
      writestatement(arg, used,
        restricted, stop)
  else if arg = open0 then
    openstatement(used,
      restricted, stop)
  else if arg = receive0 then
    receivestatement(used,
      restricted, stop)
  else { arg = send0 }
    sendstatement(used,
      restricted, stop)
end;

{ ProcedureStatement =
    ProcedureIdentifier
      ActualParameterPart |
    StandardProcedureStatement
      . }

procedure procedurestatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var paramlength, cno: integer;
  procobj: objectt; procstr: structure;
begin
  { token = (procedure)
      identifier }
  checkbody;
  findobj(idarg, procobj, cno);
  if procobj^.kind = standardproc
    then
      standardprocedurestatement(procobj^.arg,
        used, restricted, stop)
    else
      begin
        procstr := procobj^.typex;
        expect(identifier1,
          [leftparenthesis1]
            + stop);
        actualparameterpart(
          procobj, paramlength,
          used, restricted,
          stop);
        if within(procobj) then
          procobj^.recursive :=
            true;
        with procobj^ do
          if kind = procedur then
            emit3(mcall2,
               blocklevel - methodlevel,
               methodlabel)
          else { xproc }
            emit3(xmcall2, typex^.cno,
               typex^.descr);
{WAS:
        push(procobj^.varlength);
        pop(procstr^.paramlength
          + procobj^.varlength)
}
        pop(procstr^.paramlength);
      end
end;

{ ExternalIdent =
    OverloadedIdent | QualIdent .
  ExternalProcedureStatement =
    ExternalIdent
      ActualParameterPart
      . }

procedure ExternalProcedureStatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var paramlength, cno: integer;
  procobj: objectt; procstr: structure;
  isinherited: boolean;
  stop2: tokens;

begin
  { token =
    ProcedureQualIdent or
    "^" ProcedureIdentifier (of an isinherited method)
  }
  stop2 := [leftparenthesis1] + stop;
  isinherited := (token = inherit1);
  if isinherited then begin
    expect(inherit1, [identifier1] + stop);
    findimport(
      mainmodule^.xbase, {copy of xbase}
      idarg, procobj, cno);
    expect(identifier1, stop2);
  end else
    qualident(procobj,
      cno, stop2);
  checkbody;
  if procobj^.kind in procedures then
    begin
      procstr := procobj^.typex;
      cno := procstr^.cno;
      actualparameterpart(
        procobj, paramlength,
        used, restricted,
        stop);
      if isinherited then
        emit3(imcall2,
          cno, procobj^.methodlabel)
      else 
        emit3(xmcall2,
          cno, procstr^.descr);
      push(procobj^.varlength);
      pop(procstr^.paramlength
        + procobj^.varlength);
    end
  else
    begin
      newclosure(used);
      kinderror(procobj);
    end
end;

{ IfStatement =
    "if" Expression
      "then" Statement
        [ "else" Statement ] . }

procedure ifstatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var label1, label2: integer;
  exprtype: structure;
  used2: closure;
  stop2, stop3: tokens;
begin
  stop2 :=
    statementtokens + stop;
  stop3 :=
    [then1, else1] + stop2;
  expect(if1, expressiontokens
    + stop3);
  expression(exprtype, used,
    restricted, stop3);
  checktypes(exprtype,
    typeboolean);
  expect(then1,
    [else1] + stop2);
  newlabel(label1);
  emit2(do2, label1);
  pop(1);
  statement(used2,
    restricted,
    [else1] + stop);
  addclosure(used, used2);
  if token = else1 then
    begin
      expect(else1, stop2);
      newlabel(label2);
      emit2(goto2, label2);
      emit2(defaddr2, label1);
      statement(used2,
        restricted, stop);
      addclosure(used, used2);
      emit2(defaddr2, label2)
    end
  else emit2(defaddr2, label1)
end;

{ WhileStatement =
    "while" Expression
      "do" Statement . }

procedure whilestatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var label1, label2: integer;
  exprtype: structure;
  used2: closure;
  stop2, stop3: tokens;
begin
  stop2 :=
    statementtokens + stop;
  stop3 := [do1] + stop2;
  newlabel(label1);
  emit2(defaddr2, label1);
  expect(while1,
    expressiontokens + stop3);
  expression(exprtype, used,
    restricted, stop3);
  checktypes(exprtype,
    typeboolean);
  expect(do1, stop2);
  newlabel(label2);
  emit2(do2, label2);
  pop(1);
  statement(used2,
    restricted, stop);
  addclosure(used, used2);
  emit2(goto2, label1);
  emit2(defaddr2, label2)
end;

{ RepeatStatement =
    "repeat" StatementSequence
      "until" Expression . }

procedure repeatstatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var exprtype: structure;
  labelx: integer;
  used2: closure;
  stop2, stop3: tokens;
begin
  stop2 := expressiontokens
    + stop;
  stop3 := [until1] + stop2;
  newlabel(labelx);
  emit2(defaddr2, labelx);
  expect(repeat1,
    statementtokens + stop3);
  statementsequence(used,
    restricted, stop3);
  expect(until1, stop2);
  expression(exprtype, used2,
    restricted, stop);
  addclosure(used, used2);
  checktypes(exprtype,
    typeboolean);
  emit2(do2, labelx);
  pop(1)
end;

{ ForClause =
    "for" EntireVariableAccess
      ":=" Expression . }

procedure forclause(
  var vartype: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var exprtype: structure;
  used2: closure;
  stop2, stop3: tokens;
begin
  stop2 :=
    expressiontokens + stop;
  stop3 := [becomes1] + stop2;
  expect(for1, stop3);
  entirevariableaccess(
    vartype, used, true, stop3);
  checktype(vartype, ordtypes);
  expect(becomes1, stop2);
  expression(exprtype, used2,
    restricted, stop);
  addclosure(used, used2);
  checktypes(vartype, exprtype);
  emit1(for2); pop(1)
end;

{ UpClause =
    "to" Expression
      "do" Statement . }

procedure upclause(
  vartype: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var label1, label2: integer;
  exprtype: structure;
  used2: closure;
  stop2, stop3: tokens;
begin
  stop2 :=
    statementtokens + stop;
  stop3 := [do1] + stop2;
  newlabel(label1);
  newlabel(label2);
  expect(to1, expressiontokens
    + stop3);
  expression(exprtype, used,
    restricted, stop3);
  checktypes(vartype, exprtype);
  expect(do1, stop2);
  emit2(defaddr2, label1);
  emit2(to2, label2);
  statement(used2,
    restricted, stop);
  addclosure(used, used2);
  emit2(endto2, label1);
  emit2(defaddr2, label2);
  pop(2)
end;

{ DownClause =
    "downto" Expression
      "do" Statement . }

procedure downclause(
  vartype: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var label1, label2: integer;
  exprtype: structure;
  used2: closure;
  stop2, stop3: tokens;
begin
  stop2 :=
    statementtokens + stop;
  stop3 := [do1] + stop2;
  newlabel(label1);
  newlabel(label2);
  expect(downto1,
    expressiontokens + stop3);
  expression(exprtype, used,
    restricted, stop3);
  checktypes(vartype, exprtype);
  expect(do1, stop2);
  emit2(defaddr2, label1);
  emit2(downto2, label2);
  statement(used2,
    restricted, stop);
  addclosure(used, used2);
  emit2(enddown2, label1);
  emit2(defaddr2, label2);
  pop(2)
end;

{ ForStatement =
    ForClause ForOption .
  ForOption =
    UpClause | DownClause . }

procedure forstatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var vartype: structure;
  used2: closure;
begin
  forclause(vartype,
    used, restricted,
    [downto1, to1] + stop);
  if token = to1 then
    upclause(vartype, used2,
      restricted, stop)
  else
    downclause(vartype, used2,
      restricted, stop);
  addclosure(used, used2)
end;

{ CaseConstant =
    OrdinalConstant . }

procedure caseconstant(
  exprtype: structure;
  statlabel: integer;
  var length: integer;
  var table: casetable;
  stop: tokens);
var value: integer;
  found: boolean;
  typex: structure;
begin
  ordinalconstant(value,
    typex, stop);
  checktypes(typex, exprtype);
  if length < maxcase then
    begin
      insertcase(value,
        statlabel, length,
        table, found);
      if found then
          error(case3)
    end
  else haltt(maxcase5)
end;

{ CaseConstantList =
    CaseConstant
      [ "," CaseConstant ]* . }

procedure caseconstantlist(
  exprtype: structure;
  statlabel: integer;
  var length: integer;
  var table: casetable;
  stop: tokens);
var stop2: tokens;
begin
  stop2 := [comma1] + stop;
  caseconstant(
    exprtype, statlabel,
    length, table, stop2);
  while token = comma1 do
    begin
      expect(comma1,
        constanttokens + stop2);
      caseconstant(
        exprtype, statlabel,
        length, table, stop2)
    end
end;

{ CaseListElement =
    CaseConstantList
      ":" Statement . }

procedure caselistelement(
  exprtype: structure;
  endlabel: integer;
  var length: integer;
  var table: casetable;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var statlabel: integer;
  stop2: tokens;
begin
  newlabel(statlabel);
  stop2 :=
    statementtokens + stop;
  caseconstantlist(
    exprtype, statlabel,
    length, table,
    [colon1] + stop2);
  expect(colon1, stop2);
  emit2(defaddr2, statlabel);
  statement(used, restricted,
    stop);
  emit2(goto2, endlabel)
end;

{ CaseList =
    CaseListElement
      [ ";" CaseListElement ]* .
        [ ";" ] . }

procedure caselist(
  exprtype: structure;
  endlabel: integer;
  var length: integer;
  var table: casetable;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var used2: closure;
  stop2: tokens;
begin
  stop2 := [semicolon1] + stop;
  length := 0;
  caselistelement(
    exprtype, endlabel,
    length, table,
    used, restricted,
    stop2);
  while token = semicolon1 do
    begin
      { use "stop" here }
      expect(semicolon1,
        constanttokens + stop);
      if token in constanttokens
        then
          begin
            caselistelement(
              exprtype, endlabel,
              length, table,
              used2, restricted,
              stop2);
            addclosure(used,
              used2)
           end
    end
end;

{ CaseStatement =
    "case" Expression "of"
      CaseList "end" . }

procedure casestatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var exprtype: structure;
  beginlabel, endlabel,
  firstline, length:
    integer;
  table: casetable;
  used2: closure;
  stop2, stop3: tokens;
begin
  stop2 := [end1] + stop;
  stop3 :=
    constanttokens + stop2;
  newlabel(beginlabel);
  newlabel(endlabel);
  firstline := lineno;
  expect(case1, [of1] +
    expressiontokens + stop3);
  expression(exprtype, used,
    restricted, [of1] + stop3);
  checktype(exprtype, ordtypes);
  expect(of1, stop3);
  emit2(goto2, beginlabel);
  pop(1);
  caselist(exprtype, endlabel,
    length, table, used2,
    restricted, stop2);
  addclosure(used, used2);
  emit2(defaddr2, beginlabel);
  emitcase(length,
    table, firstline);
  emit2(defaddr2, endlabel);
  expect(end1, stop)
end;

{ StatementSequence =
    Statement
      [ ";" Statement ]* . }

procedure statementsequence{
  var used: closure;
  restricted: boolean;
  stop: tokens};
var used2: closure;
  stop2: tokens;
begin
  stop2 := [semicolon1] + stop;
  statement(used,
    restricted, stop2);
  while token = semicolon1 do
    begin
      expect(semicolon1,
        statementtokens + stop2);
      statement(used2,
        restricted, stop2);
      addclosure(used, used2)
    end
end;

{ CompoundStatement =
    "begin" StatementSequence
       "end" . }

procedure compoundstatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var stop2: tokens;
begin
  stop2 := [end1] + stop;
  expect(begin1,
    statementtokens + stop2);
  statementsequence(used,
    restricted, stop2);
  expect(end1, stop)
end;

{ ProcessStatement =
    StatementSequence . }

procedure processstatement(
    exitlabel: integer;
    var used: closure;
    restricted: boolean;
    stop: tokens);
var templabel, endlabel:
  integer;
begin
  newblock(block[blocklevel].heading);
  newlabel(templabel);
  newlabel(endlabel);
  emit5(process2, {blockno} 0,
    templabel, endlabel, lineno);
  statementsequence(used,
    restricted, stop);
  emit3(defarg2, templabel,
    block[blocklevel].
      maxtemp);
  emit3(endprocess2,
    exitlabel, lineno);
  emit2(defaddr2, endlabel);
  if restricted and unrestricted(used)
    then
      error(parallel3);
  endblock
end;

{ ProcessStatementList =
    ProcessStatement
      [ "|" ProcessStatement ]*
        . }

procedure processstatementlist(
  exitlabel: integer;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var used2: closure;
  stop2: tokens;
begin
  stop2 := [bar1] + stop;
  processstatement(
    exitlabel, used,
    restricted, stop2);
  while token = bar1 do
    begin
      expect(bar1,
        statementtokens + stop2);
      processstatement(
        exitlabel, used2,
        restricted, stop2);
      if restricted and not
        (disjoint(
           used.targetobj,
           used2.targetobj)
        and
         disjoint(
           used.targetobj,
           used2.exprobj)
        and
         disjoint(
           used.exprobj,
           used2.targetobj))
        then error(parallel3);
      addclosure(used, used2)
    end
end;

{ ParallelStatement =
    "parallel"
      ProcessStatementList
        "end" . }

procedure parallelstatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var exitlabel: integer;
  stop2: tokens;
begin
  stop2 := [end1] + stop;
  emit1(parallel2);
  push(1);
  expect(parallel1,
    statementtokens + stop2);
  newlabel(exitlabel);
  processstatementlist(
    exitlabel, used,
    restricted, stop2);
  expect(end1, stop);
  emit2(endparallel2, lineno);
  emit2(defaddr2, exitlabel);
  pop(1)
end;

{ ProcessIndexRange =
    Expression "to"
      Expression . }

procedure processindexrange(
  var typex: structure;
  var used: closure;
  restricted: boolean;
  stop: tokens);
var type2: structure;
  used2: closure;
  stop2: tokens;
begin
  stop2 := expressiontokens
    + stop;
  expression(typex, used,
    restricted,
    [to1] + stop2);
  expect(to1, stop2);
  expression(type2, used2,
    restricted, stop);
  addclosure(used, used2);
  checktype(typex, ordtypes);
  checktypes(typex, type2)
end;

{ IndexVariableDeclaration =
    VariableIdentifier ":="
      ProcessIndexRange . }

procedure
  indexvariabledeclaration(
    var used: closure;
    restricted: boolean;
    stop: tokens);
var id: identifier;
  typex: structure;
  varx: objectt;
  stop2: tokens;
begin
  stop2 := expressiontokens
    + stop;
  expectid(id, [becomes1]
    + stop2);
  expect(becomes1, stop2);
  processindexrange(typex,
    used, restricted, stop);
  newblock(block[blocklevel].heading);
  declareobj(id, variable, varx);
  varx^.typex := typex;
  variableaddressing(1, varx);
  include(used.exprobj, varx)
end;

{ ForallStatement =
    "forall"
      IndexVariableDeclaration
        "do" Statement . }

procedure forallstatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var endlabel, templabel:
  integer; used2: closure;
  stop2, stop3: tokens;
begin
  stop2 := statementtokens
    + stop;
  stop3 := [do1] + stop2;
  expect(forall1,
    [identifier1] + stop3);
  indexvariabledeclaration(
    used, restricted, stop3);
  if restricted and unrestricted(used)
    then
      error(forall3);
  expect(do1, stop2);
  newlabel(templabel);
  newlabel(endlabel);
  emit5(forall2, {blockno} 0,
    templabel, endlabel, lineno);
  statement(used2, restricted,
    stop);
  if restricted and
    (not empty(used2.targetobj)
      or
    unrestricted(used2))
      then
        error(forall3);
  addclosure(used, used2);
  emit3(defarg2, templabel,
    block[blocklevel].
      maxtemp);
  emit2(endall2, lineno);
  emit2(defaddr2, endlabel);
  endblock;
  pop(2)
end;

{ UnrestrictedStatement =
    SicClause Statement .
  SicClause =
    "[" "sic" "]" . }

procedure
  unrestrictedstatement(
    var used: closure;
    stop: tokens);
const restricted = false;
var stop2, stop3: tokens;
begin
  stop2 := statementtokens
    + stop;
  stop3 := [rightbracket1]
    + stop2;
  expect(leftbracket1,
    [sic1] + stop3);
  expect(sic1, stop3);
  expect(rightbracket1, stop2);
  statement(used, restricted,
    stop)
end;

{ AssumeStatement =
    "assume" Expression . }

procedure assumestatement(
  var used: closure;
  restricted: boolean;
  stop: tokens);
var typex: structure;
begin
  expect(assume1,
    expressiontokens + stop);
  expression(typex, used,
    restricted, stop);
  checktypes(typex,
    typeboolean);
  emit2(assume2, lineno);
  pop(1)
end;

{ Statement =
    AssignmentStatement |
    ProcedureStatement |
    ExternalProcedureStatement |
    IfStatement |
    WhileStatement |
    RepeatStatement |
    ForStatement |
    CaseStatement |
    CompoundStatement |
    ParallelStatement |
    ForallStatement |
    UnrestrictedStatement |
    AssumeStatement |
    EmptyStatement .
  EmptyStatement = . }

procedure statement{
  var used: closure;
  restricted: boolean;
  stop: tokens};
var obj: objectt; cno: integer;
begin
  if token = identifier1 then
    begin
      findobj(idarg, obj, cno);
      if obj^.kind in leftparts
        then
          assignmentstatement(
            used, restricted,
            stop)
      else if obj^.kind in
        procedures then
          procedurestatement(
            used, restricted,
            stop)
      else if obj^.kind = module then
        ExternalProcedureStatement(
          used, restricted, stop)
      else
        begin
          newclosure(used);
          kinderror(obj);
          expect(identifier1, stop)
        end
    end
  else if token = if1 then
    ifstatement(used,
      restricted, stop)
  else if token = while1 then
    whilestatement(used,
      restricted, stop)
  else if token = repeat1 then
    repeatstatement(used,
      restricted, stop)
  else if token = for1 then
    forstatement(used,
      restricted, stop)
  else if token = case1 then
    casestatement(used,
      restricted, stop)
  else if token = begin1 then
    compoundstatement(used,
      restricted, stop)
  else if token = parallel1 then
    parallelstatement(used,
      restricted, stop)
  else if token = forall1 then
    forallstatement(used,
      restricted, stop)
  else if token = leftbracket1
  then
    unrestrictedstatement(used,
      stop)
  else if token = assume1 then
    assumestatement(used,
      restricted, stop)
  else if token = inherit1 then
    ExternalProcedureStatement(
          used, restricted, [inherit1] + stop)
  else { empty statement }
    begin
      newclosure(used);
      syntaxcheck(stop)
    end
end;

{ StatementPart =
    CompoundStatement . }

procedure statementpart{
  var templength: integer;
  var used: closure;
  stop: tokens};

begin {statementpart}
  { restricted is a
    global constant }
  allocate;
  compoundstatement(used,
    restricted, stop);
  if blocklevel > mainlevel then
    deallocate;
  templength :=
    block[blocklevel].maxtemp
end;

{ BaseModuleList =
     [ Identifier ":=" ] Identifier
     [ "," [ Identifier ":=" ] Identifier ] . }

procedure basemodulelist(
  ownerobj: objectt;
  stop: tokens);
var impid, extid: identifier;
  fname: stringg;
  imported: objectt;
  done: boolean;
  stop2: tokens;

  procedure appendbase(ownerobj, imported: objectt);
  var imp: objectt;
  begin
    while ownerobj^.xbase <> nil do begin
      imp := imported;
      while imp <> nil do
        if imp^.id = ownerobj^.xbase^.id then
          haltt(import3)
        else
          imp := imp^.xbase;
      ownerobj := ownerobj^.xbase;
    end;
    ownerobj^.xbase := imported;
  end;

begin
  stop2 := [comma1, becomes1] + stop;
  nexttoken;
  done := false;
  repeat
    expectid(impid, stop2);
    if token = becomes1 then begin
      nexttoken;
      expectid(extid, stop2);
    end else extid := impid;
    import(impid, extid, {isbase} true, imported);
    with imported^ do
      emitmoduledesc(loadprivate2, id, key, lineno);
    appendbase(ownerobj, imported);
    if token = comma1 then nexttoken
    else if token = identifier1 then error(syntax3)
    else done := true
  until done;
end;

{ ModuleKeyWord =
    "module" | "module" .
  ModuleHeading =
    ModuleKeyword ModuleIdentifier
     [ "(" BaseModuleList ")" ] ";" . }

procedure moduleheading(
  var moduleid: identifier;
  var maincontxt: objectt;
  stop: tokens);
var stop2, stop3: tokens;
    obj: objectt; found: boolean;

  function differentident(var s1, s2: identifier): boolean;
  var i: integer; ch1, ch2: char;
  begin
    i := 0;
    repeat
      i := i + 1;
      ch1 := s1[i]; small(ch1);
      ch2 := s2[i]; small(ch2);
    until (ch1 <> ch2) or (ch1 = chr(null));
    differentident := (ch1 <> chr(null)) or (ch2 <> chr(null));
  end;

begin
  stop2 :=
    [leftparenthesis1] + stop;
  stop3 :=
    [rightparenthesis1] + stop;
  expect(module1,
    [identifier1] + stop2);
  expectid(moduleid, stop2);
  if differentident(modulename, moduleid) then
    error(name6);
  declareobj(moduleid, module, maincontxt);
  with maincontxt^ do begin
    impno := 0; marked := false;
    lastcomp := nil; xbase := nil
  end;
  newblock(maincontxt);
  if token = leftparenthesis1 then begin
    basemodulelist(maincontxt, stop3);
    expect(rightparenthesis1, stop)
  end;
  expect(semicolon1, stop);
end;

{ ImportedModuleList =
     [ Identifier ":=" ] Identifier
     [ "," [ Identifier ":=" ] Identifier ] . }

procedure importedmodulelist(
  stop: tokens);
var impid, extid: identifier;
  fname: stringg;
  imported: objectt;
  done: boolean;
  stop2: tokens;
begin
  stop2 := [comma1, becomes1] + stop;
  nexttoken;
  done := false;
  repeat
    expectid(impid, stop2);
    if token = becomes1 then begin
      nexttoken;
      expectid(extid, stop2);
    end else extid := impid;
    import(impid, extid, {isbase} false, imported);
      emitmoduledesc(loadshared2, extid,
        imported^.key, lineno);
    if token = comma1 then nexttoken
    else if token = identifier1 then error(syntax3)
    else done := true
  until done;
end;

{ Module =
    ModuleHeading
    [ "import" ImportedModuleList ";" ]
    Block "." . }

procedure modulet(stop: tokens);
var used: closure;
  varlabel, templabel,
  beginlabel, {varlength, now global}
  templength, keylabel: integer;
  moduleid: identifier;
  stop2, stop3, stop4, stop5: tokens;
 
begin { modulet }
  stop2 := [period1] + stop;
  stop3 := [begin1] + stop2;
  stop4 := declarationtokens
    + [semicolon1] + stop3;
  stop5 := [import1] + stop4;
  moduleheading(moduleid,
    mainmodule, stop5);
  newlabel(keylabel);
  emitmoduledesc(modulehead2,
    moduleid, keylabel, lineno);
  if token = import1 then begin
    importedmodulelist(stop4);
    expect(semicolon1, stop4);
  end;
  modulevarlen :=
    {allocate memory for module descriptors}
    modulevarlen + nofgcont + 1;
  newlabel(varlabel);
  newlabel(templabel);
  newlabel(beginlabel);
  emit5(moduleinit2,
    varlabel, templabel, beginlabel,
    lineno);
  declarationpart(modulevarlen,
    stop3);                          
  emit3(defarg2, varlabel,
    modulevarlen);
  emit2(defaddr2, beginlabel);
  install(externals);        {must follow beginlabel}
{
  installglobal;  {must follow beginlabel/
  installlocal;   {must follow installGlobal
}
  emit1(endinit2);
  emit2(moduleenter2, varlabel);
  findlen(externals);        {must follow moduleenter}
{
  findlenglobal;  {must follow moduleenter/
  findlenlocal;
}
  statementpart(templength,
    used, stop2);
  emit3(defarg2, templabel,
    templength);
  endblock;
  randomize;
  with mainmodule^ do begin
    key := random(maxint);
    lastcomp := block[mainlevel].lastobject;
  end;
  emit3(defarg2, keylabel, mainmodule^.key);
  emit1(endmodule2);
  expect(period1, stop);
end;

begin
  initialize;
  nexttoken;
  Modulet([endtext1]);
end;

procedure initparser;
begin end;

end.
