unit pScan;
interface
uses pCommon;

var
  token, argument: integer;
  realarg: real;
  textarg: stringg;
  idarg: identifier;

procedure gettoken;
procedure initscanner;

implementation
uses pTexim;
type
  charset = set of char;

var
  ch: char; afterperiod: boolean;
  alphanumeric,
  digits, endcomment, endline,
  invisible, letters, radix,
  separators: charset;

{ INPUT }

procedure nextchar;
var skipped: boolean;
begin
  repeat
    accept(ch);
    if (ch < chr(null))
      or (ch > chr(del))
    then
      skipped := true
    else
      skipped := ch in invisible
  until not skipped
end;

{ WordSymbol =
    "and" | "array" | "assume" | "begin" | "case" | "class" |
    "const" | "context" | "div" | "do" | "downto" |
    "else" | "end" | "for" | "forall" | "function" |
    "if" | "import" | "mod" | "module" | "not" | "object" |
    "of" | "or" | "parallel" | "procedure" |
    "record" | "repeat" | "sic" | "then" |
    "to" | "type" | "until" | "var" | "while" .

  UnusedWord =
    "in" | "nil" | "set" | "with" . }

{ LEXICAL ANALYSIS }

{ Comment =
    LeftBrace [ CommentElement ]*
      RightBrace .
  CommentElement =
    GraphicCharacter | NewLine |
    Comment . }

procedure comment;
begin
  (* ch = '{' *) nextchar;
  while not (ch in endcomment) do
    if ch = '{' then comment
    else
      begin
        if ch = chr(nl) then
          begin
            nextchar;
            token := newline1;
          end
        else nextchar
      end;
  if ch = '}' then nextchar
  else error(comment3)
end;

{ Word =
    WordSymbol | Identifier .
  Identifier =
    Letter [ Letter | Digit ]* . }

procedure word;
var symbol: boolean;
  i, k: integer;
begin
  { ch in letters }
  i := 1;
  while ch in alphanumeric do
    begin
      small(ch);
      if i < maxident then begin
        idarg[i] := ch; i := i + 1
      end;
      nextchar;
    end;
  for k := i to maxident do idarg[k] := chr(null);
  searchkw(idarg, token);
end;

function scaled(r: real;
  s: integer): real;
{ scaled(r,s) = r*(10**s) }
var maxx, min: real;
begin
  maxx := maxreal / 10.0;
  while s > 0 do
    begin
      if r <= maxx then
        r := r * 10.0
      else error(number3);
      s := s - 1
    end;
  min := 10.0 * minreal;
  while s < 0 do
    begin
      if r >= min then
        r := r / 10.0
      else r := 0.0;
      s := s + 1
    end;
  scaled := r
end;

{ DigitSequence =
    Digit [ Digit ]* . }

procedure digitsequence(
  var r: real;
  var n: integer);
{ r := digitsequence;
  n := length(r) }
var d: real;
begin
  r := 0.0; n := 0;
  if ch in digits then
    while ch in digits do
      begin
        d := ord(ch) - ord('0');
        r := 10.0 * r + d;
        n := n + 1;
        nextchar
      end
  else error(number3)
end;

{ UnsignedScaleFactor =
    DigitSequence . }

procedure unsignedscalefactor(
  var s: integer);
{ s := scalefactor }
var r: real; n: integer;
begin
  digitsequence(r, n);
  if r > maxint then
    begin
      error(number3);
      s := 0
    end
  else s := trunc(r)
end;

{ ScaleFactor =
    [ Sign ]
      UnsignedScaleFactor .
  Sign =
    "+" | "-" . }

procedure scalefactor(
  var s: integer);
{ s := scalefactor }
begin
  if ch = '+' then
    begin
      nextchar;
      unsignedscalefactor(s)
    end
  else if ch = '-' then
    begin
      nextchar;
      unsignedscalefactor(s);
      s := - s
    end
  else unsignedscalefactor(s)
end;

{ UnsignedInteger =
    DigitSequence . }

procedure unsignedinteger(
  r: real);
var i: integer;
begin
  if r >  maxint then
    begin
      error(number3);
      i := 0
    end
  else i := trunc(r);
  token := intconst1; argument := i
end;

{ UnsignedNumber =
    UnsignedReal |
    UnsignedInteger .
  UnsignedReal =
    IntegerPart RealOption .
  IntegerPart =
    DigitSequence .
  RealOption =
    "." FractionalPart
      [ ScalingPart ] |
    ScalingPart .
  FractionalPart =
    DigitSequence .
  ScalingPart =
    Radix ScaleFactor .
  Radix =
    "e" | "E" . }

procedure unsignednumber;
var i, f, r: real;
  s, n: integer;
begin
  digitsequence(i, n);
  if ch = '.' then
    begin
      nextchar;
      if ch = '.' then
        begin
          { input = i..
            and ch = '.' }
          unsignedinteger(i);
          afterperiod := true
        end
      else
        begin
          digitsequence(f, n);
          r := i + scaled(f, -n);
          { r = i.f }
          if ch in radix then
            begin
              nextchar;
              scalefactor(s);
              r := scaled(r, s)
              {r = i.f*(10**s) }
            end;
          token := realconst1; realarg := r
        end
    end
  else if ch in radix then
    begin
      nextchar;
      scalefactor(s);
      r := scaled(i, s);
      { r = i*(10**s) }
      token := realconst1; realarg := r
    end
  else unsignedinteger(i)
end;

{ CharElement =
    GraphicCharacter |
    ApostropheImmage .
  ApostropheImmage =
    "''" . }

procedure charelement(
  var length: integer);
begin
  if length = maxstring then
    haltt(maxstring5);
  length := length + 1;
  textarg[length] := ch;
  nextchar
end;

{ CharacterString =
    "'" StringElements "'" .
  StringElements =
    CharElement
      [ CharElement ]* .
  CharacterConstant =
    "'" CharElement "'" . }

procedure characterstring;
type state =
  (extend, accept, reject);
var length: integer; s: state;
begin
  { ch = apostrophe }
  textarg := nulstring;
  length := 0;
  nextchar; s := extend;
  while s = extend do
    if ch in endline then
      s := reject
    else if ch = apostrophe then
      begin
        nextchar;
        if ch = apostrophe then
          charelement(length)
        else s := accept
      end
    else
      charelement(length);
  if (s = accept) and (length > 0)
    then
      if length = 1 then begin
        token := charconst1;
        argument := ord(textarg[1])
      end else begin
        token := stringconst1;
        argument := length;
      end
    else token := unknown1
end;

{ TokenField =
    [ Separator ]* Token .
  Token =
    Literal | Identifier |
    SpecialSymbol | UnknownToken |
    EndText .
  Literal =
    UnsignedNumber |
    CharacterString .
  SpecialSymbol =
    "(" | ")" | "*" | "+" | "," |
    "-" | "." | "/" | ":" | ";" |
    "<" | "=" | ">" | "[" | "]" |
    ".." | ":=" | "<=" | "<>" |
    ">=" | "|" | | "^" | WordSymbol .
  UnknownToken =
    UnusedWord | UnusedCharacter .
  UnusedCharacter =
    "!" | """ | "#" | "$" | "%" |
    "&" | "?" | "@" | "\" |
    "_" | "`" | "~" . }

procedure gettoken;
begin
  while ch in separators do
    if ch = sp then nextchar
    else if ch = chr(nl) then
      begin
        nextchar;
        token := newline1;
      end
    else (* ch = '{' *) comment;
  case ch of
    { letters } 'A'..'Z', 'a'..'z': word;
    { digits }  '0'..'9': unsignednumber;
    apostrophe: characterstring;
    '+': begin
           token := plus1;
           nextchar
         end;
    '-': begin
           token := minus1;
           nextchar
         end;
    '*': begin
           token := asterisk1;
           nextchar;
         end;
    '/': begin
           token := slash1;
           nextchar
         end;
    '<': begin
           nextchar;
           if ch = '=' then begin
             token := notgreater1;
             nextchar
           end else if ch = '>' then begin
             token := notequal1;
             nextchar
           end else token := less1
         end;
    '=': begin
           token := equal1;
           nextchar
         end;
    '>': begin
           nextchar;
           if ch = '=' then begin
             token := notless1;
             nextchar
           end else token := greater1
         end;
    ':': begin
           nextchar;
           if ch = '=' then begin
             token := becomes1;
             nextchar
           end else token := colon1
         end;
    '(': begin
           token := leftparenthesis1;
           nextchar
         end;
    ')': begin
           token := rightparenthesis1;
           nextchar
         end;
    '[': begin
           token := leftbracket1;
           nextchar
         end;
    ']': begin
           token := rightbracket1;
           nextchar
         end;
    ',': begin
           token := comma1;
           nextchar
         end;
    '.': if afterperiod then begin
           token := doubledot1;
           nextchar;
           afterperiod := false
         end else begin
           nextchar;
           if ch = '.' then begin
             token := doubledot1;
             nextchar
           end else token := period1
         end;
    ';': begin
           token := semicolon1;
           nextchar
         end;
    '|': begin
           token := bar1;
           nextchar
         end;
    '^': begin
           token := inherit1;
           nextchar
         end;
    else
      begin
        if (token = unknown1) and (ch <> chr(etx)) then
          begin
            token := unknown1;
            nextchar
          end
        else
          token := endtext1
      end
  end
end;

procedure initscanner;
var i: integer;
begin
  digits := ['0'..'9'];
  letters :=
    ['a'..'z'] + ['A'..'Z'];
  alphanumeric := letters + digits;
  endcomment := ['}', chr(etx)];
  endline := [chr(nl), chr(etx)];
  invisible :=
    [chr(0)..chr(31), chr(127)]
      - [chr(nl), chr(etx)];
  radix := ['e', 'E'];
  separators := [sp, chr(nl), '{'];
  afterperiod := false;
  ch := chr(nl);
end;

end.