unit pTexim;
interface
uses pCommon, pAssem;

{ table types and variables }

type
  objset = ^ atom;
  messagenode = ^ messagerecord;
  objectt = ^ objectrecord;
  structure = ^ structrecord;
  class = (
      { structure form }
    arraytype, booltype, channeltype, chartype, enumtype,
    inttype, methodtype, realtype, classtype, xclasstype,
    void,

      { object kind }
    constant, field, funktion, procedur, xfunc, xproc,
    module, standardfile, standardfunc, standardproc,
    typ, valueparameter, variable, varparameter,
    undefined);
  objectkind = constant .. undefined;
  kinds = set of objectkind;
  atom =
    record
      next: objset;
      item: objectt
    end;
  closure =
    record
      exprobj, targetobj: objset;
    end;
  messagerecord =
    record
      prev: messagenode;
      typex: structure;
    end;
  objectrecord =
    record
      id: identifier;
      prev: objectt;
      typex: structure;
      marked: boolean;
      case kind: objectkind of
        constant:
          (value, value2: integer);
          { ordvalue, realvalue, stringpointer }
        typ:
          (typelevel: integer);
        field:
          (fielddispl: integer;);
        valueparameter, variable, varparameter:
          (varlevel, vardispl: integer);
        funktion, procedur,
        xfunc, xproc:
          (methodlevel, methodlabel,
             varlength: integer;
           recursive: boolean;
           implicit: closure); {can go to structure}
        standardfile, standardfunc,
          standardproc:
          (arg: integer);
        module:
          (key,
           expno {exported module number},
           impno {global imported number}: integer;
           lastcomp: objectt;
           xbase: objectt);
    end;
  structform = arraytype .. void;
  forms = set of structform;
  structrecord =
    record
      obj: objectt;
      cno: integer;
      ref: integer;
      dyncomps: integer; {number of dynamic components}
      descr: integer;
        {diplacement to structure descriptor
        for dynamic types and methods}
      case form: structform of
        booltype, chartype, enumtype, inttype:
          (minvalue, maxvalue: integer;
          {lastenum: objectt});
        realtype:
          ( );
        arraytype:
          (lowerbound, upperbound, arraylength: integer; 
           indextype, elementtype: structure;
           );
        classtype, xclasstype:
          (recordlength: integer;
           lastfield: objectt;
           baseclass: structure;
          );
        channeltype:
          (messagelist:
             messagenode);
        methodtype:
          (paramlength: integer;
           lastparam: objectt;
           restricted: boolean;
           resulttype: structure;
           );
        void:
          ( )
    end;
  blockrecord =
    record
      templength, maxtemp: integer;
      heading, lastobject: objectt
    end;
  blocktable = array [0..maxlevel] of
    blockrecord;
  imprecord =
    record
      impmodule: objectt;
      isbase: boolean;
    end;
  imptable = array [0..maximps] of imprecord;

var
  block: blocktable;
  blocklevel: integer;
  mainmodule: objectt;
  inputfile, outputfile: objectt;
  typeboolean, typechar,
  typeinteger, typereal,
  typestring,
  typevoid: structure;
  glbcont: imptable;
  nofgcont: integer;   { number of imports }

{ initialization }

procedure inittexim;

{ scanner routines }
procedure searchkw(var idarg: identifier;
  var token: integer);

procedure export;
{
  (mainmodule: objectt);
}
procedure import(
  var impid, extid: identifier;
  isbase: boolean; var imported: objectt);

{ conversions }
procedure loadreal(
  obj: objectt;
  var value: real);
procedure storereal(
  obj: objectt;
  value: real);
procedure loadstring(
  obj: objectt;
  var value: stringg);
procedure storestring(
  obj: objectt;
  value: stringpointer);

{ object sets and closures }
procedure newset(
  var x: objset);
procedure newclosure(
  var used: closure);

{ scope analysis }
procedure searchobj(
  var id: identifier;
  levelno: integer;
  var obj: objectt;
  var impno: integer);
procedure findobj(
  var id: identifier;
  var obj: objectt;
  var impno: integer);
procedure findfield(
  str: structure;
  var id: identifier;
  var res: objectt);
procedure findimport(
    cont: objectt;
    var id: identifier;
    var res: objectt;
    var impno: integer);
procedure findbase(
    var id: identifier;
    var res: objectt;
    var impno: integer);
procedure enterobj(
  var id: identifier;
  kind: objectkind;
  var obj: objectt);
procedure declareobj(
  var id: identifier;
  kind: objectkind;
  var obj: objectt);
procedure newblock(
  heading: objectt);
procedure endblock;
function within(proc: objectt)
  : boolean;
procedure checkbody;
procedure newstruct(
  form: structform;
  strobj: objectt;
  var str: structure);
function dynamic(str: structure): boolean;

implementation

{ scanner constants, types, and variables }
const
  wordsize = 10; maxkw = 40;

type
  shortword = packed array[1..wordsize]of char;
  kwrec = record
    token: integer;
    word: shortword
  end;
  kwarray = array[0..maxkw] of kwrec;
  swarray = array[1..maxstandard] of identifier;
var
  kwtable: kwarray;
  keywords: integer;
  swtable: swarray;

{export / import constants, types, and variables}
const
  sftag = 123;
  firststr = ord(void) + 1;
  notyetexp = 0;

{export/import tags}

  error6       =    0;
  constant6    =    1;
  enumconst6   =    2;
  typex6       =    3;
  typey6       =    4;
  variable6    =    5;
  method6      =    6;

  channel6     =    7;
  classx6      =    8;
  array6       =    9;
  methodstr6   =   10;
  classy6      =   11;
  enumtype6    =   12;
  paramlist6   =   13;
  valueparam6  =   14;
  varparam6    =   15;
  fieldlist6   =   16;
  field6       =   17;
  hiddenfield6 =   18;
  messagelist6 =   19;
  message6     =   20;

  sharedmodule6  =   21;
  basemodule6    =   22;

type
  symbolfile = file of char;
  structtable = array [0..maxstruct] of structure;

var
  exportf, oldexpf: symbolfile;
    {used for export only}
  nofexp: integer;
  strno: integer;
  struct:  structtable;

{ scanner routines }

procedure enterkw(
  kw: shortword;
  token: integer);
var i, k: integer; w: shortword; ch: char;
begin
  i := 1; ch := kw[i];
  while ch <> sp do begin
    w[i] := ch; i := i + 1; ch := kw[i]
  end;
  for k := i to wordsize do w[k] := chr(null);
  keywords := keywords + 1;
  kwtable[keywords].token := token;
  kwtable[keywords].word := w;
end;

procedure sortwords;
  var i, j: integer;
begin
  for i := keywords downto 2 do
    for j := 1 to i - 1 do
      if kwtable[j].word > kwtable[j + 1].word then
        begin {swap}
          kwtable[0] := kwtable[j + 1];
          kwtable[j + 1] := kwtable[j];
          kwtable[j] := kwtable[0]
        end; {if}
end;

procedure entersw(
  sw: shortword;
  i0: integer);
var i, k: integer; w: identifier; ch: char;
begin
  i := 1; ch := sw[i];
  while ch <> sp do begin
    w[i] := ch; i := i + 1; ch := sw[i]
  end;
  for k := i to maxident do w[k] := chr(null);
  swtable[i0] := w;
end;

procedure initwords;
var i: integer;
begin
  keywords := 0;
  { enter keywords }
  enterkw('and       ', and1);
  enterkw('array     ', array1);
  enterkw('assume    ', assume1);
  enterkw('begin     ', begin1);
  enterkw('case      ', case1);
  enterkw('class     ', class1);
  enterkw('const     ', const1);
  enterkw('context   ', module1);
  enterkw('div       ', div1);
  enterkw('do        ', do1);
  enterkw('downto    ', downto1);
  enterkw('else      ', else1);
  enterkw('end       ', end1);
  enterkw('for       ', for1);
  enterkw('forall    ', forall1);
  enterkw('function  ', function1);
  enterkw('if        ', if1);
  enterkw('import    ', import1);
  enterkw('mod       ', mod1);
  enterkw('module    ', module1);
  enterkw('not       ', not1);
  enterkw('of        ', of1);
  enterkw('or        ', or1);
  enterkw('parallel  ', parallel1);
  enterkw('procedure ', procedure1);
  enterkw('record    ', class1);
  enterkw('repeat    ', repeat1);
  enterkw('sic       ', sic1);
  enterkw('then      ', then1);
  enterkw('to        ', to1);
  enterkw('type      ', type1);
  enterkw('until     ', until1);
  enterkw('var       ', var1);
  enterkw('while     ', while1);
  sortwords;

  { enter standard identifiers }
  entersw('abs       ', abs0);
  entersw('arctan    ', arctan0);
  entersw('boolean   ', boolean0);
  entersw('char      ', char0);
  entersw('chr       ', chr0);
  entersw('cos       ', cos0);
  entersw('eof       ', eof0);
  entersw('eoln      ', eoln0);
  entersw('exp       ', exp0);
  entersw('false     ', false0);
  entersw('integer   ', integer0);
  entersw('ln        ', ln0);
  entersw('maxint    ', maxint0);
  entersw('maxstring ', maxstring0);
  entersw('null      ', null0);
  entersw('odd       ', odd0);
  entersw('open      ', open0);
  entersw('ord       ', ord0);
  entersw('pred      ', pred0);
  entersw('read      ', read0);
  entersw('random    ', random0);
  entersw('readln    ', readln0);
  entersw('real      ', real0);
  entersw('receive   ', receive0);
  entersw('round     ', round0);
  entersw('send      ', send0);
  entersw('sin       ', sin0);
  entersw('sqr       ', sqr0);
  entersw('sqrt      ', sqrt0);
  entersw('string    ', string0);
  entersw('succ      ', succ0);
  entersw('true      ', true0);
  entersw('trunc     ', trunc0);
  entersw('write     ', write0);
  entersw('writeln   ', writeln0);
end;

procedure searchkw(var idarg: identifier;
  var token: integer);
  var i, leftbound, rightbound, middleindex: integer;
    notfound: boolean;
    id: shortword;
begin
  for i := 1 to wordsize do id[i] := idarg[i];
  notfound := true;
  leftbound := 1;
  rightbound := keywords;
  while notfound and (leftbound <= rightbound) do
    begin
      middleindex := (leftbound + rightbound) div 2;
      if id < kwtable[middleindex].word then
        rightbound := middleindex - 1
      else if id > kwtable[middleindex].word then
        leftbound := middleindex + 1
      else
        notfound := false;
    end; {while}
  if notfound then
    token := identifier1
  else
    token := kwtable[middleindex].token;
end;

{ conversions }

procedure loadreal(
  obj: objectt;
  var value: real);
var dual: dualreal;
begin
  dual.split := true;
  dual.b := obj^.value;
  dual.c := obj^.value2;
  dual.split := false;
  value := dual.a
end;

procedure storereal(
  obj: objectt;
  value: real);
var dual: dualreal;
begin
  dual.split := false;
  dual.a := value;
  dual.split := true;
  obj^.value := dual.b;
  obj^.value2 := dual.c
end;

procedure loadstring(
    obj: objectt;
    var value: stringg);
var ptr: stringpointer;
    dual: dualstring;
begin
  dual.split := true;
  dual.b := obj^.value;
  dual.c := obj^.value2;
  dual.split := false;
  ptr := dual.a;
  value := ptr^
end;

procedure storestring(
  obj: objectt;
  value: stringpointer);
var dual: dualstring;
begin
  dual.split := false;
  dual.a := value;
  dual.split := true;
  obj^.value := dual.b;
  obj^.value2 := dual.c
end;

     { var sets and closures }

procedure newset(
  var x: objset);
{ x := [] }
begin x := nil end;

procedure newclosure(
  var used: closure);
{ used := [] }
begin
  newset(used.targetobj);
  newset(used.exprobj);
end;

     { scope analysis }

procedure findbase(
    var id: identifier;
    var res: objectt;
    var impno: integer);
  var
    more: boolean;
    xbase: objectt;
begin
  xbase := block[mainlevel].heading^.xbase;
  more := xbase <> nil;
  impno := 0; res := nil;
  while more do begin
    if xbase = nil then
      more := false
    else if xbase^.id = id then begin
      more := false; res := xbase;
      impno := xbase^.impno;
    end else begin
      xbase := xbase^.xbase;
      more := xbase <> nil;
    end;
  end;
end;

procedure searchbase(
    xbase: objectt;
    var id: identifier;
    var res: objectt;
    var impno: integer);
  var
    more: boolean; obj: objectt;
begin
  more := xbase <> nil;
  impno := 0; obj := nil;
  while more do begin
    obj := xbase^.lastcomp;
    impno := xbase^.impno;
    while more do
      if obj = nil then
        more := false
      else if obj^.id = id then
        more := false
      else
        obj := obj^.prev;
    if obj = nil then begin
      xbase := xbase^.xbase;
      more := xbase <> nil
    end;
  end;
  res := obj;
end;

procedure searchobj(
  var id: identifier;
  levelno: integer;
  var obj: objectt;
  var impno: integer);
var more: boolean;
begin
  more := true; impno := 0;
  obj :=  block[levelno].lastobject;
  while more do
    if obj = nil then
      more := false
    else if obj^.id = id then
      more := false
    else
      obj := obj^.prev;
  if (obj = nil) and (levelno = mainlevel) then
    searchbase(
      block[levelno].heading^.xbase,
      id, obj, impno);
end;

procedure enterobj(
  var id: identifier;
  kind: objectkind;
  var obj: objectt);
begin
  new(obj);
  obj^.id := id; {obj^.marked := false;}
  obj^.prev :=
    block[blocklevel].lastobject;
  obj^.kind := kind;
  block[blocklevel].lastobject :=
    obj
end;

procedure declareobj(
  var id: identifier;
  kind: objectkind;
  var obj: objectt);
var unused: integer;
  other: objectt;
begin
  if id <> nameless then
    begin
      searchobj(id, blocklevel,
        other, unused);
      if other <> nil {found} then
        error(ambiguous3)
    end;
  enterobj(id, kind, obj);
end;

procedure findobj(
  var id: identifier;
  var obj: objectt;
  var impno: integer);
var more: boolean;
  levelno: integer;
begin
  more := true;
  levelno := blocklevel;
  while more do
    begin
      searchobj(id, levelno, obj, impno);
      if (obj <> nil) {found} or
        (levelno < mainlevel) then
          more := false
      else levelno := levelno - 1
    end;
  if obj = nil {not found} then
    begin
      error(undefined3);
      declareobj(id, undefined, obj)
    end
end;

procedure findfield(
  str: structure;
  var id: identifier;
  var res: objectt);
var
  more: boolean;
  fld: objectt;
begin
  if str^.form = xclasstype then
    str := str^.baseclass;
  more := true;
  while more do begin
    fld := str^.lastfield;
    while more do
      if fld = nil then
        more := false
      else if fld^.id = id then
        more := false
      else
        fld := fld^.prev;
    if fld = nil then begin
      str := str^.baseclass;
      more := str <> typevoid
    end;
  end;
  res := fld;
end;

procedure findimport(
    cont: objectt;
    var id: identifier;
    var res: objectt;
    var impno: integer);
  var
    more: boolean;
begin
  searchbase(cont, id, res, impno);
  if res = nil then begin
    error(undefined3);
    new(res);
    res^.id := id;
    res^.kind := undefined;
    res^.prev := cont^.lastcomp;
    cont^.lastcomp := res
  end;
end;

function within(proc: objectt)
  : boolean;
var levelno: integer;
  more: boolean;
begin
  more := true;
  levelno := blocklevel;
  while more do
    if block[blocklevel].heading
        = proc then
      begin
        more := false;
        within := true
      end
    else if levelno < mainlevel {= 0} then
      begin
        more := false;
        within := false
      end
    else levelno := levelno - 1
end;

procedure checkbody;
var proc: objectt;
begin
  proc :=
    block[blocklevel].heading;
  if proc <> nil then
    if proc^.kind = funktion then
       error(body3)
end;

procedure newstruct(form: structform;
  strobj: objectt;
  var str: structure);
begin new(str);
  str^.form := form; str^.cno := 0;
  str^.ref := 0; str^.dyncomps := 0;
  str^.descr := 0; str^.obj := strobj;
end;

function dynamic(str: structure): boolean;
begin
  with str^ do
    dynamic :=
      (form = xclasstype) or
      (dyncomps > 0);
end;

procedure newblock(
  heading: objectt);
var current: blockrecord;
begin
  if blocklevel = maxlevel then
    haltt(maxlevel5);
  blocklevel := blocklevel + 1;
  current.templength := 0;
  current.maxtemp := 0;
  current.heading := heading;
  current.lastobject := nil;
  block[blocklevel] := current
end;

procedure endblock;
begin
  blocklevel := blocklevel - 1
end;

{ initialization of the standard block }

procedure declstandard(i0: integer;
    kind: objectkind;
    var obj: objectt);
  var id: identifier; ch: char;
begin
  id := swtable[i0];
  declareobj(id, kind, obj);
  obj^.arg := i0;
  obj^.typex := typevoid
end;

procedure initstruct(var typ: structure;
  f: structform);
begin
  new(typ);
  with typ^ do begin
    form := f;
    ref := ord(f);
    cno := 0;
    dyncomps := 0;
  end;
end;

procedure declstringtype(var str: structure);
  var obj: objectt;
begin
  declstandard(string0, typ, obj);
  newstruct(arraytype, obj, str);
  obj^.typex := str; 
  str^.ref := ord(arraytype);
  str^.lowerbound := 1;
  str^.upperbound := maxstring;
  str^.arraylength := maxstring;
  str^.indextype := typeinteger;
  str^.elementtype := typechar;
end;

procedure declrealtype(var str: structure);
  var obj: objectt;
begin
  declstandard(real0, typ, obj);
  newstruct(realtype, obj, str);
  obj^.typex := str;
  str^.ref := ord(realtype);
end;

procedure declordinaltype(
    i0: integer;
    f: structform;
    minvalue, maxvalue: integer;
    var str: structure
);
  var obj: objectt;
begin
  declstandard(i0, typ, obj);
  newstruct(f, obj, str);
  obj^.typex := str; 
  str^.ref := ord(f);
  str^.minvalue := minvalue;
  str^.maxvalue := maxvalue;
end;

procedure declconst(
    i0: integer;
    typex: structure;
    value: integer);
  var obj: objectt;
    id: identifier;
begin
  id := swtable[i0];
  declareobj(id, constant, obj);
  obj^.typex := typex;
  obj^.value := value;
end;

procedure inittables;
var p: objectt; i: integer;
begin
  strno := 0; nofgcont := 0;
  for i := 1 to maxident do nameless[i] := chr(null);
  blocklevel := mainlevel - 2;
  newblock(nil);

  initstruct(typevoid, void);

  { standard files }
  declareobj(nameless, standardfile,
    inputfile);
  declareobj(nameless, standardfile,
    outputfile);

  { standard types }
  declordinaltype(boolean0, booltype, ord(false), ord(true),
    typeboolean);
  declordinaltype(char0, chartype, null, del,
    typechar);
  declordinaltype(integer0, inttype, minint, maxint,
    typeinteger);
  declstringtype(typestring);
  declrealtype(typereal);

  { standard constants }
  declconst(false0, typeboolean, ord(false));
  declconst(maxint0, typeinteger, maxint);
  declconst(maxstring0, typeinteger, maxstring);
  declconst(null0, typechar, null);
  declconst(true0, typeboolean, ord(true));

  { standard procedures }
  declstandard(open0, standardproc, p);
  declstandard(read0, standardproc, p);
  declstandard(readln0, standardproc, p);
  declstandard(receive0, standardproc, p);
  declstandard(send0, standardproc, p);
  declstandard(write0, standardproc, p);
  declstandard(writeln0, standardproc, p);
  { standard functions }
  declstandard(abs0, standardfunc, p);
  declstandard(arctan0, standardfunc, p);
  declstandard(chr0, standardfunc, p);
  declstandard(cos0, standardfunc, p);
  declstandard(eof0, standardfunc, p);
  declstandard(eoln0, standardfunc, p);
  declstandard(exp0, standardfunc, p);
  declstandard(ln0, standardfunc, p);
  declstandard(odd0, standardfunc, p);
  declstandard(ord0, standardfunc, p);
  declstandard(pred0, standardfunc, p);
  declstandard(random0, standardfunc, p);
  declstandard(round0, standardfunc, p);
  declstandard(sin0, standardfunc, p);
  declstandard(sqr0, standardfunc, p);
  declstandard(sqrt0, standardfunc, p);
  declstandard(succ0, standardfunc, p);
  declstandard(trunc0, standardfunc, p);
end;

     { import }

procedure resetimport
  (var importf: symbolfile; var found: boolean);
begin
  {$I-} reset(importf); {$I+}
  found := (ioresult = 0);
end;

procedure importbyte
  (var importf: symbolfile; var value: integer);
  var ch: char;
begin
  read(importf, ch);
  value := ord(ch)
end;

procedure importbool
  (var importf: symbolfile; var value: boolean);
  var ch: char;
begin
  read(importf, ch);
  value := ord(ch) = ord(true)
end;

procedure importint
  (var importf: symbolfile; var value: integer);
var dual: dualinteger;
begin
  dual.split := true;
  read(importf, dual.b);
  read(importf, dual.c);
  dual.split := false;
  value := dual.a;
end;

procedure importident
  (var importf: symbolfile; var name: identifier);
  var i, k: integer; s: char;
begin i := 1;
  repeat read(importf, s); name[i] := s; i := i + 1
  until s = chr(null);
  for k := i to maxident do name[k] := chr(null);
end;

procedure importstring
  (var importf: symbolfile; var str: stringg);
  var i, k: integer; s: char;
begin i := 1;
  str := nulstring;
  repeat read(importf, s);
    str[i] := s; i := i + 1
  until s = chr(null);
end;

procedure import(
    var impid {name}, extid: identifier;
    isbase: boolean; var imported: objectt);
  type
    fieldlist = array [0..maxnest] of objectt;
    msglist = array [0..maxnest] of messagenode;
  var
      importf: symbolfile;
      found, done: boolean;
      i, c, s, class, k: integer;
      noflcont, strno: integer;
      obj, ob0: objectt;
      str: structure;
      msg: messagenode;
      ch: char;
      si: integer;           
      contname: identifier;
      filename: stringg;
      loccont:  imptable;
      lastfld: fieldlist; fldlev: integer;
      lastmsg: msglist; msglev: integer;
      lastpar, lastenum: objectt;
        {if procedure types can be defined,
        lastpar must be an array, like lastfld}
      stringptr: stringpointer;

  procedure reverseobjects(p: objectt; var q: objectt);
    var r: objectt;
  begin q := nil;
    while p <> nil do begin
      r := p^.prev; p^.prev := q; q := p; p := r
    end
  end;

  procedure reversemessages(
    p: messagenode;
    var q: messagenode);
    var r: messagenode;
  begin q := nil;
    while p <> nil do begin
      r := p^.prev; p^.prev := q; q := p; p := r
    end
  end;

  procedure newglobal(isbase: boolean);
  begin
    new(obj);
    if nofgcont < maximps then begin
      glbcont[nofgcont].impmodule := obj;
      glbcont[nofgcont].isbase := isbase;
      nofgcont := nofgcont + 1
    end else
      haltt(maximport5);
    with obj^ do begin
      expno := notyetexp;
      id := contname;
      kind := module;
      {marked := false;}
      key := k;        {a1}
      impno:= nofgcont; {a0}
      lastcomp := nil; {dsc}
      xbase := nil;
    end;
  end;

procedure newlocal;
begin
  if noflcont < maximps then begin
    loccont[noflcont].impmodule := obj;
    noflcont := noflcont + 1
  end else
    error(maximport5)
end;

begin {import}
  noflcont := 0; strno := firststr;
  {parlev := -1; }fldlev := -1; msglev := -1;
  makefilename(extid, symext, filename);
  assign(importf, filename);
  resetimport(importf, found);
  if found then
    begin
      read(importf, ch);
      if ord(ch) = sftag then begin
        importint(importf, k);
        importident(importf, contname);
        newglobal(isbase);
        newlocal;
        done := eof(importf);
        while not done do begin { read next item from symbol file }
          {read(importf, ch); class := ord(ch);}
          importbyte(importf, class);
          case class of
            error6:    haltt(implementation5);
            constant6,
            enumconst6,
            typex6,
            typey6,
            variable6,
            method6:
              begin {objectt}
                new(obj); c := 0;
                obj^.marked := true;
                importbyte(importf, s);
                case class of
                  constant6:
                    begin
                      obj^.kind := constant;
                      obj^.typex := struct[s];
                      case obj^.typex^.form of
                        booltype, chartype, inttype:
                          importint(importf, obj^.value);
                        realtype:
                          begin
                            importint(importf, obj^.value);
                            importint(importf, obj^.value2)
                          end;
                        enumtype:
                          haltt(implementation5);
                        arraytype: {stringconst}
                          begin
                            new(stringptr);
                            importstring(importf, stringptr^);
                            storestring(obj, stringptr);
                         end
                      end
                    end;
                  enumconst6:
                    begin
                      obj^.typex := struct[s];
                      obj
                      ^.kind := constant;
                      importint(importf, obj^.value);
                    end;
                  typex6, typey6:
                    begin
                      obj^.typelevel := mainlevel;
                      obj^.kind := typ;
                      obj^.typex := struct[s];
                      {read(importf, ch); c := ord(ch);}
                      if obj^.typex^.obj = nil then
                        obj^.typex^.obj := obj;
                      obj^.marked := class = typex6;
{                      obj^.alloc/dealloc label := 0;  needed?}
                    end;
                  variable6:
                    begin
                      obj^.varlevel := mainlevel;
                      obj^.typex := struct[s];
                      obj^.kind := variable;
                      importint(importf, obj^.vardispl);
                    end;
                  method6:
                    begin
                      obj^.methodlevel := mainlevel;
                      obj^.typex := struct[s];
                      if obj^.typex^.resulttype = typevoid then
                        obj^.kind := xproc
                      else
                        obj^.kind := xfunc;
                      importint(importf, obj^.varlength);
                      importint(importf, obj^.methodlabel);
                      newclosure(obj^.implicit);
                    end;
                end;
                importbyte(importf, c);
                importident(importf, obj^.id);

                ob0 := loccont[c].impmodule^.lastcomp;
                found := false;
                while (ob0 <> nil) and not found do
                  if ob0^.id <> obj^.id then
                    ob0 := ob0^.prev
                  else
                    found := true;
                if not found then { insert objectt }
                  with loccont[c].impmodule^ do begin
                    obj^.prev := {loccont[c].impmodule^.}lastcomp;
                    {loccont[c].impmodule^.}lastcomp := obj
                  end
                else if obj^.kind = typ then
                  struct[s] := ob0^.typex
              end; { objectt }
            channel6,
            classx6,
            array6,
            methodstr6,
            classy6,
            enumtype6:
              begin {structure}
                newstruct(void, nil, str);
                read(importf, ch);
                str^.cno := loccont[ord(ch)].impmodule^.impno;
                if class = channel6 then
                  begin
                    str^.form := channeltype;
                    reversemessages(lastmsg[msglev], str^.messagelist);
                    {is reversing redundant?}
                    msglev := msglev - 1;
                  end
                else if class = classx6 then
                  begin
                    str^.form := xclasstype;
                    read(importf, ch);
                    str^.baseclass := struct[ord(ch)];
                    importint(importf, str^.recordlength);
                  end
                else if class = classy6 then
                  begin
                    str^.form := classtype;
                    read(importf, ch);
                    str^.baseclass := struct[ord(ch)];
                    importint(importf, str^.recordlength);
                    reverseobjects(lastfld[fldlev], str^.lastfield);
                    fldlev := fldlev - 1;
                  end
                else if class = array6 then
                  begin
                    str^.form := arraytype;
                    read(importf, ch);
                    str^.elementtype := struct[ord(ch)];
                    read(importf, ch);
                    str^.indextype := struct[ord(ch)];
                    importint(importf, str^.arraylength);
                    importint(importf, str^.upperbound);
                    importint(importf, str^.lowerbound);
                  end
                else if class = methodstr6 then
                  begin
                    str^.form := methodtype;
                    read(importf, ch);
                    str^.resulttype := struct[ord(ch)];
                    reverseobjects(lastpar{[parlev]}, str^.lastparam);
                    {parlev := parlev - 1;}
                    importint(importf, str^.paramlength);
                    importbool(importf, str^.restricted);
                  end
                else {class = enumtype6}
                  begin
                    str^.form := enumtype;
                    importint(importf, str^.minvalue);
                    importint(importf, str^.maxvalue);
                  end;
                struct[strno] := str; inc(strno);
                importint(importf, str^.dyncomps);
                importint(importf, str^.descr);
              end;
            paramlist6:
              lastpar := nil;
            varparam6, valueparam6:
              begin
                new(obj);
                if class = valueparam6 then
                  obj^.kind := valueparameter
                else obj^.kind := varparameter;
                read(importf, ch);
                obj^.typex := struct[ord(ch)];
                importint(importf, obj^.vardispl);
                importident(importf, obj^.id);
                obj^.varlevel := mainlevel + 1;
                obj^.prev := lastpar;
                lastpar := obj
              end;
            fieldlist6:
              begin
                If fldlev < maxnest-1 then
                  begin
                    fldlev := fldlev + 1;
                    lastfld[fldlev] := nil
                  end
                else haltt(maximport5);
              end;
            field6, hiddenfield6:
              begin
                new(obj);
                obj^.kind := field;
                read(importf, ch); obj^.typex := struct[ord(ch)];
                importint(importf, obj^.fielddispl);
                importident(importf, obj^.id);
                obj^.marked := (class = field6);
                obj^.prev := lastfld[fldlev];
                lastfld[fldlev] := obj
              end;
            messagelist6:
              begin
                if msglev < maxnest-1 then
                  begin
                    msglev := msglev + 1;
                    lastmsg[msglev] := nil
                  end
                else haltt(maximport5);
              end;
            message6:
              begin
                new(msg);
                read(importf, ch);
                msg^.typex := struct[ord(ch)];
                msg^.prev := lastmsg[msglev];
                lastmsg[msglev] := msg;
              end;
            sharedmodule6, basemodule6: { module anchor }
              begin
                importint(importf, k); importident(importf, contname);
                if class = basemodule6 then begin
                  newglobal(true {isbase});
                  ob0 := loccont[0].impmodule;
                  obj^.xbase := ob0^.xbase;
                  ob0^.xbase := obj;
                end else begin {class = sharedmodule6 }
                  i := 0;
                  while (i < nofgcont) and
                      ((contname <> glbcont[i].impmodule^.id) or
                       glbcont[i].isbase)
                  do
                    i := i + 1;
                  if i < nofgcont then { module already present }
                    begin
                      if k <> glbcont[i].impmodule^.key then
                        begin
                          haltt(version6);
                          done := true;
                        end;
                      obj := glbcont[i].impmodule;
                    end
                  else
                    newglobal(false {isbase});
                end; { shared module }
                newlocal;
              end;
            else
              begin
                haltt(implementation5);
                done := true
              end;
          end;
          done := eof(importf)
        end { while };
        if isbase then
          imported := loccont[0].impmodule
        else
          declareobj(impid, module, imported);
        with loccont[0].impmodule^ do begin
          imported^.marked := isbase;
          imported^.key := {loccont[0].impmodule^.}key;
          imported^.lastcomp := {loccont[0].impmodule^.}lastcomp;
          imported^.impno  := {loccont[0].impmodule^.}impno;
          imported^.typex := typevoid;
          imported^.xbase := {loccont[0].impmodule^.}xbase;
        end
      end else haltt(file6) { not a symbol file }
    end
  else haltt(file6)   { sym file not found }
end;

     { export }

procedure exportbyte(value: integer);
  var ch: char;
begin
  ch := chr(value);
  write(exportf, ch)
end;

procedure exportbool(value: boolean);
  var ch: char;
begin
  ch := chr(ord(value));
  write(exportf, ch)
end;

procedure exportint(value: integer);
var dual: dualinteger;
begin
  dual.split := false;
  dual.a := value;
  dual.split := true;
  write(exportf, dual.b);
  write(exportf, dual.c);
end;
                   
procedure exportident(var id: identifier);
  var i: integer; s: char;
begin i := 1;
  repeat s := id[i]; write(exportf, s); i := i + 1
  until s = chr(null)
end;

procedure exportstring(var str: stringg);
  var i: integer; s: char;
begin i := 1;
  repeat s := str[i]; write(exportf, s); i := i + 1
  until s = chr(null)
end;

procedure exportstruct(str: structure); forward;

procedure exportpars(par: objectt);
begin
  exportbyte(paramlist6);
  while (par <> nil) {and (par^.vardispl > 0)} do begin
    exportstruct(par^.typex);
    if par^.kind = valueparameter then
      exportbyte(valueparam6)
    else exportbyte(varparam6);
    exportbyte(par^.typex^.ref);
    exportint(par^.vardispl);
    exportident(par^.id);
    par := par^.prev
  end
end;

procedure exportfields(fld: objectt);
const
  nullident = null;

begin
  exportbyte(fieldlist6);
  while fld <> nil do begin
    with fld^, fld^.typex^ do
      if marked or dynamic(typex) then begin
        exportstruct(typex);
        if marked then
          exportbyte(field6)
        else
          exportbyte(hiddenfield6);
        exportbyte(ref);
        exportint(fielddispl);
        if marked then
          exportident(id)
        else
          exportbyte(nullident);
      end;
    fld := fld^.prev;
  end
end;

procedure exportmessages(
  msg: messagenode);
begin
  exportbyte(messagelist6);
  while msg <> nil do begin
    exportstruct(msg^.typex);
    exportbyte(message6);
    exportbyte(msg^.typex^.ref);
    msg := msg^.prev
  end
end;

procedure exportanchor(
  var c: integer; isbase: boolean);
var em: integer; cont: objectt;
begin
  if c > 0 then begin
    cont := glbcont[c-1].impmodule;
    em := cont^.expno;
    if em = notyetexp then begin
      glbcont[c-1].impmodule^.expno := nofexp;
      c := nofexp; nofexp := nofexp + 1;
      if isbase then exportbyte(basemodule6)
      else exportbyte(sharedmodule6);
      exportint(cont^.key);
      exportident(cont^.id)
    end else c := em
  end
end;

procedure exportstruct(str: structure);
  var c: integer;
    baseclass: structure;
    visibletype: boolean;
    enum: objectt;
begin
  if str <> nil then
    if str^.ref = 0 then begin
      c := str^.cno;
      if c > 0 then
        exportanchor(c, {isbase} false);
      if str^.obj <> nil then
        visibletype :=
          (str^.obj^.kind = typ) and str^.obj^.marked;
      case str^.form of
        enumtype:
          begin
            exportbyte(enumtype6);
            exportbyte(c);
            exportint(str^.minvalue);
            exportint(str^.maxvalue);
          end;
        channeltype:
          begin
            exportmessages(str^.messagelist);
            exportbyte(channel6);
            exportbyte(c);
          end;
        arraytype:
          begin
            exportstruct(str^.elementtype);
            exportstruct(str^.indextype);
            exportbyte(array6);
            exportbyte(c);
            exportbyte(str^.elementtype^.ref);
            exportbyte(str^.indextype^.ref);
            exportint(str^.arraylength);
            exportint(str^.upperbound);
            exportint(str^.lowerbound);
          end;
        classtype:
          begin
            exportstruct(str^.baseclass);
            exportfields(str^.lastfield);
            exportbyte(classy6);
            exportbyte(c);
            exportbyte(str^.baseclass^.ref);
            exportint(str^.recordlength);
          end;
        xclasstype:
          begin
            exportstruct(str^.baseclass);
            exportbyte(classx6);
            exportbyte(c);
            exportbyte(str^.baseclass^.ref);
            exportint(str^.recordlength);
          end;
        methodtype:
          begin
            exportstruct(str^.resulttype);
            exportpars(str^.lastparam);
            exportbyte(methodstr6);
            exportbyte(c);
            exportbyte(str^.resulttype^.ref);
            exportint(str^.paramlength);
            exportbool(str^.restricted);
          end;
      end;
      exportint(str^.dyncomps);
      exportint(str^.descr);
      if str^.obj <> nil then
        if str^.obj^.kind = typ then begin
          if str^.obj^.marked  then
            exportbyte(typex6)
          else
            exportbyte(typey6);
          exportbyte(strno);
          exportbyte(c);
          exportident(str^.obj^.id)
        end;
      str^.ref := strno; strno := strno + 1;
      if strno > maxstruct then haltt(maxexport5)
    end
end;

procedure exportobjects(
  xcont: objectt);
var
  str: structure; obj: objectt;
  f: structform; k: objectkind;
  c: integer; stringvalue: stringg;
begin { xcont^.kind = module }
  if xcont^.kind <> module then
    haltt(implementation5);
  if xcont^.xbase <> nil then begin
    {export base}
    exportanchor(
      xcont^.xbase^.impno, {isbase} true);
    exportobjects(xcont^.xbase);
  end;
  c := xcont^.impno;
  obj := xcont^.lastcomp;
  while obj <> nil do begin
    k := obj^.kind;
    if obj^.marked then
      case k {obj^.kind} of
        constant:
          begin
            f := obj^.typex^.form;
            if f = enumtype then
              with obj^ do begin
                exportstruct(typex);
                exportbyte(enumconst6);
                exportbyte(typex^.ref);
                exportint(value);
              end
          else begin
              exportbyte(constant6);
              exportbyte(ord(f));
              case f of
                void: ;
                booltype, chartype, inttype:
                  exportint(obj^.value);
                realtype:
                  begin
                    exportint(obj^.value);
                    exportint(obj^.value2);
                  end;
                arraytype {stringconst}:
                  begin
                    loadstring(obj, stringvalue);
                    exportstring(stringvalue);
                  end;
              end;
            end;
            exportbyte(c);
            exportident(obj^.id)
          end;
        typ:
          begin
            exportstruct(obj^.typex);
            if (obj^.typex^.obj <> obj)
                 {and (obj^.typex^.obj <> nil) redundant?} then
              begin
                exportbyte(typex6); exportbyte(obj^.typex^.ref);
                exportbyte(c); exportident(obj^.id)
              end
          end;
        variable:
          with obj^ do begin
            exportstruct(typex);
            exportbyte(variable6);
            exportbyte(typex^.ref);
            exportint(vardispl);
            exportbyte(c);
            exportident(id)
          end;
        xproc, xfunc:
          begin
            str := obj^.typex;
            exportstruct(str);
            exportbyte(method6);
            exportbyte(str^.ref);
            exportint(obj^.varlength);
            if obj^.typex^.cno > 0 then
              exportint(obj^.methodlabel)
            else
              exportint(table[obj^.methodlabel]);
            exportbyte(c);
            exportident(obj^.id)
          end;
        standardfile, standardfunc, standardproc,
        valueparameter, varparameter, undefined,
        field, funktion, procedur, module:
          haltt(implementation5);
      end;
    obj := obj^.prev
  end
end;

procedure export
  {
  (mainmodule: objectt)
  };

  var
    filename: stringg;
{
    i: integer;
    oldkey: integer;
}
    typex: structure;

{
  procedure register;
    var b, oldb: char;
      fileexists, samefiles, done: boolean;
      newexport: symbolfile;

    procedure skipbytes(
        var sf: symbolfile; n: integer);
      var ch: char;
    begin
      while n > 0 do begin
        read(sf, ch); n := n - 1
      end
    end;

  begin / register /
    makefilename(modulename, symext, filename);
    assign(oldexpf, filename);
    resetimport(oldexpf, fileexists);
    if fileexists then
      begin / compare to the new one /
        reset(exportf);
        skipbytes(oldexpf, 1); / skip sftag /
        importint(oldexpf, oldkey);
        skipbytes(exportf, 1 + intbytes); / skip sftag and newkey /
        done := eof(oldexpf) or eof(exportf);
        while not done do begin
          read(oldexpf, oldb); read(exportf, b);
          done := (b <> oldb) or eof(oldexpf) or eof(exportf);
        end;
        samefiles := eof(exportf) and eof(oldexpf);
        close(oldexpf)
      end
    else samefiles := false; / because old file does not exist /
    if samefiles then
      mainmodule^.key := oldkey
    else
      begin
        assign(newexport, filename);
        rewrite(newexport); reset(exportf);
        repeat read(exportf, b); write(newexport, b);
        until eof(exportf);
        close(newexport);
      end;
  end;
}

begin
  { nonstandard assign }
{
  assign(exportf, tempfile);
}

  makefilename(modulename, symext, filename);
  assign(exportf, filename);
  rewrite(exportf);
  exportbyte(sftag);
  strno := firststr;
  nofexp := notyetexp + 1;
  exportint(mainmodule^.key);
  exportident(modulename);
  exportobjects(mainmodule);
{
  register;
}
  close(exportf)
end;

procedure initimpex;
begin
  struct[ord(arraytype)]   := typestring;
  struct[ord(booltype)]    := typeboolean;
  struct[ord(chartype)]    := typechar;
  struct[ord(channeltype)] := nil;  {nils should be removed after testing }
  struct[ord(enumtype)]    := nil;
  struct[ord(inttype)]     := typeinteger;
  struct[ord(methodtype)]  := nil;
  struct[ord(realtype)]    := typereal;
  struct[ord(classtype)]   := nil;
  struct[ord(xclasstype)]  := nil;
  struct[ord(void)]   := typevoid;
end;

procedure inittexim;
begin
  initwords;
  inittables;
  initimpex;
end;

end.